
import cPickle
import gzip
import os
import sys
import time
import cv2
import random
import subprocess as sp
import numpy

from scipy import optimize

from PIL import Image

import theano 
import theano.tensor as T

import deep
from deep import LogisticRegression
from deep import MLP, HiddenLayer

def normalize(X, low, high, dtype=None):
	"""Normalizes a given array in X to a value between low and high."""
	X = numpy.asarray(X)
	minX, maxX = numpy.min(X), numpy.max(X)
	# normalize to [0...1].
	X = X - float(minX)
	X = X / float((maxX - minX))
	# scale to [low...high].
	X = X * (high-low)
	X = X + low
	if dtype is None:
		return numpy.asarray(X)
	return numpy.asarray(X, dtype=dtype)

def shared_dataset(data_xy, borrow=True):
	""" Function that loads the dataset into shared variables

	The reason we store our dataset in shared variables is to allow
	Theano to copy it into the GPU memory (when code is run on GPU).
	Since copying data into the GPU is slow, copying a minibatch everytime
	is needed (the default behaviour if the data is not in a shared
	variable) would lead to a large decrease in performance.
	"""
	data_x, data_y = data_xy    
	shared_x = theano.shared(numpy.asarray(data_x,
										   dtype=theano.config.floatX),
							 borrow=borrow)
	shared_y = theano.shared(numpy.asarray(data_y,
										   dtype=theano.config.floatX),
							 borrow=borrow)
	# When storing data on the GPU it has to be stored as floats
	# therefore we will store the labels as ``floatX`` as well
	# (``shared_y`` does exactly that). But during our computations
	# we need them as ints (we use labels as index, and if they are
	# floats it doesn't make sense) therefore instead of returning
	# ``shared_y`` we will have to cast it to int. This little hack
	# lets ous get around this issue
	return shared_x, T.cast(shared_y, 'int32')


def read_data_dir(path, splitsize=None, splitstart=None, nclasses=None, sz=None):
	"""Reads the images in a given folder, resizes images on the fly if size is given.

	Args:
		path: Path to a folder with subfolders representing the subjects (persons).
		splitsize: Number of images from each directory to read
		splitstart: Offset within each directory to start reading images
		nclasses: Number of image subdirectories (classes) to read
		sz: A tuple with the size Resizes

	Returns:
		A list [X,y]

			X: The images, which is a Python list of numpy arrays.
			y: The corresponding labels (the unique number of the subject, person) in a Python list.
	"""
	c = 0
	X,y = [], []	
	for dirname, dirnames, filenames in os.walk(path):
		ctr1 = 0
		for subdirname in dirnames:
			if nclasses is not None and ctr1 >= nclasses:
				break
			ctr1 += 1
			ctr2 = 0
			subject_path = os.path.join(dirname, subdirname)
			for filename in os.listdir(subject_path):
				if filename.startswith(".") or not (filename.endswith(".png") or filename.endswith(".pgm")):
					continue				
				if splitstart is not None and ctr2 < splitstart:
					ctr2 += 1
					continue
				if splitsize is not None and splitstart is not None and ctr2 >= splitsize+splitstart:
					break
				ctr2 += 1				
				try:
					im = cv2.imread(os.path.join(subject_path, filename), cv2.IMREAD_GRAYSCALE)
					# resize to given size (if given)
					if (sz is not None):
						im = cv2.resize(im, sz)
					X.append(normalize(numpy.asarray(im, dtype=numpy.uint8).flatten(), 0, 1, theano.config.floatX))					
					y.append(c)
				except IOError, (errno, strerror):
					print "I/O error({0}): {1}".format(errno, strerror)
				except:
					print "Unexpected error:", sys.exc_info()[0]
					raise
			c = c+1
	return [numpy.asarray(X, dtype=theano.config.floatX), numpy.asarray(y, dtype=theano.config.floatX)]

def data_dir_shape(path):
	for dirname, dirnames, filenames in os.walk(path):
		for subdirname in dirnames:
			subject_path = os.path.join(dirname, subdirname)
			for filename in os.listdir(subject_path):
				if filename.startswith(".") or not (filename.endswith(".png") or filename.endswith(".pgm")):
					continue
				try:
					im = cv2.imread(os.path.join(subject_path, filename), cv2.IMREAD_GRAYSCALE)					
					return im.shape
				except IOError, (errno, strerror):
					print "I/O error({0}): {1}".format(errno, strerror)
				except:
					print "Unexpected error:", sys.exc_info()[0]
					raise
	raise Exception('Could not get shape in %s' % path)

def make_status_image(attacker_raw, mask, attacker_noised, victim, shape, scale=1):

	height, width = shape	
	combined = numpy.concatenate((normalize(attacker_raw.reshape(shape), 0, 255, dtype=numpy.uint8), normalize(mask.reshape(shape), 0, 255, dtype=numpy.uint8), normalize(attacker_noised.reshape(shape), 0, 255, dtype=numpy.uint8), normalize(victim.reshape(shape), 0, 255, dtype=numpy.uint8)), axis=1)

	return cv2.resize(normalize(combined, 0, 255, dtype=numpy.uint8), (scale*4*width, scale*height))

def mask_fn(idx, img, shape):
	
	height, width = shape

	width_cons = True	
	height_cons = img[0,idx] >= 0.5

	# width_cons = idx%width > width/7 and idx%width < 5*width/6
	# height_cons = idx/width > height/7 and idx/width < 6*height/7

	# width_cons = True
	# height_cons = True

	return width_cons and height_cons or True

def open_video(shape):

	height, width = shape

	FFMPEG_BIN = "ffmpeg"
	command = [ FFMPEG_BIN,
			'-y', # (optional) overwrite output file if it exists
			'-f', 'rawvideo',
			'-vcodec','rawvideo',
			'-pix_fmt', 'gray',
			'-s', '%dx%d' % (1*4*width, 1*height), # size of one frame			
			# '-r', '24', # frames per second
			'-i', '-', # The imput comes from a pipe
			'-an', # Tells FFMPEG not to expect any audio
			'-vcodec', 'mpeg4', 
			'attack.mp4' ]
	pipe = sp.Popen(command, stdin=sp.PIPE)
	print dir(pipe)

	return pipe

def balance_image(image, phi=1, theta=1):

	image = (255.0/phi)*(image/(255.0/theta))**0.5
	return numpy.asarray(cam_face,dtype=numpy.uint8)	

def run_attack3(data_dir="../images/political", nclass=2, splitsize=5, target_index=3, attack_label=1, attack_rate=100, attack_patience=100000, attack_reg=0, make_video=False):

	im_shape = data_dir_shape(data_dir)
	height, width = im_shape

	print im_shape

	pipe = None
	if make_video:
		pipe = open_video(im_shape)

	# train_set_x, train_set_y = shared_dataset(read_data_dir("../images/att_faces", splitsize=splitsize, splitstart=0, nclasses=nclass))
	# valid_set_x, valid_set_y = shared_dataset(read_data_dir("../images/att_faces", splitsize=splitsize, splitstart=0, nclasses=nclass))
	# test_set_x, test_set_y = shared_dataset(read_data_dir("../images/att_faces", splitsize=splitsize, splitstart=splitsize, nclasses=nclass))        

	train_set_x, train_set_y = shared_dataset(read_data_dir("../images/political", splitsize=5, splitstart=0, nclasses=2))
	valid_set_x, valid_set_y = shared_dataset(read_data_dir("../images/political", splitsize=5, splitstart=0, nclasses=2))
	test_set_x, test_set_y = shared_dataset(read_data_dir("../images/political", splitsize=5, splitstart=5, nclasses=2))

	# optional - read file saved from webcam
	# cam_face = cv2.resize(cv2.imread("webcam_image.png", cv2.IMREAD_GRAYSCALE), (width, height))	
	# cam_face = theano.shared(value=numpy.asarray(normalize(numpy.asarray(cam_face, dtype=numpy.uint8), 0, 1, theano.config.floatX).reshape(1, height*width), dtype=theano.config.floatX), borrow=True)

	cv2.startWindowThread() 

	mfile = file('mlp_model.pkl', 'rb')
	classifier = cPickle.load(mfile)
	mfile.close()

	print "\n... building attack model\n"

	i = T.scalar('i', dtype='int32')
	reg = T.scalar('reg', dtype='floatX')
	rvar = T.matrix('rvar')
	x = classifier.input
	y = T.vector('y')	

	victim_idx = attack_label*splitsize+1
	victim_image = theano.function([], train_set_x[victim_idx:victim_idx+1])()	
	target = test_set_x[target_index:target_index+1]	
	original_image = theano.function([], target)()
	target_shape = theano.function([], target.shape)()    	
	#target = theano.shared(value=numpy.random.sample(height*width).astype(theano.config.floatX).reshape(target_shape), borrow=True)
	target = theano.shared(value=0.5*numpy.ones(target_shape, dtype=theano.config.floatX), borrow=True)
	r = theano.shared(value=numpy.zeros(target_shape, dtype=theano.config.floatX), name='r', borrow=True)	
	#r = theano.shared(value=numpy.random.sample(height*width).astype(theano.config.floatX).reshape(target_shape), name='r', borrow=False)	

	mask_native = [1.0 if mask_fn(j, original_image, im_shape) else 0.0 for j in range(0,height*width)]
	mask = theano.shared(value=numpy.asarray(mask_native, dtype=theano.config.floatX), name='mask', borrow=False)
	
	orig_label = theano.function([], test_set_y[target_index])()	

	p_y_given_xr = theano.clone(classifier.logRegressionLayer.p_y_given_x, replace={x: T.clip(x+(mask*rvar), 0, 1)})	
	nll = -T.log(p_y_given_xr)[0][i]
	attack_cost = T.sqr(nll-0.5) - reg*T.mean(T.sqr(rvar))
	current_label = theano.function([], T.argmax(p_y_given_xr, axis=1), givens={x: target, rvar: r})

	current_label_t = theano.function([x], T.argmax(p_y_given_xr, axis=1), givens={rvar: r})

	cost_f = theano.function(inputs=[i, reg], 
			outputs=attack_cost,
			givens={ rvar: r, x: target })        
	classprob = theano.function(inputs=[i], 
			outputs=p_y_given_xr[0][i],
			givens={rvar: r, x: target })        

	ag_r = T.grad(cost=attack_cost, wrt=rvar)

	# attack_updates = [(r, T.clip(r - attack_rate*ag_r, 0, 1))]
	attack_updates = [(r, mask*(r - attack_rate*ag_r))]
	derive_attack = theano.function(inputs=[i, reg],
			outputs=attack_cost,
			updates=attack_updates,
			givens={ rvar: r, x: target })

	print "original class label is:", orig_label
	print "\n\n... running attack\n"

	start_time = time.clock()
	reg_epoch = 500
	cur_reg_epoch = 0
	last_correct_image = None
	last_correct_mask = None
	last_correct_cost = None
	backwards_time = 0
	backwards_cutoff = 1000
	for iter in range(0,attack_patience):

		attack_image_data = theano.function([], T.clip(target + r, 0, 1))()
		attack_mask_data = theano.function([], r)()
		status_image = make_status_image(original_image, attack_mask_data, attack_image_data, victim_image, im_shape)		
		if iter % 10 == 0 and make_video:
			pipe.stdin.write( status_image.tostring() )

		# run the next step of the attack
		cur_cost = derive_attack(attack_label, attack_reg)
		# r_actual = cv2.bilateralFilter(r.get_value().reshape(im_shape), 3, 3, 3).reshape(target_shape)
		#r_actual = cv2.medianBlur(r.get_value().reshape(im_shape), 5).reshape(target_shape)
		#r_actual = cv2.GaussianBlur(r.get_value().reshape(im_shape), (3,3), 0).reshape(target_shape)
		#r_actual = cv2.blur(r.get_value().reshape(im_shape), (7,7)).reshape(target_shape)
		# r.set_value(r_actual)

		# every so many iterations, update the status information and display the current attack image
		if iter % 100 == 0:			
			cv2.destroyAllWindows()
			cv2.imshow("iter", status_image)
			sys.stdout.flush()

			print "iter=%d, class prob=%.3f, cost=%.3f, class label=%d, reg=%f" % (iter, classprob(attack_label), cost_f(attack_label, attack_reg), current_label(), attack_reg)
		
		if last_correct_cost is not None and cur_cost > last_correct_cost:
			backwards_time += 1		
		if backwards_time >= backwards_cutoff and False:
			break

		# if we have been successful so far with the attack, increase the regularization constant
		# to encourage less overall noise in the attack image
		if current_label() == attack_label:
			
			if last_correct_cost is None:
				last_correct_cost = cur_cost
			if last_correct_cost >= cur_cost:
				last_correct_image = theano.function([], T.clip(target + r, 0, 1))()
				last_correct_mask = theano.function([], r)()
				last_correct_cost = cur_cost
				backwards_time = 0

			if cur_reg_epoch == 0:

				# lop off some of the noise
				# mask.set_value([random.randint(0, 1) for j in range(0,height*width)])

				attack_reg = min(2*attack_reg, 512)
				cur_reg_epoch = reg_epoch
				# print "increased attack_reg to", attack_reg
			
			cur_reg_epoch -= 1

		elif current_label != attack_label and attack_reg > 2:
			attack_reg = max(2, attack_reg/1.1)
			cur_reg_epoch = reg_epoch
			# print "backing off regularizer; new value =", attack_reg

	end_time = time.clock()
	print '\n\nAttack ran for %.1fs' % (end_time - start_time)
	print "After attack, class is:", current_label()

	# as a final sanity check, re-normalize and construct the attack image, and run it
	# through the original classifier to see what its predicted label is
	attack_image_data = theano.function([], T.clip(target + r, 0, 1))() if last_correct_image is None else last_correct_image
	attack_mask_data = theano.function([], r)() if last_correct_mask is None else last_correct_mask
	after_label = theano.function([], T.argmax(classifier.logRegressionLayer.p_y_given_x, axis=1), givens={x: attack_image_data})()	
	attack_image_data.resize(im_shape)
	rescaled = normalize(attack_image_data, 0, 255, dtype=numpy.uint8)

	print "\n\nAfter image reconstruction, class is:", after_label
	print "SUCCESS!" if after_label == attack_label else "ATTACK FAILED"

	# and save the final status image to disk
	status_image = make_status_image(original_image, last_correct_mask, attack_image_data, victim_image, im_shape)
	im = Image.fromarray(status_image)
	im.save('attack-result.png')

	if make_video:
		pipe.terminate()

def run_attack2(data_dir="../images/political", nclass=2, splitsize=5, target_index=3, attack_label=0, attack_rate=1, attack_patience=100000, attack_reg=2, make_video=False):

	im_shape = data_dir_shape(data_dir)
	height, width = im_shape

	print im_shape

	pipe = None
	if make_video:
		pipe = open_video(im_shape)

	# train_set_x, train_set_y = shared_dataset(read_data_dir("../images/att_faces", splitsize=splitsize, splitstart=0, nclasses=nclass))
	# valid_set_x, valid_set_y = shared_dataset(read_data_dir("../images/att_faces", splitsize=splitsize, splitstart=0, nclasses=nclass))
	# test_set_x, test_set_y = shared_dataset(read_data_dir("../images/att_faces", splitsize=splitsize, splitstart=splitsize, nclasses=nclass))        

	train_set_x, train_set_y = shared_dataset(read_data_dir("../images/political", splitsize=5, splitstart=0, nclasses=2))
	valid_set_x, valid_set_y = shared_dataset(read_data_dir("../images/political", splitsize=5, splitstart=0, nclasses=2))
	test_set_x, test_set_y = shared_dataset(read_data_dir("../images/political", splitsize=5, splitstart=5, nclasses=2))

	# optional - read file saved from webcam
	# cam_face = cv2.resize(cv2.imread("webcam_image.png", cv2.IMREAD_GRAYSCALE), (width, height))	
	# cam_face = theano.shared(value=numpy.asarray(normalize(numpy.asarray(cam_face, dtype=numpy.uint8), 0, 1, theano.config.floatX).reshape(1, height*width), dtype=theano.config.floatX), borrow=True)

	cv2.startWindowThread() 

	mfile = file('mlp_model.pkl', 'rb')
	classifier = cPickle.load(mfile)
	mfile.close()

	print "\n... building attack model\n"

	i = T.scalar('i', dtype='int32')
	reg = T.scalar('reg', dtype='floatX')
	rvar = T.matrix('rvar')
	x = classifier.input
	y = T.vector('y')	

	victim_idx = attack_label*splitsize+1
	victim_image = theano.function([], train_set_x[victim_idx:victim_idx+1])()	
	target = test_set_x[target_index:target_index+1]	
	original_image = theano.function([], target)()
	target_shape = theano.function([], target.shape)()    		
	r = theano.shared(value=numpy.zeros(target_shape, dtype=theano.config.floatX), name='r', borrow=True)	
	#r = theano.shared(value=numpy.random.sample(height*width).astype(theano.config.floatX).reshape(target_shape), name='r', borrow=False)	

	mask_native = [1.0 if mask_fn(j, original_image, im_shape) else 0.0 for j in range(0,height*width)]
	mask = theano.shared(value=numpy.asarray(mask_native, dtype=theano.config.floatX), name='mask', borrow=False)
	
	orig_label = theano.function([], test_set_y[target_index])()	

	p_y_given_xr = theano.clone(classifier.logRegressionLayer.p_y_given_x, replace={x: T.clip(x+(mask*rvar), 0, 1)})	
	nll = -T.log(p_y_given_xr)[0][i]
	attack_cost = nll + reg*T.mean(rvar**2)	
	current_label = theano.function([], T.argmax(p_y_given_xr, axis=1), givens={x: target, rvar: r})

	current_label_t = theano.function([x], T.argmax(p_y_given_xr, axis=1), givens={rvar: r})

	cost_f = theano.function(inputs=[i, reg], 
			outputs=attack_cost,
			givens={ rvar: r, x: target })        
	classprob = theano.function(inputs=[i], 
			outputs=p_y_given_xr[0][i],
			givens={rvar: r, x: target })        

	ag_r = T.grad(cost=attack_cost, wrt=rvar)

	# attack_updates = [(r, T.clip(r - attack_rate*ag_r, 0, 1))]
	attack_updates = [(r, mask*(r - attack_rate*ag_r))]
	derive_attack = theano.function(inputs=[i, reg],
			outputs=attack_cost,
			updates=attack_updates,
			givens={ rvar: r, x: target })

	print "original class label is:", orig_label
	print "\n\n... running attack\n"

	start_time = time.clock()
	reg_epoch = 200
	cur_reg_epoch = 0
	last_correct_image = None
	last_correct_mask = None
	last_correct_cost = None
	backwards_time = 0
	backwards_cutoff = 500
	for iter in range(0,attack_patience):

		attack_image_data = theano.function([], T.clip(target + r, 0, 1))()
		attack_mask_data = theano.function([], r)()
		status_image = make_status_image(original_image, attack_mask_data, attack_image_data, victim_image, im_shape)		
		if iter % 10 == 0 and make_video:
			pipe.stdin.write( status_image.tostring() )

		# run the next step of the attack
		derive_attack(attack_label, attack_reg)
		#r_actual = cv2.bilateralFilter(r.get_value().reshape(im_shape), 7, 5, 5).reshape(target_shape)
		#r_actual = cv2.medianBlur(r.get_value().reshape(im_shape), 5).reshape(target_shape)
		#r_actual = cv2.GaussianBlur(r.get_value().reshape(im_shape), (9,9), 0).reshape(target_shape)
		#r_actual = cv2.blur(r.get_value().reshape(im_shape), (7,7)).reshape(target_shape)
		#r.set_value(r_actual)

		# every so many iterations, update the status information and display the current attack image
		if iter % 100 == 0:			
			cv2.destroyAllWindows()
			cv2.imshow("iter", status_image)
			sys.stdout.flush()

			print "iter=%d, class prob=%.3f, cost=%.3f, class label=%d, reg=%f" % (iter, classprob(attack_label), cost_f(attack_label, attack_reg), current_label(), attack_reg)

		cur_prob = classprob(attack_label)		
		if last_correct_cost is not None and cur_prob > last_correct_cost:
			backwards_time += 1		
		if backwards_time >= backwards_cutoff:
			break

		# if we have been successful so far with the attack, increase the regularization constant
		# to encourage less overall noise in the attack image
		if current_label() == attack_label:
			
			if last_correct_cost is None:
				last_correct_cost = cur_prob
			if last_correct_cost >= cur_prob:
				last_correct_image = theano.function([], T.clip(target + r, 0, 1))()
				last_correct_mask = theano.function([], r)()
				last_correct_cost = cur_prob
				backwards_time = 0

			if cur_reg_epoch == 0:

				# lop off some of the noise
				# mask.set_value([random.randint(0, 1) for j in range(0,height*width)])

				attack_reg *= 2
				cur_reg_epoch = reg_epoch
				# print "increased attack_reg to", attack_reg
			
			cur_reg_epoch -= 1

		elif current_label != attack_label and attack_reg > 2:
			attack_reg = max(2, attack_reg/1.1)
			cur_reg_epoch = reg_epoch
			# print "backing off regularizer; new value =", attack_reg

	end_time = time.clock()
	print '\n\nAttack ran for %.1fs' % (end_time - start_time)
	print "After attack, class is:", current_label()

	# as a final sanity check, re-normalize and construct the attack image, and run it
	# through the original classifier to see what its predicted label is
	attack_image_data = theano.function([], T.clip(target + r, 0, 1))() if last_correct_image is None else last_correct_image
	attack_mask_data = theano.function([], r)() if last_correct_mask is None else last_correct_mask
	after_label = theano.function([], T.argmax(classifier.logRegressionLayer.p_y_given_x, axis=1), givens={x: attack_image_data})()	
	attack_image_data.resize(im_shape)
	rescaled = normalize(attack_image_data, 0, 255, dtype=numpy.uint8)

	print "\n\nAfter image reconstruction, class is:", after_label
	print "SUCCESS!" if after_label == attack_label else "ATTACK FAILED"

	# and save the final status image to disk
	status_image = make_status_image(original_image, last_correct_mask, attack_image_data, victim_image, im_shape)
	im = Image.fromarray(status_image)
	im.save('attack-result.png')

	if make_video:
		pipe.terminate()

def run_attack(dataset='../images/mnist.pkl.gz'):	

	datasets = deep.load_data(dataset)

	train_set_x, train_set_y = datasets[0]
	valid_set_x, valid_set_y = datasets[1]
	test_set_x, test_set_y = datasets[2]

	mfile = file('lr_model.pkl', 'rb')
	classifier = cPickle.load(mfile)
	mfile.close()

	print   	 

	print "\n... building attack model\n"
	i = T.scalar('i', dtype='int32')
	rvar = T.vector('rvar')


	target_index = 100
	target = test_set_x[target_index]
	r = theano.shared(value=numpy.ones((28 * 28,),
						dtype=theano.config.floatX),
						name='r', borrow=True)
	rprint = theano.function([], theano.printing.Print('real class label is: ')(test_set_y[target_index]))

	attack_image = T.clip(target+rvar, 0, 1)
	attack_cost = -1*classifier.p_y_i(attack_image, i) + 15*T.mean((rvar ** 2))
	cost_f = theano.function(inputs=[i, rvar], outputs=attack_cost)
	classprob = theano.function(inputs=[i], 
					outputs=classifier.p_y_i(attack_image, i),
					givens={rvar: r},
					on_unused_input='ignore')    

	attack_label = 1
	attack_rate = 0.1
	attack_patience = 100000

	# def obj_f(rvect):

	ag_r = T.grad(cost=attack_cost, wrt=rvar)
	#attack_updates = [(r, T.clip(r - attack_rate*ag_r, 0, 1))]
	attack_updates = [(r, r - attack_rate*ag_r)]
	derive_attack = theano.function(inputs=[i],
			outputs=attack_cost,
			updates=attack_updates,
			givens={rvar: r})

	rprint()
	print
	print "initial class probabilities:"
	for iv in range(0,10):
		print "[%d, %f]," % (iv, classprob(iv)),

	print "\n\n... running attack\n"
	start_time = time.clock()
	for iter in range(0,attack_patience):
		if iter % 10000 == 0:
			print "(%d, %.3f, %.3f)," % (iter, classprob(attack_label), cost_f(attack_label, r.get_value())),
			sys.stdout.flush()
		derive_attack(attack_label)    
	print
	print

	print "final class probabilities:"
	for iv in range(0,10):
		print "[%d, %f]," % (iv, classprob(iv)),

	end_time = time.clock()
	print
	print
	print >> sys.stderr, ('Attack ran for %.1fs' % ((end_time - start_time)))

	print "\n\nAfter attack, class is:", theano.function([], T.argmax(T.nnet.softmax(T.dot(T.clip(target + r, 0, 1), classifier.W) + classifier.b), axis=1))()

	# print
	# print
	# theano.function([], theano.printing.Print()(target))()
	# print
	# print
	# theano.function([], theano.printing.Print()(target+r))()
	# print
	# print
	# theano.function([], theano.printing.Print()(T.mean(r ** 2)))()
	# theano.function([], theano.printing.Print()(r))()

	attack_image_data = theano.function([], T.clip(target + r, 0, 1))()
	attack_image_data.resize((28,28))
	rescaled = (255.0 / attack_image_data.max() * (attack_image_data - attack_image_data.min())).astype(numpy.uint8)

	im = Image.fromarray(rescaled)
	im.save('attack.png')    


	original_image = theano.function([], target)()
	original_image.resize((28,28))
	rescaled = (255.0 / original_image.max() * (original_image - original_image.min())).astype(numpy.uint8)

	im = Image.fromarray(rescaled)
	im.save('original.png')    

if __name__ == '__main__':
	run_attack()	