import sys
import csv
import math
import json
import pdb
import random
import itertools

import numpy as np
import numpy.random

import treeinversion

from sklearn import metrics
from bigml.model import Model
from bigml.api import BigML

class BigMLModelInfo(treeinversion.ModelInfo):

  def __init__(self, model_object):

    self.training_size = model_object["max_rows"] # yes, this is overwritten. not sure which to use yet
    self.model = Model(model_object)
    self.objective_field = model_object["objective_field"]

    obj = model_object["model"]
    self.field_names = [obj["fields"][fkey]["name"] for fkey in obj["fields"].keys()]

    self.training_size = sum([cat[1] for cat in obj["distribution"]["training"]["categories"]])
    self.name_to_id = {obj["fields"][fkey]["name"]: fkey for fkey in obj["fields"].keys()}
    self.id_to_name = {fkey: obj["fields"][fkey]["name"] for fkey in obj["fields"].keys()}
    self.root = obj["root"]
    self.objective_name = self.id_to_name[self.objective_field]
    self.objective_index = sum([i for i in range(len(self.field_names)) if self.field_names[i] == self.objective_name])

    self.marginals = {obj["fields"][fkey]["name"]: None for fkey in obj["fields"].keys()}
    self.field_values = {obj["fields"][fkey]["name"]: None for fkey in obj["fields"].keys()}
    self.splits = {obj["fields"][fkey]["name"]: None for fkey in obj["fields"].keys()}
    self.types = {obj["fields"][fkey]["name"]: None for fkey in obj["fields"].keys()}

    for fkey in obj["fields"].keys():

      name = obj["fields"][fkey]["name"]
      self.types[name] = obj["fields"][fkey]["optype"]

      if obj["fields"][fkey]["optype"] == "numeric":

        descriptor = "bins" if obj["fields"][fkey]["summary"].has_key("bins") else "counts"
        converter = lambda x: float(x)
        self.splits[name] = obj["fields"][fkey]["summary"]["splits"][:] if obj["fields"][fkey]["summary"].has_key("splits") else None
        self.field_values[name] = [converter(cat[0]) for cat in obj["fields"][fkey]["summary"][descriptor]]
        tot = sum([float(cat[1]) for cat in obj["fields"][fkey]["summary"][descriptor]])
        self.marginals[name] = {cat[0]: float(cat[1])/tot for cat in obj["fields"][fkey]["summary"][descriptor]}

      elif obj["fields"][fkey]["optype"] == "categorical":

        descriptor = "categories"
        converter = lambda x: str(x)
        self.field_values[name] = [converter(cat[0]) for cat in obj["fields"][fkey]["summary"][descriptor]]
        tot = sum([float(cat[1]) for cat in obj["fields"][fkey]["summary"][descriptor]])
        self.marginals[name] = {cat[0]: float(cat[1])/tot for cat in obj["fields"][fkey]["summary"][descriptor]}

      elif obj["fields"][fkey]["optype"] == "text":

        descriptor = "tag_cloud"
        converter = lambda x: str(x).lower()
        tags = [cat[0] for cat in obj["fields"][fkey]["summary"][descriptor]]
        ps = {cat[0]: float(cat[1])/self.training_size for cat in obj["fields"][fkey]["summary"][descriptor]}

        self.field_values[name] = []
        self.marginals[name] = {}
        for i in range(1, len(tags)+1):
          for c in itertools.combinations(tags, i):
            val = " ".join(list(c))
            self.field_values[name].append(val)
            self.marginals[name][val] = math.exp(sum([math.log(ps[v]) for v in list(c)]))    

def process_value(val, field_name, info):

  if info.types[field_name] == "numeric":

    if isinstance(val, basestring):
      val = val.strip().replace(",", "")

    numval = float(val)
    val_diffs = [abs(numval - float(fieldval)) for fieldval in info.field_values[field_name]]
    newval = info.field_values[field_name][val_diffs.index(min(val_diffs))]

    return newval

  elif info.types[field_name] == "categorical":

    for fieldval in info.field_values[field_name]:
      newval = fieldval if fieldval.lower() == val.lower() else val
      if newval == fieldval:
        return newval

  elif info.types[field_name] == "text":

    newval = val.lower()
    return newval

  else:

    assert "Error: unrecognized field operation on %s (\"%s\")" % (field_name, info.types[field_name]) and False

def read_csv(filename, info, config):

  data = []
  classname = info.objective_name
  attributes = info.field_names
  classindex = sum([i for i in range(len(attributes)) if attributes[i] == classname])

  with open(filename, 'rU') as f:

    reader = csv.reader(f, delimiter = ",", quotechar = "\"")

    attrline = reader.next()
    attrline = [el.strip() for el in attrline]

    fixed_attrline = []
    for csv_attr in attrline:
      newattr = csv_attr
      for attr in attributes:
        if csv_attr.lower().startswith(attr.lower()):
          newattr = attr
          break
      fixed_attrline.append(newattr)
    attrline = fixed_attrline
    new_attributes = {attrline[i]: i for i in range(len(attrline))}

    for row in reader:
      drow = {}
      missing_class = False
      has_missing = False
      for i in range(len(attributes)):
        missing = len(str(row[new_attributes[str(attributes[i])]]).strip()) == 0
        missing_class = missing_class or (missing and (attributes[i] == classname or attributes[i] == config["target"]))
        has_missing = has_missing or missing
        if not missing:
          processed_val = process_value(row[new_attributes[str(attributes[i])]], attributes[i], info)
          if processed_val is not None:
            drow[attributes[i]] = processed_val

      if not missing_class and not has_missing:
        data.append(drow)

  print "read %d rows" % len(data)
  return data

def get_bigml_tree(config, split, shuffle=False):

  with open(config["tree"]) as json_data:
    obj = json.load(json_data)
    info = BigMLModelInfo(obj)

  if not config.has_key("seed"):
    config["seed"] = 1
  rseed = config["seed"]

  data = read_csv(config["datafile"], info, config)

  if shuffle:
    random.shuffle(data)
  if split < 1.0:
    training = data[:int(round(len(data)*split))]
    test = data[int(round(len(data)*split)):]
  else:
    training = data[:]
    test = data[:]

  return obj, lambda x: info.model.predict(x), info, training, test

def main():

  config = {"tree": sys.argv[1], "datafile": sys.argv[2], "split": 1.0, "target": "Porn Watched [#Sex] (Have you seen an x-rated movie in the last year?)"}
  obj, model, info, training, test = get_bigml_tree(config, 1.0)
  pdb.set_trace()

if __name__ == '__main__':

  main()
