import sys
import csv
import math
import json
import pdb
import random
import itertools

import numpy as np
import numpy.random

import runtreeinversion
import treeinversion
import bigmlutil
import orangeutil

def read_csv(config):

  filename = config["datafile"]

  data = []
  with open(filename, 'rU') as f:

    reader = csv.reader(f, delimiter = ",", quotechar = "\"")

    attrline = reader.next()
    attributes = [el.strip() for el in attrline]
    new_attributes = {attributes[i]: i for i in range(len(attributes))}
    vals = {feat: set() for feat in attributes}
    margs = {feat: {} for feat in attributes}

    for row in reader:
      drow = {}
      for i in range(len(attributes)):
        missing = len(str(row[new_attributes[str(attributes[i])]]).strip()) == 0
        if not missing:
          processed_val = row[new_attributes[str(attributes[i])]]
          drow[attributes[i]] = processed_val
          vals[attributes[i]].add(processed_val)
          if margs[attributes[i]].has_key(processed_val):
            margs[attributes[i]][processed_val] += 1.0
          else:
            margs[attributes[i]][processed_val] = 0.0
        else:
          drow[attributes[i]] = None

      data.append(drow)

  N = float(len(data))
  for feat in attributes:
    for val in margs[feat].keys():
      margs[feat][val] /= N

  print "read %d rows" % len(data)
  return data, attributes, vals, margs

def bayesopt_predict(data, sample, attributes, vals, target, missing_feats=[], margs=None, use_margs=True):

  if sample[target] is None:
    return None

  missing = missing_feats[:]
  if target not in missing:
    missing.append(target)
  known = [str(feat) for feat in attributes if not feat in missing]
  if margs is None:
    margs = {feat: {val: 1.0 for val in vals[feat]} for feat in attributes}

  matching = {val: 0 for val in vals[target]}
  for row in data:
    if row[target] is None:
      continue
    matches = reduce(lambda matches, feat: matches and (sample[feat] is None or row[feat] is None or row[feat] == sample[feat]), known, True)
    if matches:
      matching[row[target]] += 1.0 if not use_margs else margs[target][row[target]]

  maxmatch = max(matching.values())
  nmax = sum([1 if matching[v] == maxmatch else 0 for v in matching.keys()])
  if nmax > 1 and margs is not None:
    for val in vals[target]:
      if matching[val] == maxmatch:
        matching[val] = float(matching[val])*margs[target][val]
      else:
        matching[val] = 0.0

  return max(matching, key=lambda v: matching[v])

def eval_predictability(config, data, attributes, vals, target, missing=[], margs=None, print_updates=None):

  mod_config = dict(config)
  mod_config["target"] = target

  guesses = []
  matches = 0
  for row, i in zip(data, range(len(data))):
    guess = bayesopt_predict(data, row, attributes, vals, target, missing, margs)
    if guess == row[target]:
      matches += 1
    if guess is not None and row[target] is not None:
      guesses.append(guess)
    if print_updates is not None and i % print_updates == 1 and len(guesses) > 0:
      if len(vals[target]) == 2:
        stats = runtreeinversion.statistics(guesses, data[:i+1], mod_config, count=0, time_ms=0, do_flip=False, do_print=False)
        print stats
      else:
        stats = runtreeinversion.simplestatistics(guesses, data[:i+1], target, count=0, time_ms=0, do_print=False)
        print stats

  if len(vals[target]) == 2:
    stats = runtreeinversion.statistics(guesses, data, mod_config, count=0, time_ms=0, do_flip=False, do_print=False)
  else:
    stats = runtreeinversion.simplestatistics(guesses, data, target, count=0, time_ms=0, do_print=False)

  return stats

def eval_data(config):

  data, attributes, vals, margs = read_csv(config)

  print "inversion target (missing=None):"
  print eval_predictability(config, data, attributes, vals, str(config["target", "fnlwgt"]), margs=margs, print_updates=1000)
  print

  print "class value (missing=None):"
  print eval_predictability(config, data, attributes, vals, str(config["class"]), margs=margs, print_updates=1000)
  print

  print "-"*60
  print

  for feat in attributes:
    missing = [feat]

    print "inversion target (missing=%s):" % (str(missing))
    print eval_predictability(config, data, attributes, vals, str(config["target"]), missing=missing, margs=margs, print_updates=1000)
    print

  print "-"*60
  print

  for feat in attributes:
    missing = [feat]

    print "class value (missing=%s):" % (str(missing))
    print eval_predictability(config, data, attributes, vals, str(config["class"]), missing=missing, margs=margs, print_updates=1000)
    print

def predicate_matches(predicate, row, info):

  fieldname = info.id_to_name[predicate["field"]]
  if not row.has_key(fieldname):
    return True

  val = predicate["value"]
  rowval = row[fieldname]

  if predicate["operator"] == "=":
    return val == rowval
  elif predicate["operator"] == "!=":
    return val != rowval
  elif predicate["operator"] == "<=":
    return float(rowval) <= float(val)
  elif predicate["operator"] == "<":
    return float(rowval) < float(val)
  elif predicate["operator"] == ">=":
    return float(rowval) >= float(val)
  elif predicate["operator"] == ">":
    return float(rowval) > float(val)

  assert False and "Unknown predicate operator: \"%s\"" % predicate["operator"]

def path_matches(path, row, info):

  predicates = [node["predicate"] for node in path[1:len(path)]]

  leaf = path[len(path)-1]
  leaf_pred = {"operator": "=", "value": leaf["output"], "field": info.name_to_id[info.objective_name]}
  predicates.append(leaf_pred)

  return reduce(lambda matches, predicate: matches and predicate_matches(predicate, row, info), predicates, True)

def eval_tree(config):

  assert config.has_key("tree") and "Given configuration does not mention a tree..."

  with open(config["tree"]) as json_data:
    treeobj = json.load(json_data)
    info = bigmlutil.BigMLModelInfo(treeobj)

  interest_feat = str(config["target"])
  interest_val = str(config["pos_label"])

  data = bigmlutil.read_csv(config["datafile"], info, config)
  reduced = [row for row in data if row[interest_feat] == interest_val]

  partial_config = {interest_feat: interest_val}

  paths = []
  allpaths = treeinversion.paths(info.root)
  for path in allpaths:
    has_predicate = False
    for node in path:
      assert node.has_key("predicate")

      predicate = node["predicate"]
      if isinstance(predicate, bool):
        continue

      fieldname = info.id_to_name[predicate["field"]]
      if fieldname != interest_feat:
        continue

      assert str(predicate["operator"]) == "=" or str(predicate["operator"]) == "!="

      op = str(predicate["operator"])
      if op == "=" and predicate["value"] != interest_val:
        continue
      if op == "!=" and predicate["value"] == interest_val:
        continue
      has_predicate = True
      break

    if has_predicate:
      paths.append(path[:])

  total_affected = 0
  correct_affected = 0
  row_matches = 0
  path_sets = {i: [] for i in range(len(reduced))}
  for path in paths:
    leaf = path[len(path)-1]
    total_affected += leaf["count"]
    correct_cnt = sum([int(el[1]) for el in leaf["objective_summary"]["categories"] if el[0] == leaf["output"]])
    correct_affected += correct_cnt
    for (row, i) in zip(reduced, range(len(reduced))):
      if path_matches(path, row, info):
        row_matches += 1
        path_sets[i].append(path)

  print "value affects %d of %d" % (total_affected, info.training_size)
  print "correctly predicts %d" % correct_affected
  print "row matches %d" % row_matches
  print "max matches %d" % max([len(path_sets[i]) for i in range(len(reduced))])

  pdb.set_trace()

def main():

  with open(sys.argv[1]) as config_fp:
    config = json.load(config_fp)

  eval_data(config)
  # eval_tree(config)

if __name__ == '__main__':

  main()
