import sys
import csv
import json
import math

from sklearn import metrics
from bigml.model import Model
from bigml.api import BigML

class ModelInfo(object):

	def __init__(self, model_object, objective_field=None):

		self.training_size = model_object["max_rows"] # yes, this is overwritten. not sure which to use yet
		self.model = Model(model_object)
		self.objective_field = model_object["objective_field"] if objective_field is None else objective_field

		obj = model_object["model"]
		self.field_names = [obj["fields"][fkey]["name"] for fkey in obj["fields"].keys()]
		self.field_values = {obj["fields"][fkey]["name"]: [cat[0] for cat in obj["fields"][fkey]["summary"]["categories"]] for fkey in obj["fields"].keys()}
		self.training_size = sum([cat[1] for cat in obj["distribution"]["training"]["categories"]])
		self.marginals = {obj["fields"][fkey]["name"]: {cat[0]: float(cat[1])/float(self.training_size) for cat in obj["fields"][fkey]["summary"]["categories"]} for fkey in obj["fields"].keys()}
		self.name_to_id = {obj["fields"][fkey]["name"]: fkey for fkey in obj["fields"].keys()}
		self.id_to_name = {fkey: obj["fields"][fkey]["name"] for fkey in obj["fields"].keys()}
		self.root = obj["root"]		
		self.objective_name = self.id_to_name[self.objective_field]
		self.objective_index = sum([i for i in range(len(self.field_names)) if self.field_names[i] == self.objective_name])

def remove_quoted_commas(line):

	in_quote = False
	new_line = ""
	
	for i in range(len(line)):
		if line[i] == '\"':
			in_quote = not in_quote
			continue
		if line[i] == ',' and in_quote:
			continue
		new_line = "%s%s" % (new_line, line[i])
	return new_line

def read_csv(filename, attributes, classindex):

	data = []

	with open(filename, 'rb') as f:
		dialect = csv.Sniffer().sniff(f.read(1024))
		f.seek(0)
		reader = csv.reader(f, dialect)
		attrline = reader.next()
		attrline = [el.strip() for el in attrline]
		new_attributes = {attrline[i]: i for i in range(len(attrline))}

		for row in reader:
			drow = {}
			for i in range(len(attributes)):
				drow[attributes[i]] = row[new_attributes[attributes[i]]]		

			data.append(drow)

	return data

def marginal_prob(row, fields, info):

	margs = info.marginals
	return math.exp(sum([math.log(margs[field][row[field]]) for field in fields]))

# def fillin(row, missing, doms):

# 	newrows = [dict(row)]
# 	for m in missing:		
# 		for el in doms[m]:
# 			additions = []			
# 			for r in newrows:
# 				if r[m] == el:
# 					continue
# 				newr = dict(r)
# 				r[m] = el
# 				additions.append(newr)
# 			newrows.extend(additions)
	
# 	for i in range(len(newrows)-1):
# 		for j in range(i+1, len(newrows)):
# 			if str(newrows[i]) == str(newrows[j]):
# 				newrows.pop(i)
# 				break

# 	return newrows

def fillin(row, missing, info):

	if len(missing) == 0:
		yield row
	else:
		for val in info.field_values[missing[0]]:
			row[missing[0]] = val
			for newrow in fillin(row, missing[1:], info):
				yield newrow

def row_marginal(row1, row2, info):

	fields = [] #[field in info.field_values if field != info.objective_name]
	row = dict(row1)
	for field in info.field_names:
		if field == info.objective_name:
			continue
		if row1[field] != row2[field]:
			fields.append(field)
			del row[field]

	if len(fields) > 2:
		return 0.0
	else:
		fills = fillin(row, fields, info)

		return sum([marginal_prob(filled, fields, info) for filled in fills])

def row_matches(row1, row2, info, threshold=1):

	for field in info.field_names:
		if field == info.objective_name:
			continue
		if row1[field] != row2[field]:
			threshold -= 1
		if threshold == 0:
			return False
	return True

def joint_prediction(row, train, info):

	matches = {val: 0.0 for val in info.field_values[info.objective_name]}

	for sample in train:
		if row_matches(row, sample, info): # and sample[info.objective_name] == "Yes":
			matches[sample[info.objective_name]] += 1
		# matches[sample[info.objective_name]] += row_marginal(row, sample, info)
		# print matches

	if max(matches.values()) == 0.0:
		matches["No"] = 1
	
	return (max(matches.keys(), key=lambda v: matches[v]), max(matches.values()))

def num_matches(row, train, info, threshold=1):

	n = 0
	for sample in train:
		n += 1 if row_matches(row, sample, info, threshold) else 0

	return n

with open(sys.argv[1]) as json_data:
	obj = json.load(json_data)
	info = ModelInfo(obj)

data = read_csv(sys.argv[2], info.field_names, info.objective_index)
train = read_csv(sys.argv[3], info.field_names, info.objective_index)

real_objective = [row[info.objective_name] for row in data]
predicted_objective = []
cheaters_caught = 0
cheaters_missed = 0
false_accused = 0
mispredicted = 0
n_matches = []
for row in data:

	n_matches.append(num_matches(row, train, info))

	predicted = joint_prediction(row, train, info)[0]
	predicted_objective.append(predicted)
	if predicted != row[info.objective_name]:
		mispredicted += 1
		if row[info.objective_name] == "Yes":
			cheaters_missed += 1
		else:
			false_accused += 1
	else:
		if row[info.objective_name] == "Yes":
			cheaters_caught += 1		
	print "joint misclassified: %d of %d (%.2f, %d, %d, %d)" % (mispredicted, len(predicted_objective), float(mispredicted)/float(len(predicted_objective)), cheaters_caught, cheaters_missed, false_accused)
print "joint misclassified: %d of %d (%.2f, %d, %d, %d)" % (mispredicted, len(predicted_objective), float(mispredicted)/float(len(predicted_objective)), cheaters_caught, cheaters_missed, false_accused)
print "mean matches:", float(sum(n_matches))/float(len(n_matches))
quit()

real_objective = [row[info.objective_name] for row in data]
cheaters_caught = 0
cheaters_missed = 0
false_accused = 0
predicted_objective = []
mispredicted = 0
for row in data:
	predicted = info.model.predict(row)	
	predicted_objective.append(predicted)
	if predicted != row[info.objective_name]:
		mispredicted += 1
		if row[info.objective_name] == "Yes":
			cheaters_missed += 1
		else:
			false_accused += 1
	else:
		if row[info.objective_name] == "Yes":
			cheaters_caught += 1		

print "joint misclassified: %d of %d (%.2f, %d, %d, %d)" % (mispredicted, len(predicted_objective), float(mispredicted)/float(len(predicted_objective)), cheaters_caught, cheaters_missed, false_accused)
