import os
import sys
import random
import itertools
import pydot
import pdb
import argparse

import sklearn.datasets as skdata
import numpy as np

from sklearn import tree
from sklearn.externals.six import StringIO 

from dill.source import getsource

# global counter to track the number of queries made to the model
npredict = 0

def round_to(n, prec):
    return round(round(n/prec)*prec, len(str(prec).split('.')[1]))

'''
    Make a tree from the iris dataset, and render the tree to iris.pdf.
    By default, this only uses two dimensions of the data, to make examples 
    easier to draw on paper.

    arguments:
        r1, r2: row-slice [r1:r2] of the dataset to use for training
        cols: subset of the columns to use for training

    returns: the tree, training data
'''
def make_iris(r1=None, r2=None, cols=[2,3]):

    iris = skdata.load_iris()
    data = iris.data
    data_cols = zip(*data)
    train_cols = [[el for el in data_cols[col]] for col in cols]
    data = np.asarray(zip(*train_cols))
    # data = np.asarray(zip([el for el in data_cols[2]], [el for el in data_cols[3]]))  

    if r1 is None:
        r1=random.randint(0, len(data)-30)
    if r2 is None:
        r2=random.randint(r1+10, len(data))

    traindata = data[r1:r2]

    itree = tree.DecisionTreeClassifier(random_state=0)
    itree = itree.fit(traindata, iris.target[r1:r2])    

    dot_data = StringIO() 
    tree.export_graphviz(itree, out_file=dot_data) 
    graph = pydot.graph_from_dot_data(dot_data.getvalue()) 
    graph.write_pdf("iris.pdf") 

    print "\n... wrote model to iris.pdf"

    return itree, data, iris.target

'''
    Queries a model and updates the counter.

    arguments:
        c: model
        x: query point

    returns: model response
'''
def predict(c, x):

    global npredict

    npredict += 1
    return c.predict(x)[0]

'''
    Returns the midpoint of a rectangular subspace.

    arguments:
        reg: an array of tuples corresponding to the mins, max'es 
             of the region along each dimension.
        epsilon: numeric granularity
        idx: if specified, the midpoint is only taken along this dimension.

    returns: the midpoint
'''
def midpoint(reg, epsilon, idx=None):

    mins, maxs = zip(*reg)

    assert len(mins) == len(maxs)

    if idx is None:
        return [round_to(float(maxs[i] + mins[i])/2.0, epsilon) for i in range(len(mins))]
    else:
        return [round_to(float(maxs[i] + mins[i])/2.0, epsilon) if i == idx else mins[i] for i in range(len(mins))] 

'''
    NOTE: This function is deprecated. I'm commenting on it just to give insight into
          the trajectory of our thinking on this heuristic.

    Finds the next model split in feature space from a given point, along a given dimension.
    Basically, this works by searching the target dimension for a transition point,
    where moving in one direction will result in a change in model response. This marks the
    boundary of a region in feature space used for classification.

    arguments:
        clsfr: the model
        pt: the starting point
        idx: target dimension
        orig_label: model response at starting point
        mins: lower-bounds for search along each dimension
        maxs: upper-bounds for search along each dimension
        epsilon: numeric granularity

    returns: a point at which the classifier response transitions when incremented
             along the target dimension
'''
def findsplit(clsfr, pt, idx, orig_label, mins, maxs, epsilon):     

    worklist = [(mins[idx], maxs[idx])]
    done = []
    curpt = pt[:]
    newpt = pt[:]

    cnt = 0

    while len(worklist) > 0:
        cnt += 1        
        (curmin, curmax) = worklist.pop()
        done.append((curmin, curmax))       
        
        curpt[idx] = round_to((curmax + curmin)/2, epsilon)     
        if predict(clsfr, curpt) != orig_label:         
            if (curmin, curpt[idx]) not in done:
                worklist.append((curmin, curpt[idx]))
            continue

        newpt = curpt[:]
        newpt[idx] += epsilon       
        if predict(clsfr, newpt) != orig_label:
            return curpt

        if curmin < curpt[idx] and (curmin, curpt[idx]) not in done:            
            worklist.append((curmin, curpt[idx]))
        if curmax > curpt[idx] and (curpt[idx], curmax) not in done:            
            worklist.append((curpt[idx], curmax))

    return None

'''
    Infer a decision tree that approximates the given model (c).

    NOTE: This is deprecated, it was a first stab at the problem.
          In the interest of time, I won't comment on it further.
'''
def makenode(c, mins, maxs, dim, epsilon, indent="", maxdepth=None):

    assert len(mins) == len(maxs) == dim

    # idx = 0
    pt = mins[:]

    orig_label = predict(c, pt)

    if maxdepth is not None:
        maxdepth -= 1
        if maxdepth == 0:
            return (" {{ %s }} " % str(orig_label), lambda _: orig_label)

    indices = range(dim)
    # random.shuffle(indices)
    for idx in indices:
        
        split = findsplit(c, pt, idx, orig_label, mins, maxs, epsilon)      
        if split is None:           
            continue

        left_pt = [maxs[i] if i != idx else round_to(split[i]-epsilon, epsilon) for i in range(dim)]
        right_pt = [mins[i] if i != idx else round_to(split[i]+epsilon, epsilon) for i in range(dim)]
        (left_child_str, left_child_fn) = makenode(c, mins, left_pt, dim, epsilon, "  %s" % indent, maxdepth)
        (right_child_str, right_child_fn) = makenode(c, right_pt, maxs, dim, epsilon, "  %s" % indent, maxdepth)

        strep = "\n%sx[%d] <= %f --> %s,\n%sx[%d] > %f --> %s " % (indent, idx, split[idx], left_child_str, indent, idx, split[idx], right_child_str)
        fn = lambda x: left_child_fn(x) if x[idx] <= split[idx] else right_child_fn(x)

        return (strep, fn)

    return (" {{ %s }} " % str(orig_label), lambda _: orig_label)

'''
    Given a region structure, return a membership indicator function.

    arguments:
        region: an array of tuples corresponding to the mins, max'es 
                of the region along each dimension.

    returns: a function that takes a point (as an array), returns True to indicate membership in region
'''
def region_function(region):

    mins, maxs = zip(*region)

    assert len(mins) == len(maxs)

    dim = len(mins)
    return lambda x, mins=mins, maxs=maxs, dim=dim: sum([1 if x[idx] >= mins[idx] and x[idx] <= maxs[idx] else 0 for idx in range(dim)]) == dim

'''
    Given a set of region structures (a cover), determine membership
    of pt.

    arguments:
        pt: the point in question
        covered_reg: the cover

    returns: membership indicator
'''
def is_covered(pt, covered_reg):

    for reg in covered_reg:
        reg_fn = region_function(reg)
        if reg_fn(pt):
            return True

    return False

'''
    "Updates" a region based on new information learned
    during inference. The heuristic works "from left to right",
    so if we start from full feature space and proceed to carve
    out a subspace used for classification, then we want to
    update the search space to exclude this subspace.

    arguments:
        old: previous feature space (as a region structure)
        new: feature space whose "left" portion properly excludes inferred partitions
        epsilon: numeric granularity

    returns: new regions structure updated as described.
'''
def update_region(old, new, epsilon):

    assert len(old) == len(new)

    reg = []

    dim = len(old)
    for i in range(dim):
        old_min, old_max = old[i]
        new_min, new_max = new[i]

        up_min, up_max = old[i]
        if new_max < old_max:
            up_min = new_max+epsilon

        reg.append((up_min, up_max))

    return reg

# rg1 <= rg2
def is_region_subset(rg1, rg2):

    assert len(rg1) == len(rg2)

    for i in range(len(rg1)):
        if rg1[i][0] < rg2[i][0] or rg1[i][1] > rg2[i][1]:
            return False

    return True

'''
    Identifies whether a given region is contained in a cover.
    This function isn't technically correct, as the target
    region could straddle adjacent regions in the cover.
    TODO: fix it, if we decide that this heuristic is worthwhile.

    arguments:
        rg: target region
        cover: covered area

    returns: Boolean indicator
'''
def is_region_covered(rg, cover):

    for cover_rg in cover:

        if is_region_subset(rg, cover_rg):
            return True

    return False

'''
    Identifies whether a given point lies on the edge of
    feature space.
    The heuristic only works on bounded feature spaces

    arguments:
        pt: the point in question
        dom: the full domain of the feature space
        idx: if specified, the predicate only operates along the given dimension

    returns: Boolean indicator
'''
def is_edge_pt(pt, dom, idx=None):

    assert len(pt) == len(dom)

    for i in range(len(pt)):
        if idx is not None and i != idx:
            continue
        if pt[i] >= dom[i][1]:
            return True

    return False

'''
    Finds the next partition (or region) of the feature space used by the
    model for classification. "Next" is relative to a given point. The
    heuristic works from "left to right" along each dimension independently,
    and walks monotonically in fixed increments (of epsilon) until it
    witnesses a change in the model's response. When it does, it pivots to a
    new dimension and performs a similar search. If it ever hits the edge of
    feature space, it changes sign and backs up. The region constructed in
    this way gradually expands until it covers a rectangular subspace on which
    the model's response is constant (** subject to caveats).

    ** Some types of non-convex regions are still correctly identified by this
       heuristic as a set of adjacent rectangular regions, inferred over
       multiple calls to this function. This algorithm will NOT correctly
       identify regions wholly contained in others. A stochastic modification
       to this algorithm could probably do OK with such models, but in general,
       exact inference for these requires an exhaustive search.    

    arguments:
        c: the model
        pt: the starting point
        mins: lower-bounds of feature space along each dimension
        maxs: upper-bounds of feature space ''
        dim: the number of feature-space dimensions
        epsilon: numeric granularity

    returns: a region structure over which the model response is invariant,
             subject to the caveats described above.

'''
def get_region(c, pt, mins, maxs, dim, epsilon):

    orig_label = predict(c, pt)

    region = zip(pt, pt)    

    indices = range(dim)
    for idx in indices:

        sign = 1
        if is_edge_pt(pt, zip(mins, maxs), idx):
            sign = -1

        v = pt[idx]     
        while mins[idx] <= v and v <= maxs[idx]:
            new_pt = pt[:]
            new_pt[idx] = round_to(v, epsilon)
            new_label = predict(c, new_pt)          
            if new_label == orig_label:
                region[idx] = (region[idx][0], max(region[idx][1], new_pt[idx]))
            else:
                break
            v += sign*epsilon
            v = round_to(v, epsilon)            

    return region

'''
    The following functions (makefn_selector, makefn_plus, makefn_minus,
    makefn_identity, makefn_composition, make_corner_basis_fns, corner2fn,
    eval_cornerfn, make_corner_fns) set up some scaffolding that allows us to
    pivot in all the necessary directions from each feature space region we
    identify, regardless of the number of dimensions.

    It uses some awesomeness from itertools to do this, but it's a little
    tricky. In the interest of time, I won't explain how these functions work
    in detail, but I will explain how the result is used below. If necessary
    in the future, I can explain how it works to whomever's interested.
'''
def makefn_selector(index, sel):
    return lambda x, index=index, sel=sel: [x[i][sel] if i == index else x[i] for i in range(len(x))]

def makefn_plus(index, epsilon):
    return lambda x, index=index, epsilon=epsilon: [round_to(x[i]+epsilon, epsilon) if i == index else x[i] for i in range(len(x))]

def makefn_minus(index, epsilon):
    return lambda x, index=index, epsilon=epsilon: [round_to(x[i]-epsilon, epsilon) if i == index else x[i] for i in range(len(x))]

def makefn_identity():
    return lambda x: x

def makefn_composition(fnset):

    if len(fnset) == 0:
        return lambda x: x

    fn = fnset.pop()
    return lambda x, fnset=fnset: fn(makefn_composition(fnset)(x))

def make_corner_basis_fns(dim, epsilon):

    fnset_sel = [[makefn_selector(index, sel) for sel in range(2)] for index in range(dim)]
    fnset_plus = [makefn_plus(index, epsilon) for index in range(dim)]
    fnset_minus = [makefn_minus(index, epsilon) for index in range(dim)]
    fnset_id = [makefn_identity() for index in range(dim)]

    fnset = []
    for i in range(dim):
        fnset.append([])
        for j in range(2):
            fnset[i].append(lambda x, i=i, j=j: fnset_minus[i](fnset_sel[i][j](x)))         
            fnset[i].append(lambda x, i=i, j=j: fnset_id[i](fnset_sel[i][j](x)))
            fnset[i].append(lambda x, i=i, j=j: fnset_plus[i](fnset_sel[i][j](x)))

    return fnset
    

def corner2fn(corner, functions):

    # 0 -> -minus
    # 1 -> -identity
    # 2 -> -plus
    # 3 -> +minus
    # 4 -> +identity
    # 5 -> +plus

    dim = len(corner)
    fns = [functions[i][corner[i]] for i in range(dim)]
    
    return makefn_composition(fns)

def eval_cornerfn(corner, functions, x):

    dim = len(corner)

    for i in range(dim):
        x = functions[i][corner[i]](x)

    return x

def make_corner_fns(dim, epsilon):

    corner_basis_fns = make_corner_basis_fns(dim, epsilon)
    corner_fns = []
    corners = []

    for corner in itertools.product([1,4,5], repeat=dim):
        corner_fns.append(corner2fn(corner, corner_basis_fns))
        corners.append(corner)

    return (corners, corner_fns)

'''
    Predict the model's response on a given point from the inferred region
    cover.

    arguments:
        region_fns: the region cover as a set of tuples (region membership function, response label, region structure)
        pt: the point in question

    returns: predicted model response
'''
def predict_from_cover(region_fns, pt):

    for fn, l, rg in region_fns:
        if fn(pt):
            return l

    return None

'''
    Infer an approximation of the model given by itree. The primary heuristic
    is described in the doc for get_region, and its use in the comments of this function.

    arguments:
        itree: the model
        mins, maxs: upper and lower-bounds for the feature space along each dimension
        epsilon: numeric granularity; used when searching and pivoting between regions

    returns: a black-box approximation of the given classifier, as a list of
        triples; for each discovered feature-space region a membership
        indicator function, the model's response label, and the corresponding
        region structure.
'''
def infer_model_from_blackbox(itree, mins, maxs, epsilon):

    assert len(mins) == len(maxs)
    dim = len(mins)

    # set up the feature space
    domain = zip(mins, maxs)    

    # use the itertools magic to build the scaffolding for
    # region pivots
    corner_basis_fns = make_corner_basis_fns(dim, epsilon)
    domain_fn = region_function(domain)
    corners, corner_fns = make_corner_fns(dim, epsilon)

    # initial solution
    pts = [(mins[:], domain)]
    covered = []

    # we apply the heuristic using a worklist algorithm
    while len(pts) > 0:

        # grab the next point whose model region we will identify
        pt, dom = pts[0]        
        pts = pts[1:] if len(pts) > 0 else []

        # if we've managed to cover this point in our inferred model, move on
        if is_covered(pt, covered):
            continue
        
        dmins, dmaxs = zip(*dom)
        
        # get the response-invariant region in feature space containing this point     
        rg = get_region(itree, pt, dmins, dmaxs, dim, epsilon)
        if rg is None:
            continue
        if not is_region_covered(rg, covered):
            covered.append(rg)
        else:
            continue

        # we've found a new region. now we update our working set to make sure
        # that we get all of the adjacent regions.
        # this is where the "scaffolding" from above comes into play.
        # we identify all of the corners of this region, and take the points
        # that lie just past the corners if they're not already covered by
        # a previously-inferred region.
        # these are added to the working set.
        for corner, fn in zip(corners, corner_fns):
            new_pt = eval_cornerfn(corner, corner_basis_fns, rg)
            if not is_covered(new_pt, covered) and domain_fn(new_pt): # and not is_edge_pt(new_pt, domain):
                pts.append((new_pt, update_region(dom, rg, epsilon)))                   

    # for debugging, print the inferred regions and their model responses
    print "\ninferred model:"
    for reg in covered:
        print "\t", reg, "-->", predict(itree, midpoint(reg, epsilon))

    # construct the inferred classifier.
    # it's a list of triples, containing for each discovered region a membership indicator function,
    # the response label, and the corresponding region structure.
    covered_fns = [(region_function(cov), predict(itree, midpoint(cov, epsilon)), cov) for cov in covered]

    return covered_fns

def process_args():
    parser = argparse.ArgumentParser(description=
        """
        Black-box model extraction for classifiers with real-valued features and nominal responses.
        Loads Fisher's Iris dataset, and selects a random subset of rows on which to train a decision tree.        
        """)

    parser.add_argument('--ncols', metavar='ncols', type=int, default=2, help='number of columns to use for training (default=2)')
    parser.add_argument('--epsilon', metavar='epsilon', type=float, default=0.05, help='numeric granularity (default=0.05)')
    parser.add_argument('--seed', metavar='seed', type=int, default=1, help='random seed (default=1)')
    
    return parser.parse_args()

def main():

    args = process_args()

    random.seed(args.seed)
    
    itree, data, target = make_iris(cols=range(4)[(4-args.ncols):4])
    dim = len(data[0])  

    epsilon = args.epsilon

    mins = [round_to(min(data[:,i]), epsilon) - epsilon for i in range(dim)]
    maxs = [round_to(max(data[:,i]), epsilon) for i in range(dim)]  

    covered_fns = infer_model_from_blackbox(itree, mins, maxs, epsilon)

    # now evaluate the inferred model on the original data. 
    # here, we say that we have an error whenever the inferred model diverges from the original.    
    print
    print '='*60
    print "\n... %d model queries" % npredict
    wrong = 0
    actwrong = 0
    for (row, response) in zip(data, target):
        row = [round_to(el, epsilon) for el in row]
        if predict(itree, row) != predict_from_cover(covered_fns, row):
            wrong += 1
    print "... error rate=%f [%d mispredictions]\n" % (float(wrong)/float(len(data)), wrong)    

if __name__ == '__main__':

  main()
