import sys
import csv
import math
import json

import numpy as np
import numpy.random

from sklearn import metrics
from bigml.model import Model
from bigml.api import BigML

class ModelInfo(object):

	def __init__(self, model_object):

		self.training_size = model_object["max_rows"] # yes, this is overwritten. not sure which to use yet
		self.model = Model(model_object)
		self.objective_field = model_object["objective_field"]

		obj = model_object["model"]
		self.field_names = [obj["fields"][fkey]["name"] for fkey in obj["fields"].keys()]
		self.field_values = {obj["fields"][fkey]["name"]: [cat[0] for cat in obj["fields"][fkey]["summary"]["categories"]] for fkey in obj["fields"].keys()}
		self.training_size = sum([cat[1] for cat in obj["distribution"]["training"]["categories"]])
		self.marginals = {obj["fields"][fkey]["name"]: {cat[0]: float(cat[1])/float(self.training_size) for cat in obj["fields"][fkey]["summary"]["categories"]} for fkey in obj["fields"].keys()}
		self.name_to_id = {obj["fields"][fkey]["name"]: fkey for fkey in obj["fields"].keys()}
		self.id_to_name = {fkey: obj["fields"][fkey]["name"] for fkey in obj["fields"].keys()}
		self.root = obj["root"]		
		self.objective_name = self.id_to_name[self.objective_field]
		self.objective_index = sum([i for i in range(len(self.field_names)) if self.field_names[i] == self.objective_name])

with open(sys.argv[1]) as json_data:
	obj = json.load(json_data)
	info = ModelInfo(obj)

def firstpath(node, path=[]):

	if not node.has_key("children"):
		return path+[node]
	else:
		return firstpath(node["children"][0], path+[node])

def paths(node, path=[]):

	if not node.has_key("children"):
		yield path+[node]
	else:
		for child in node["children"]:
			for newpath in paths(child, path+[node]):
				yield newpath

def objective_configs(path, info):

	leaf = path[len(path)-1]
	assert leaf.has_key("objective_summary")

	for cat in leaf["objective_summary"]["categories"]:
		for i in range(int(cat[1])):
			yield {info.objective_name: cat[0]}

def path_str(path, info):

	r = ""

	for node in path:
		if not isinstance(node["predicate"], dict):
			r = "%s(root)" % r
		else:
			count = node["count"]
			op = node["predicate"]["operator"]
			pfield = info.id_to_name[node["predicate"]["field"]]
			pval = node["predicate"]["value"]
			r = "%s\n\t-> [%s %s %s] n=%s" % (r, pfield, op, pval, str(count))
			if not node.has_key("children"):
				for cat in node["objective_summary"]["categories"]:
					r = "%s\n\t\t-> [%s = %s] n=%s" % (r, info.objective_name, cat[0], str(cat[1]))

	return r

def row_str(row):

	r = ""
	for field in row.keys():
		r = "%s\t%s = %s\n" % (r, field, row[field])

	return r

avg_possible = {field: 0.0 for field in info.field_names}

def mode_from_samples(samples, field, info):

	counts = {val: info.marginals[field][val] for val in info.field_values[field]}

	for samp in samples:
		counts[samp[field]] += 1.0

	# for val in info.field_values[field]:
	# 	counts[val] /= info.marginals[field][val]
	# print counts
	return max(counts, key=lambda v: counts[v])

def samples_from_path_and_partial(path, partial_config, info):

	samples = []
	possible_values = {field: info.field_values[field][:] for field in info.field_names}
	for field in partial_config.keys():
		possible_values[field] = [partial_config[field]]

	for node in path:
		
		assert node.has_key("predicate")
		
		predicate = node["predicate"]
		if not isinstance(predicate, dict):
			continue

		fieldname = info.id_to_name[predicate["field"]]
		if predicate["operator"] == "=":
			if predicate["field"] in partial_config.keys():
				if predicate["value"] != partial_config[info.id_to_name[predicate["field"]]]:
					possible_values[fieldname] = []
			else:
				possible_values[fieldname] = [predicate["value"]]
		elif predicate["operator"] == "!=":
			if predicate["value"] in possible_values[fieldname]:			
				possible_values[fieldname].remove(predicate["value"])
		else:
			assert False and "Unrecognized predicate--this only works on nominal values."

	for field in info.field_names:
		avg_possible[field] += len(possible_values[field])

	# for field in info.field_names:
	# 	print "possible[%s]: %s" % (field, possible_values[field])
	# 	for val in possible_values[field]:
	# 		print "\t--%s" % val
	# 		print "\tmarg[%s]: %f" % (val, info.marginals[field][val])
	# print '-'*60
	path_mass = {field: sum([info.marginals[field][val] for val in possible_values[field]]) for field in info.field_names}
	min_mass = min([path_mass[field] for field in info.field_names])
	if min_mass == 0.0:		
		return []
	path_marginals = {field: [info.marginals[field][val]/path_mass[field] for val in possible_values[field]] for field in info.field_names}
	path_entropy = sum([-1.0*sum([p*math.log(p, 2) for p in path_marginals[field]]) for field in info.field_names if field != info.objective_name])	
	counting_uncertainty = max([len(possible_values[field]) for field in possible_values.keys() if field != info.objective_name])

	if path_entropy > 0.0:
		return []
	print possible_values["have_you_ever_cheated_on_your_significant_other_"]

	for samp in objective_configs(path, info):
		if samp[info.objective_name] != partial_config[info.objective_name]:
			continue
		for field in possible_values.keys():
			if field == info.objective_name:
				continue
			samp[field] = numpy.random.choice(possible_values[field], p=path_marginals[field])			
		samples.append(samp)		

	return samples


def samples_from_path(path, info):

	samples = []
	possible_values = {field: info.field_values[field][:] for field in info.field_names}

	for node in path:
		
		assert node.has_key("predicate")
		
		predicate = node["predicate"]
		if not isinstance(predicate, dict):
			continue

		fieldname = info.id_to_name[predicate["field"]]
		if predicate["operator"] == "=":
			possible_values[fieldname] = [predicate["value"]]
		elif predicate["operator"] == "!=":
			possible_values[fieldname].remove(predicate["value"])
		else:
			assert False and "Unrecognized predicate--this only works on nominal values."

	for field in info.field_names:
		avg_possible[field] += len(possible_values[field])

	path_mass = {field: sum([info.marginals[field][val] for val in possible_values[field]]) for field in info.field_names}
	path_marginals = {field: [info.marginals[field][val]/path_mass[field] for val in possible_values[field]] for field in info.field_names}
	path_entropy = sum([-1.0*sum([p*math.log(p, 2) for p in path_marginals[field]]) for field in info.field_names if field != info.objective_name])	
	counting_uncertainty = max([len(possible_values[field]) for field in possible_values.keys() if field != info.objective_name])
	# if path_entropy > 5.0:
	# 	return samples

	for samp in objective_configs(path, info):
		for field in possible_values.keys():
			if field == info.objective_name:
				continue
			samp[field] = numpy.random.choice(possible_values[field], p=path_marginals[field])
			# samp[field] = possible_values[field][max(range(len(path_marginals[field])), key=lambda v: path_marginals[field][v])] 
		samples.append(samp)		

	return samples

def read_csv(filename, attributes, classindex):

	data = []

	with open(filename, 'rb') as f:
		dialect = csv.Sniffer().sniff(f.read(1024))
		f.seek(0)
		reader = csv.reader(f, dialect)
		attrline = reader.next()
		attrline = [el.strip() for el in attrline]
		new_attributes = {attrline[i]: i for i in range(len(attrline))}

		for row in reader:
			drow = {}
			for i in range(len(attributes)):
				drow[attributes[i]] = row[new_attributes[attributes[i]]]		

			data.append(drow)

	return data

def write_csv(filename, data, attributes):

	with open(filename, 'wb') as f:
		first = True
		for attr in attributes:
			quotes = "\"" if (',' in attr) else ""
			f.write("%s%s%s%s" % ("," if not first else "", quotes, attr, quotes))
			first = False
		f.write("\n")
		for row in data:
			first = True
			for attr in attributes:
				quotes = "\"" if (',' in row[attr]) else ""
				f.write("%s%s%s%s" % ("," if not first else "", quotes, row[attr], quotes))
				first = False
			f.write("\n")

cnt = 0
allpaths = paths(info.root)
samples = []

for path in allpaths:
	# print path_str(path, info)
	# print
	# for i in range(1):
	samples.extend(samples_from_path(path, info))
	# print '-'*60
	cnt += 1
	# if cnt > 10:
	# 	break

misclass = 0
caught = 0
missed = 0
data = read_csv(sys.argv[2], info.field_names, info.objective_index)
known_fields = [field for field in info.field_names if  field != "have_you_ever_cheated_on_your_significant_other_"]
for row in data:
	samples = []
	partial = {field: row[field] for field in known_fields}
	# print partial
	# print '-'*60
	allpaths = paths(info.root)
	for path in allpaths:
		samples.extend(samples_from_path_and_partial(path, partial, info))
	guess = mode_from_samples(samples, "have_you_ever_cheated_on_your_significant_other_", info)
	if guess != row["have_you_ever_cheated_on_your_significant_other_"]:
		misclass += 1
		if guess == "No":
			missed += 1
	else:
		if guess == "Yes":
			caught += 1
	print len(samples),
	print "misclass: %d (%.4f), caught=%d, missed=%d" % (misclass, float(misclass)/float(len(data)), caught, missed)

# for field in info.field_names:
# 	avg_possible[field] /= float(cnt)
# 	print "%s: %.2f" % (field, avg_possible[field])

# write_csv('/tmp/regen.csv', samples, info.field_names)