import sys
import os
import argparse
import gzip
import math
import random
import subprocess as sp
import time
import json
import pdb
import shutil

import pylearn2
import theano
import theano.tensor.signal
import numpy as np
# import numdifftools as nd
# import algopy
import cv2
import cPickle

import scipy
from scipy.optimize import fmin_l_bfgs_b
from scipy.optimize import fmin_bfgs

from collections import namedtuple

from pylearn2.config import yaml_parse
from pylearn2.utils import serial
from pylearn2.utils import safe_zip
from pylearn2.utils.data_specs import DataSpecsMapping

from theano.tensor.signal.conv import conv2d

from multiencoder import EncoderCascade

from PIL import Image

def gauss2D_filter(degree, image_shape):

    shape = (degree, degree)
    sigma = math.pow(degree, 2.0)

    m,n = [(ss-1.)/2. for ss in shape]
    y,x = np.ogrid[-m:m+1,-n:n+1]
    h = np.exp( -(x*x + y*y) / (2.*sigma*sigma) )
    h[ h < np.finfo(h.dtype).eps*h.max() ] = 0
    sumh = h.sum()
    if sumh != 0:
        h /= sumh

    filter_mat = h.astype(theano.config.floatX)

    image_bits = theano.tensor.matrix('image_bits', dtype=theano.config.floatX)

    blur_expr = conv2d(image_bits.reshape(image_shape), filter_mat, filter_shape=filter_mat.shape, subsample=(1,1))
    blur_fn = theano.function([image_bits], blur_expr)

    return image_bits, blur_expr, blur_fn

def pixelate_img(orig_img, npix=10):

    height, width = orig_img.shape
    img = np.copy(orig_img)

    for x in range(0, height, npix):
        for y in range(0, width, npix):
            ux = min(x+npix, height)
            uy = min(y+npix, width)
            img[x:ux, y:uy] = np.mean(orig_img[x:ux, y:uy])

    return img

def normalize(X, low, high, dtype=theano.config.floatX):
    """Normalizes a given array in X to a value between low and high."""
    X = np.asarray(X)
    # minX, maxX = max(np.min(X), 1), max(np.max(X), 1)
    minX, maxX = np.min(X), np.max(X)
    if minX == maxX:
        minX = low
        maxX = high
    # print minX, ",", maxX
    # normalize to [0...1].
    X = X - float(minX)
    X = X / float((maxX - minX))
    # scale to [low...high].
    X = X * (high-low)
    X = X + low
    if dtype is None:
        return np.asarray(X)
    return np.asarray(X, dtype=dtype)

def sharpen(image, iterations=1):
    sharpened = np.copy(image)
    for i in range(iterations):
        blurred = cv2.GaussianBlur(sharpened, (3, 3), 0)
        sharpened = cv2.addWeighted(sharpened, 1.5, blurred, -0.5, 0)

    return sharpened

def onehot_label(model, label):

    if model is None or label is None:
        return None

    cost = model.get_default_cost()
    data_specs = cost.get_data_specs(model)
    mapping = DataSpecsMapping(data_specs)

    space_tuple = mapping.flatten(data_specs[0], return_tuple=True)
    source_tuple = mapping.flatten(data_specs[1], return_tuple=True)

    assert source_tuple.index("targets") >= 0
    target_space = space_tuple[source_tuple.index("targets")]

    assert isinstance(target_space, pylearn2.space.VectorSpace)
    dim = target_space.dim

    onehot = np.asarray([0]*dim, dtype=theano.config.floatX).reshape(1, dim)
    onehot[0,label] = 1.0

    return onehot

def get_model_feature_shape(model):

    cost = model.get_default_cost()
    data_specs = cost.get_data_specs(model)
    mapping = DataSpecsMapping(data_specs)

    space_tuple = mapping.flatten(data_specs[0], return_tuple=True)
    source_tuple = mapping.flatten(data_specs[1], return_tuple=True)

    assert source_tuple.index("features") >= 0
    feature_space = space_tuple[source_tuple.index("features")]

    # if this is a vector space, we require image dimensions from the user
    if isinstance(feature_space, pylearn2.space.VectorSpace):
        shape = (1, feature_space.dim)
    else:
        # otherwise use the model's feature space shape
        shape = feature_space.shape
        if len(shape) < 3:
            height, width = shape
            shape = (1, height, width, 1)

    return shape

def process_image_for_display(img_data, image_shape, scale, do_sharpen=False, do_normalize=True):

    if do_normalize:
        norm_img = normalize(img_data, 0, 255, dtype=np.uint8).reshape(image_shape)
    else:
        norm_img = img_data.reshape(image_shape)
    scaled_shape = tuple(map(lambda x: scale*x, image_shape[:2]))
    scaled_img = cv2.resize(norm_img, scaled_shape[::-1], interpolation=cv2.INTER_NEAREST)
    if do_sharpen:
        scaled_img = sharpen(scaled_img, 3)

    return scaled_img

def display_update(img_data, img_orig, pix_image, image_shape, scale, window_name, do_sharpen=False, conf=None):

    # print "img_data: (%f, %f), img_orig: (%f, %f), pix_image: (%f, %f)" % (np.min(img_data), np.max(img_data), np.min(img_orig), np.max(img_orig), np.min(pix_image), np.max(pix_image))    

    cur_img = None
    p_img = process_image_for_display(img_data, image_shape, scale, do_sharpen)
    label = "?" if conf is None else "%.2f" % (conf)
    cv2.putText(p_img, "p=%s" % label, (0, 15), cv2.FONT_HERSHEY_COMPLEX, 0.5, (255,255,255), lineType=cv2.CV_AA)
    if pix_image is not None:
        pix_image = process_image_for_display(pix_image, image_shape, scale, False, True)
        cv2.putText(pix_image, "blurred", (0, 15), cv2.FONT_HERSHEY_COMPLEX, 0.5, (255,255,255), lineType=cv2.CV_AA)
        cur_img = pix_image
    if img_orig is not None:
        img_orig = process_image_for_display(img_orig, image_shape, scale, False, True)
        # cv2.putText(img_orig, "target", (0, 15), cv2.FONT_HERSHEY_COMPLEX, 0.5, (255,255,255), lineType=cv2.CV_AA)
        if cur_img is not None:
            cur_img = np.concatenate((cur_img, img_orig), axis=1)
        else:
            cur_img = img_orig
    if cur_img is not None:
        cur_img = normalize(cur_img, 0, 255, dtype=np.uint8)
        cur_img = np.concatenate((cur_img, p_img), axis=1)
    cv2.imshow(window_name, cur_img)
    cv2.waitKey(1)

def display_update_encoded(img_data, decoder, image_shape, scale, window_name):

    img_data = decoder(img_data)
    cv2.imshow(window_name, process_image_for_display(img_data, image_shape, scale))
    cv2.waitKey(1)

def open_video(filename, shape, display_scale):

    height, width, nchannels = shape

    pix_fmt = 'gray' if nchannels == 1 else 'bgr24'

    FFMPEG_BIN = "ffmpeg"

    command = [ FFMPEG_BIN,
            '-y',
            '-s',
            '%dx%d' % (display_scale*width, display_scale*height),
            '-r', '24',
            # '-vcodec', 'mpeg4',
            '-c:v', 'rawvideo',
            '-f', 'rawvideo',
            '-pix_fmt', pix_fmt,
            '-i',
            '-',
            filename ]
    with open('/dev/null', 'w') as devnull:
        pipe = sp.Popen(command, stdin=sp.PIPE, stdout=devnull, stderr=devnull)

    return pipe

def pixelate_img_misclass(orig_img, model_fun, feature_shape):

    orig_label = model_fun(orig_img.reshape(feature_shape))

    height, width = orig_img.shape        

    new_label = orig_label
    npix = 25

    while new_label == orig_label and npix < 50:
        img = np.copy(orig_img)
        blur_var, blur_expr, blur_fn = gauss2D_filter(npix, orig_img.shape)                
        img = cv2.resize(blur_fn(img), dsize=(width, height))
        new_label = model_fun(img.reshape(feature_shape))
        # npix += 1
        break

    # while new_label == orig_label:
    #     img = np.copy(orig_img)
    #     for x in range(0, height, npix):
    #         for y in range(0, width, npix):
    #             ux = min(x+npix, height)
    #             uy = min(y+npix, width)
    #             img[x:ux, y:uy] = np.mean(orig_img[x:ux, y:uy])
    #     npix += 1
    #     new_label = model_fun(img.reshape(feature_shape))    

    return img, blur_var, blur_expr, blur_fn

def get_model_gradient(model, target, feature_shape, pix_img=None, blur_expr=None, blur_var=None, image_shape=None):

    cost = model.get_default_cost()
    data_specs = cost.get_data_specs(model)
    mapping = DataSpecsMapping(data_specs)
    space_tuple = mapping.flatten(data_specs[0], return_tuple=True)
    source_tuple = mapping.flatten(data_specs[1], return_tuple=True)

    r = None
    features_var = None
    targets_var = None

    theano_args = []
    for space, source in safe_zip(space_tuple, source_tuple):
        name = 'model[%s]' % source
        arg = space.make_theano_batch(name=name)
        if source == 'features':
            features_var = arg
            r = space.make_theano_batch(name='r_in')
        elif source == 'targets':
            targets_var = arg
        theano_args.append(arg)
    theano_args = tuple(theano_args)

    assert r is not None and features_var is not None and targets_var is not None

    nested_args = mapping.nest(theano_args)
    fixed_var_descr = cost.get_fixed_var_descr(model, nested_args)

    expr = cost.expr(model, nested_args, ** fixed_var_descr.fixed_vars)

    if pix_img is not None and blur_expr is not None and blur_var is not None:
        r_blur_expr = theano.clone(blur_expr, replace={blur_var: r})        
        expr = expr + 2000*theano.tensor.mean(theano.tensor.sqr(r_blur_expr - pix_img))

    cost_expr = theano.clone(expr, replace={features_var: r, targets_var: target})
    cost_grad = theano.tensor.grad(cost=cost_expr, wrt=r)

    return (cost_expr, cost_grad, r)

def get_encoded_model_gradient(model, encoder, target, feature_shape, pix_img=None, blur_expr=None, blur_var=None, image_shape=None):

    cost = model.get_default_cost()
    data_specs = cost.get_data_specs(model)
    mapping = DataSpecsMapping(data_specs)
    space_tuple = mapping.flatten(data_specs[0], return_tuple=True)
    source_tuple = mapping.flatten(data_specs[1], return_tuple=True)

    r = None
    features_var = None
    targets_var = None

    theano_args = []
    for space, source in safe_zip(space_tuple, source_tuple):
        name = 'model[%s]' % source
        arg = space.make_theano_batch(name=name)
        if source == 'features':
            features_var = arg
            r = space.make_theano_batch(name='r_in')
        elif source == 'targets':
            targets_var = arg
        theano_args.append(arg)
    theano_args = tuple(theano_args)

    assert r is not None and features_var is not None and targets_var is not None

    hiddens = theano.tensor.matrix(name='hids', dtype=theano.config.floatX)

    nested_args = mapping.nest(theano_args)
    fixed_var_descr = cost.get_fixed_var_descr(model, nested_args)

    expr = cost.expr(model, nested_args, ** fixed_var_descr.fixed_vars)
    if pix_img is not None and blur_expr is not None and blur_var is not None:
        r_blur_expr = theano.clone(blur_expr, replace={blur_var: encoder.decode_expr(hiddens)})
        expr = expr + 1000*theano.tensor.mean(theano.tensor.sqr(r_blur_expr - pix_img))
        cost_expr = theano.clone(expr, replace={features_var: encoder.decode_expr(hiddens), targets_var: target})
    else:
        cost_expr = theano.clone(expr, replace={features_var: encoder.decode_expr(hiddens), targets_var: target})

    cost_grad = theano.tensor.grad(cost=cost_expr, wrt=hiddens)

    return (cost_expr, cost_grad, hiddens)

num_it = 0
def image_inversion(model, target, feature_shape, image_shape, num_elements,
    target_image=None, rate_c=0.1, patience=10000, p_cutoff=0.99, tolerance=0.045, random_start=True, blur=False, display_scale=1,
    make_video=False, video_len=60, starting_image=None, pixelate_width=None, suppress_graphics=False, print_display=True):

    global num_it

    frame_buf = []
    target_idx = np.argmax(target[0])

    if not suppress_graphics:
        cv2.namedWindow('status', cv2.CV_WINDOW_AUTOSIZE)

    height, width, nchannels = image_shape
    costf, gradf, featv = get_model_gradient(model, target, feature_shape, pix_img=target_image)

    predict_expr = model.fprop(featv)
    predict = theano.function([featv], theano.tensor.argmax(predict_expr, axis=1)[0])
    predict_all_fn = theano.function([featv], predict_expr[0], allow_input_downcast=True)
    predict_all = lambda xpt: predict_all_fn(xpt)[target_idx]    

    pix_image = None
    blur_expr = None
    blur_var = None
    blur_fn = lambda x: None
    if starting_image is not None and pixelate_width is not None:        
        pix_image, blur_var, blur_expr, blur_fn = pixelate_img_misclass(starting_image, predict, feature_shape)
        if print_display:
            print "... blurred image label is %d" % (predict(pix_image.reshape(feature_shape)))

    costf, gradf, featv = get_model_gradient(model, target, feature_shape, pix_img=blur_fn(starting_image), blur_expr=blur_expr, blur_var=blur_var, image_shape=image_shape)
    cost = theano.function([featv], costf)
    grad = theano.function([featv], gradf)

    grad_norm = theano.function([featv], outputs=theano.tensor.sqrt(theano.tensor.sum(theano.tensor.sqr(gradf))))

    if pix_image is not None:
        pix_image = pix_image.reshape(feature_shape)

    rate = theano.tensor.scalar(name='rate', dtype=theano.config.floatX)
    if starting_image is not None:
        if pix_image is not None:
            starting_image = pix_image        
        cand_img = theano.shared(value=np.zeros(feature_shape, dtype=theano.config.floatX), name='r', borrow=True)
    elif not random_start:
        cand_img = theano.shared(value=np.zeros(feature_shape, dtype=theano.config.floatX), name='r', borrow=True)        
    else:
        cand_img = theano.shared(value=np.random.sample(num_elements).astype(theano.config.floatX).reshape(feature_shape), name='r', borrow=True)

    invert_update = [(cand_img, cand_img-rate*gradf)]
    invert_step = theano.function([rate, featv], outputs=costf, updates=invert_update)

    if print_display:
        print '... initial label is %d, going for %d' % (predict(cand_img.get_value()), target_idx)

    best_cost = 100000.0
    best_img_bits = None
    prev_costs = [100000.0]*50
    current_cost = 100000.0

    dnoise = 500
    num_it = 0
    for i in range(0, patience):

        num_it = i

        cand_img_bits = normalize(cand_img.get_value(), 0, 1, dtype=theano.config.floatX)

        if i % 100 == 0 and print_display:            
            print 'i=%d, label=%d, cost=%f, norm=%f (rate=%f)' % (i, predict(cand_img_bits), cost(cand_img_bits), grad_norm(cand_img_bits), rate_c)
        
        current_cost = cost(cand_img_bits)                

        cand_img.set_value(cand_img_bits)
        cand_img_bits = normalize(cand_img.get_value(), 0, 1, dtype=theano.config.floatX)

        if current_cost < best_cost:
            best_cost = current_cost
            best_img_bits = np.copy(cand_img_bits)
        if current_cost > max(prev_costs):
            if print_display:
                print "!!! breaking due to gradient chaos"                
            break 
        prev_costs[i%50] = current_cost

        gradv = grad(cand_img_bits)
        diff_mask = gradv

        cand_img_bits = cand_img_bits - rate_c*diff_mask

        if blur and i % 400 == 0:
            cand_img_bits = cv2.GaussianBlur(cand_img_bits, (3, 3), 0)
        if blur and i % 500 == 0:
            cand_img_bits = sharpen(cand_img_bits, 3)
        if blur and i % dnoise == 0:
            # cand_img_bits = normalize(cv2.fastNlMeansDenoising(normalize(cand_img_bits, 0, 255, dtype=np.uint8).reshape(image_shape), h=4).reshape(feature_shape), 0, 1)
            dnoise = int(dnoise * 1.5)
        cand_img.set_value(cand_img_bits)

        if grad_norm(cand_img_bits) < tolerance and False:
            break

        pval = predict_all(cand_img_bits)
        if not suppress_graphics:            
            display_update(cand_img_bits, target_image, pix_image, image_shape, display_scale, 'status', do_sharpen=True, conf=pval)

        if make_video:
            frame_buf.append(process_image_for_display(cand_img_bits, image_shape, display_scale))

        if p_cutoff is not None and predict_all(cand_img_bits) >= p_cutoff:
            break

    finish_time = time.strftime("%Y-%m-%dT%Hh%Mm", time.localtime(time.time()))

    if best_img_bits is None:
        best_img_bits = cand_img.get_value()
    cand_img_bits = normalize(best_img_bits, 0, 1, dtype=theano.config.floatX)
    final_image = process_image_for_display(cand_img_bits, image_shape, display_scale)    

    if not suppress_graphics:
        cv2.waitKey(0)

    if make_video:
        max_frames = float(24*video_len)
        frame_mod = 1 if len(frame_buf) <= max_frames else math.ceil(float(len(frame_buf))/max_frames)
        frames = [frame_buf[i] for i in range(len(frame_buf)) if i % frame_mod == 0]
        print "... writing %d (of %d total) frames to inversion-%s.avi\n\n" % (len(frames), len(frame_buf), finish_time)
        video_pipe = open_video("inversion-%s.avi" % finish_time, image_shape, display_scale)
        for frame in frames:
            video_pipe.stdin.write(frame.tostring())
        video_pipe.terminate()

    return final_image, cand_img_bits, process_image_for_display(pix_image, image_shape, display_scale) if pix_image is not None else None, current_cost

def deblur_image_inversion(model, target, feature_shape, image_shape, num_elements,
    target_image=None, rate_c=0.1, patience=10000, p_cutoff=0.99, tolerance=0.045, random_start=True, blur=False, display_scale=1,
    make_video=False, video_len=60, starting_image=None, pixelate_width=None, suppress_graphics=False, maxlabel=40):

    pvals = {}
    for t_idx in range(maxlabel):

        t_onehot = onehot_label(model, t_idx)
        _, _, _, pval = image_inversion(model, t_onehot, feature_shape, image_shape, num_elements, target_image, 0.05, 100, None, tolerance, random_start, blur, display_scale, False, video_len, starting_image, pixelate_width, suppress_graphics=True, print_display=False)
        pvals[t_idx] = pval

        print "!!!!!!! pval[%d] = %.4f" % (t_idx, pval)

    best_guess = min(pvals, key=lambda x: pvals[x])
    target_onehot = onehot_label(model, best_guess)

    print
    print '*'*60
    print "picked %d" % best_guess
    print

    return image_inversion(model, target_onehot, feature_shape, image_shape, num_elements, target_image, rate_c, patience, p_cutoff, tolerance, random_start, blur, display_scale, make_video, video_len, starting_image, pixelate_width, suppress_graphics)

n_bb_calls = 0
def blackbox_image_inversion(model, target, feature_shape, image_shape, num_elements,
    target_image=None, rate_c=0.1, patience=10000, p_cutoff=0.99, tolerance=0.045, random_start=True, blur=False, display_scale=1,
    make_video=False, video_len=60, starting_image=None, pixelate_width=None, suppress_graphics=False, print_display=True):

    global n_bb_calls
    n_bb_calls = 0
    
    frame_buf = []
    target_idx = np.argmax(target[0])

    if not suppress_graphics:
        cv2.namedWindow('status', cv2.CV_WINDOW_AUTOSIZE)

    height, width, nchannels = image_shape
    costf, gradf, featv = get_model_gradient(model, target, feature_shape, pix_img=target_image)

    predict_expr = model.fprop(featv)
    predict = theano.function([featv], theano.tensor.argmax(predict_expr, axis=1)[0])
    predict_all_fn = theano.function([featv], predict_expr[0], allow_input_downcast=True)
    def predict_all(xpt):
        global n_bb_calls
        n_bb_calls += 1
        return predict_all_fn(xpt)[target_idx]
    # predict_all = lambda xpt: predict_all_fn(xpt)[target_idx]        

    pix_image = None
    blur_expr = None
    blur_var = None
    blur_fn = lambda x: None
    if starting_image is not None and pixelate_width is not None:        
        pix_image, blur_var, blur_expr, blur_fn = pixelate_img_misclass(starting_image, predict, feature_shape)

    costf, gradf, featv = get_model_gradient(model, target, feature_shape, pix_img=blur_fn(starting_image), blur_expr=blur_expr, blur_var=blur_var, image_shape=image_shape)
    cost = theano.function([featv], costf)
    grad = theano.function([featv], gradf)

    grad_norm = theano.function([featv], outputs=theano.tensor.sqrt(theano.tensor.sum(theano.tensor.sqr(gradf))))

    if pix_image is not None:
        pix_image = pix_image.reshape(feature_shape)        

    rate = theano.tensor.scalar(name='rate', dtype=theano.config.floatX)
    if starting_image is not None:
        if pix_image is not None:
            starting_image = pix_image
        cand_img = theano.shared(value=np.zeros(feature_shape, dtype=theano.config.floatX), name='r', borrow=True)
    elif not random_start:
        cand_img = theano.shared(value=np.zeros(feature_shape, dtype=theano.config.floatX), name='r', borrow=True)
    else:
        cand_img = theano.shared(value=np.random.sample(num_elements).astype(theano.config.floatX).reshape(feature_shape), name='r', borrow=True)

    invert_update = [(cand_img, cand_img-rate*gradf)]
    invert_step = theano.function([rate, featv], outputs=costf, updates=invert_update)    

    bounds = [(None, None) for i in range(num_elements)]

    if print_display:
        print '... initial label is %d' % predict(cand_img.get_value())

    best_cost = 100000.0
    best_img_bits = None
    prev_costs = [100000.0]*1
    current_cost = 100000.0    

    for i in range(0, patience):

        cand_img_bits = normalize(cand_img.get_value(), 0, 1, dtype=theano.config.floatX)

        if i % 1 == 0 and print_display:
            print 'i=%d, label=%d, cost=%f, calls=%d (rate=%f) %s' % (i, predict(cand_img_bits), cost(cand_img_bits), n_bb_calls, rate_c, str(predict_all(cand_img_bits)))

        cand_img_bits, f_val, info_d = fmin_l_bfgs_b(lambda xpt: 1.0 - predict_all([xpt]), cand_img_bits, approx_grad=True, bounds=bounds, maxiter=1, epsilon=rate_c)
        cand_img_bits = [cand_img_bits.astype(np.float32)]
        current_cost = f_val
        cand_img.set_value(cand_img_bits)
        cand_img_bits = normalize(cand_img.get_value(), 0, 1, dtype=theano.config.floatX)

        if current_cost < best_cost:
            best_cost = current_cost
            best_img_bits = np.copy(cand_img_bits)
        if current_cost > max(prev_costs):
            if print_display:
                print "!!! breaking due to gradient chaos"                
            break 
        prev_costs[i%1] = current_cost

        if blur and i % 2 == 0:
            cand_img_bits = cv2.GaussianBlur(cand_img_bits, (3, 3), 0)
        if blur and i % 3 == 0:
            cand_img_bits = sharpen(cand_img_bits, 3)

        cand_img.set_value(cand_img_bits)        

        if grad_norm(cand_img_bits) < tolerance and False:
            break

        pval = predict_all(cand_img_bits)
        if not suppress_graphics:
            display_update(cand_img_bits, target_image, pix_image, image_shape, display_scale, 'status', do_sharpen=True, conf=pval)

        if make_video:
            frame_buf.append(process_image_for_display(cand_img_bits, image_shape, display_scale))
        
        if p_cutoff is not None and predict_all(cand_img_bits) >= p_cutoff:
            print "breaking due to cutoff"            
            break

    finish_time = time.strftime("%Y-%m-%dT%Hh%Mm", time.localtime(time.time()))

    if best_img_bits is None:
        best_img_bits = cand_img.get_value()
    cand_img_bits = normalize(best_img_bits, 0, 1, dtype=theano.config.floatX)
    final_image = process_image_for_display(cand_img_bits, image_shape, display_scale)    
    
    if not suppress_graphics:
        cv2.waitKey(0)

    if make_video:
        max_frames = float(24*video_len)
        frame_mod = 1 if len(frame_buf) <= max_frames else math.ceil(float(len(frame_buf))/max_frames)
        frames = [frame_buf[i] for i in range(len(frame_buf)) if i % frame_mod == 0]
        print "... writing %d (of %d total) frames to inversion-%s.avi\n\n" % (len(frames), len(frame_buf), finish_time)
        video_pipe = open_video("inversion-%s.avi" % finish_time, image_shape, display_scale)
        for frame in frames:
            video_pipe.stdin.write(frame.tostring())
        video_pipe.terminate()

    return final_image, cand_img_bits, process_image_for_display(pix_image, image_shape, display_scale) if pix_image is not None else None, current_cost

def deblur_blackbox_image_inversion(model, target, feature_shape, image_shape, num_elements,
    target_image=None, rate_c=0.1, patience=10000, p_cutoff=0.99, tolerance=0.045, random_start=True, blur=False, display_scale=1,
    make_video=False, video_len=60, starting_image=None, pixelate_width=None, suppress_graphics=False, maxlabel=40):

    pvals = {}
    for t_idx in range(maxlabel):

        t_onehot = onehot_label(model, t_idx)
        _, _, _, pval = blackbox_image_inversion(model, t_onehot, feature_shape, image_shape, num_elements, target_image, 0.01, 50, None, tolerance, random_start, blur, display_scale, False, video_len, starting_image, pixelate_width, suppress_graphics=True, print_display=False)
        pvals[t_idx] = pval

        print "!!!!!!! pval[%d] = %.4f" % (t_idx, pval)

    best_guess = min(pvals, key=lambda x: pvals[x])
    target_onehot = onehot_label(model, best_guess)

    print
    print '*'*60
    print "picked %d" % best_guess
    print

    return blackbox_image_inversion(model, target_onehot, feature_shape, image_shape, num_elements, target_image, rate_c, patience, p_cutoff, tolerance, random_start, blur, display_scale, make_video, video_len, starting_image, pixelate_width, suppress_graphics)

def image_encoded_inversion(model, encoder, target, feature_shape, model_feature_shape, image_shape, num_elements,
    target_image=None, rate_c=0.1, patience=10000, p_cutoff=0.99, tolerance=0.045, random_start=True, blur=False, display_scale=1,
    make_video=False, video_len=60, starting_image=None, pixelate_width=10, suppress_graphics=False, print_display=True, do_recode=True):

    global num_it

    frame_buf = []
    target_idx = np.argmax(target[0])

    if not suppress_graphics:
        cv2.namedWindow('status', cv2.CV_WINDOW_AUTOSIZE)

    height, width, nchannels = image_shape
    
    costf, gradf, featv = get_encoded_model_gradient(model, encoder, target, feature_shape)

    predict_raw_expr = model.fprop(featv)
    predict_raw = theano.function([featv], theano.tensor.argmax(predict_raw_expr, axis=1)[0])

    predict_expr = model.fprop(encoder.decode_expr(featv))
    predict_all_fn = theano.function([featv], predict_expr[0], allow_input_downcast=True)
    predict_all = lambda xpt: predict_all_fn(xpt)[target_idx]    
    likely = theano.function([featv], theano.tensor.max(predict_expr, axis=1)[0])
    predict = theano.function([featv], theano.tensor.argmax(predict_expr, axis=1)[0])        

    pix_image = None
    blur_expr = None
    blur_var = None
    blur_fn = lambda x: None

    if starting_image is not None and pixelate_width is not None:        
        pix_image, blur_var, blur_expr, blur_fn = pixelate_img_misclass(starting_image, predict_raw, model_feature_shape)            
    
    # pix_img_reshaped = None if pix_image is None else pix_image.reshape(model_feature_shape)
    costf, gradf, featv = get_encoded_model_gradient(model, encoder, target, feature_shape, pix_img=blur_fn(starting_image), blur_expr=blur_expr, blur_var=blur_var)
    cost = theano.function([featv], costf)
    grad = theano.function([featv], gradf)

    grad_norm = theano.function([featv], outputs=theano.tensor.sqrt(theano.tensor.sum(theano.tensor.sqr(gradf))))

    rate = theano.tensor.scalar(name='rate', dtype=theano.config.floatX)
    if starting_image is not None:
        if pix_image is None:
            pix_image = np.copy(starting_image)
        starting_image = encoder.encode(pix_image.reshape(model_feature_shape))        
        cand_img = theano.shared(value=np.zeros(feature_shape, dtype=theano.config.floatX), name='r', borrow=True)
        # cand_img = theano.shared(value=np.random.sample(num_elements).astype(theano.config.floatX).reshape(feature_shape), name='r', borrow=True)
    elif not random_start:
        cand_img = theano.shared(value=np.zeros(feature_shape, dtype=theano.config.floatX), name='r', borrow=True)        
    else:
        cand_img = theano.shared(value=np.random.sample(num_elements).astype(theano.config.floatX).reshape(feature_shape), name='r', borrow=True)

    invert_update = [(cand_img, cand_img-rate*gradf)]
    invert_step = theano.function([rate, featv], outputs=costf, updates=invert_update)

    if print_display:
        print '... initial label is %d' % predict(cand_img.get_value())

    best_cost = 100000.0
    best_img_bits = None
    prev_costs = [100000.0]*100
    current_cost = 100000.0

    num_it = 0
    for i in range(0, patience):

        num_it = i

        # if i % int(100*math.log(i+2,10)) == 200:
        if i == 50 and do_recode:

            cand_img_bits = cand_img.get_value()
            # pdb.set_trace()
            inds = np.arange(num_elements)
            np.random.shuffle(inds)
            pct = round(0.05*num_elements)
            # cand_img_bits[0,inds[:pct]] = 0 #np.random.random_sample(pct)
            cand_img.set_value(cand_img_bits)

            cand_img_bits = encoder.decode(cand_img)
            # cand_img_bits = normalize(cand_img_bits, 0, 1)
            # cand_img_bits = cv2.fastNlMeansDenoising(cand_img_bits.reshape(image_shape), h=6)
            cand_img_bits = normalize(cv2.fastNlMeansDenoising(normalize(cand_img_bits, 0, 255, dtype=np.uint8).reshape(image_shape), h=4), 0, 1)
            cand_img_bits = sharpen(cand_img_bits, 3)
            cand_img_bits = cv2.GaussianBlur(cand_img_bits, (3, 3), 0)
            cand_img_bits = encoder.encode(cand_img_bits.reshape(model_feature_shape))
            cand_img.set_value(cand_img_bits)            
        
        cand_img_bits = cand_img.get_value()

        # knockout learning
        if i % 200 == 99 and False:
            inds = np.arange(num_elements)
            np.random.shuffle(inds)
            pct = round(0.05*num_elements)
            cand_img_bits[0,inds[:pct]] = 0 #np.random.random_sample(pct)
            cand_img.set_value(cand_img_bits)

        current_cost = invert_step(rate_c, cand_img_bits)
        cand_img_bits = cand_img.get_value()

        if current_cost < best_cost:
            best_cost = current_cost
            best_img_bits = np.copy(cand_img_bits)
        if current_cost > max(prev_costs):
            if print_display:
                print "!!! breaking due to gradient chaos"                
            break 
        prev_costs[i%100] = current_cost        

        if i % 100 == 0 and print_display:
            print 'i=%d, label=%d, cost=%f, norm=%f' % (i, predict(cand_img_bits), current_cost, grad_norm(cand_img_bits))            

        if grad_norm(cand_img_bits) < tolerance and False:
            break

        pval = predict_all(cand_img_bits)
        if not suppress_graphics:
            dec_img = encoder.decode(cand_img)
            # dec_img = normalize(cv2.fastNlMeansDenoising(normalize(dec_img, 0, 255, dtype=np.uint8).reshape(image_shape), h=6), 0, 1)
            # dec_img = sharpen(dec_img, 3)
            # dec_img = cv2.GaussianBlur(dec_img, (3, 3), 0)            
            display_update(dec_img, target_image, pix_image, image_shape, display_scale, 'status', do_sharpen=True, conf=pval)

        if p_cutoff is not None and predict_all(cand_img_bits) >= p_cutoff:
            break

    finish_time = time.strftime("%Y-%m-%dT%Hh%Mm", time.localtime(time.time()))

    if best_img_bits is None:
        best_img_bits = cand_img.get_value()    
    cand_img_bits = encoder.decode(best_img_bits)
    cand_img_bits = normalize(cv2.fastNlMeansDenoising(normalize(cand_img_bits, 0, 255, dtype=np.uint8).reshape(image_shape), h=6), 0, 1)
    # cand_img_bits = sharpen(cand_img_bits, 3)
    cand_img_bits = cv2.GaussianBlur(cand_img_bits, (3, 3), 0)    
    final_image = process_image_for_display(cand_img_bits, image_shape, display_scale)    
    
    if not suppress_graphics:
        cv2.waitKey(0)

    return final_image, cand_img_bits, process_image_for_display(pix_image, image_shape, display_scale) if pix_image is not None else None, current_cost

def deblur_image_encoded_inversion(model, encoder, target, feature_shape, model_feature_shape, image_shape, num_elements,
    target_image=None, rate_c=0.1, patience=10000, p_cutoff=0.99, tolerance=0.045, random_start=True, blur=False, display_scale=1,
    make_video=False, video_len=60, starting_image=None, pixelate_width=None, suppress_graphics=False, maxlabel=40):

    pvals = {}
    for t_idx in range(maxlabel):

        t_onehot = onehot_label(model, t_idx)
        _, _, _, pval = image_encoded_inversion(model, encoder, t_onehot, feature_shape, model_feature_shape, image_shape, num_elements, target_image, 0.1, 100, None, tolerance, random_start, blur, display_scale, False, video_len, starting_image, pixelate_width, suppress_graphics=True, print_display=False, do_recode=False)
        pvals[t_idx] = pval

        print "!!!!!!! pval[%d] = %.4f" % (t_idx, pval)

    best_guess = min(pvals, key=lambda x: pvals[x])
    target_onehot = onehot_label(model, best_guess)

    print
    print '*'*60
    print "picked %d" % best_guess
    print

    return image_encoded_inversion(model, encoder, target_onehot, feature_shape, model_feature_shape, image_shape, num_elements, target_image, rate_c, patience, p_cutoff, tolerance, random_start, blur, display_scale, make_video, video_len, starting_image, pixelate_width, suppress_graphics)


def write_output(output_dir, model_file, final_image, pix_image, target_image_file, target_label_idx, rate, patience, p_cutoff, 
    random_start, blur, starting_image_file, pixelate_width, scale, msecs):
    
    global n_bb_calls
    global num_it

    assert final_image is not None

    output_config_file = os.path.join(output_dir, "config.json")
    output_target_image_file = os.path.join(output_dir, "target_img.png")
    output_starting_image_file = os.path.join(output_dir, "starting_image.png")    
    output_final_image_file = os.path.join(output_dir, "final_image.png")
    output_pix_image_file = os.path.join(output_dir, "blurred_image.png")

    config = {}
    config["model"] = model_file
    if target_image_file is not None:
        config["target_image"] = target_image_file
    config["target_label"] = target_label_idx
    if rate is not None:
        config["rate"] = rate
    if patience is not None:
        config["patience"] = patience
    if p_cutoff is not None:
        config["p_cutoff"] = p_cutoff
    if random_start is not None:
        config["random_start"] = random_start
    if blur is not None:
        config["blur"] = blur
    if starting_image_file is not None:
        config["starting_image"] = starting_image_file
    if pixelate_width is not None:
        config["pixelate_width"] = pixelate_width
    if msecs is not None:
        config["msecs"] = msecs
    if n_bb_calls > 0:
        config["calls"] = n_bb_calls
    if num_it > 0:
        config["num_it"] = num_it

    if not os.path.exists(output_dir):
        os.mkdir(output_dir)

    with open(output_config_file, "w") as output_config:        
        json.dump(config, output_config, sort_keys=True, indent=4)

    cv2.imwrite(output_final_image_file, final_image)
    if pix_image is not None:
        cv2.imwrite(output_pix_image_file, pix_image)

    if target_image_file is not None:
        img = cv2.resize(cv2.imread(target_image_file), None, fx=scale, fy=scale)
        cv2.imwrite(output_target_image_file, img)        
    if starting_image_file is not None:
        img = cv2.resize(cv2.imread(starting_image_file), None, fx=scale, fy=scale)
        cv2.imwrite(output_starting_image_file, img)

def process_args():
    parser = argparse.ArgumentParser(description='Model inversion for pylearn2 image classifiers.')
    parser.add_argument('--model', metavar='model_file', type=str, help='recognition model (.pkl produced by pylearn2)')
    parser.add_argument('--target_label', metavar='target_label', type=int, help='target label')
    parser.add_argument('--target_image', metavar='target_image', type=str, help='target image')
    parser.add_argument('--starting_image', metavar='starting_image', type=str, help='starting image')
    parser.add_argument('--img_width', type=int, default=None, help='image width')
    parser.add_argument('--img_height', type=int, default=None, help='image height')
    parser.add_argument('--img_channels', type=int, default=1, help='# image channels (default: 1)')
    parser.add_argument('--blackbox', action='store_true', help='use blackbox algorithm')
    parser.add_argument('--rate', type=float, default=0.1, help='attack rate (default: 0.1)')
    parser.add_argument('--patience', type=int, default=10000, help='attack patience (default: None)')
    parser.add_argument('--p_cutoff', type=int, default=0.99, help='attack patience (default: 0.99)')
    parser.add_argument('--scale', type=int, default=1, help='factor to scale display image by')
    parser.add_argument('--video_length', type=int, default=60, help='time limit for output video (default: 60)')
    parser.add_argument('--pixelate_width', type=int, default=None, help='width of starting point pixelation (default: None)')
    parser.add_argument('--random', action='store_true', help='use random starting point')
    parser.add_argument('--blur', action='store_true', help='apply gaussian blur to image')
    parser.add_argument('--video', action='store_true', help='make video on completion')
    parser.add_argument('--output_dir', metavar='output_dir', type=str, default=None, help='directory to store statistics')
    parser.add_argument('--experiment_config', metavar='config_file', type=str, default=None, help='take options from experiment config file')
    parser.add_argument('--suppress_graphics', action='store_true', help='suppress graphical status')
    parser.add_argument('--encoder', metavar='encoder', type=str, default=None, help='Encoder cascade')

    return parser.parse_args()

def main(config_filename=None):

    if config_filename is None:
        args = process_args()    
    
    config_file = args.experiment_config if config_filename is None else config_filename

    if config_filename is None:
        model_file = args.model
        target_label = args.target_label
        target_image = args.target_image
        width = args.img_width
        height = args.img_height
        channels = args.img_channels
        rate = args.rate
        patience = args.patience
        p_cutoff = args.p_cutoff
        random_start = args.random
        blur = args.blur
        scale = args.scale
        video_length = args.video_length
        make_video = args.video        
        starting_image = args.starting_image        
        pixelate_width = args.pixelate_width
        use_blackbox = args.blackbox
        output_dir = args.output_dir
        suppress_graphics = args.suppress_graphics
        encoder_file = args.encoder
    if config_file is not None:
        with open(config_file) as config_fp:
            config = json.load(config_fp)
        model_file = config["model"]
        target_label = config["target_label"] if config.has_key("target_label") else None
        target_image = config["target_image"] if config.has_key("target_image") else None
        width = config["img_width"] if config.has_key("img_width") else None
        height = config["img_height"] if config.has_key("img_height") else None
        channels = config["img_channels"] if config.has_key("img_channels") else None
        rate = config["rate"] if config.has_key("rate") else None
        patience = config["patience"] if config.has_key("patience") else 10000
        p_cutoff = config["p_cutoff"] if config.has_key("p_cutoff") else None
        random_start = config["random"] if config.has_key("random_start") else False
        blur = config["blur"] if config.has_key("blur") else False
        scale = config["scale"] if config.has_key("scale") else 2
        video_length = config["video_length"] if config.has_key("video_length") else 60
        make_video = config["video"] if config.has_key("video") else False
        starting_image = config["starting_image"] if config.has_key("starting_image") else None
        pixelate_width = config["pixelate_width"] if config.has_key("pixelate_width") else None
        use_blackbox = config["blackbox"] if config.has_key("blackbox") else False
        output_dir = config["output_dir"] if config.has_key("output_dir") else False
        suppress_graphics = config["suppress_graphics"] if config.has_key("suppress_graphics") else False
        encoder_file = config["encoder"] if config.has_key("encoder") else None

    if encoder_file is not None:
        with open(encoder_file, 'rb') as fp:
            encoder = cPickle.load(fp)
    else:
        encoder = None

    model = serial.load(model_file)
    model_feature_shape = get_model_feature_shape(model)
    if encoder is not None:
        feature_shape = (1, encoder.nhid)
    else:
        feature_shape = model_feature_shape
    num_elements = reduce(lambda prod, el: prod*el, list(feature_shape))
    image_shape = None
    if width is None or height is None:
        image_shape = tuple(list(feature_shape)[1:])
    else:
        image_shape = (height, width, channels)

    target_image_file = target_image
    if target_image is not None:            
        target_image = normalize(cv2.cvtColor(cv2.imread(target_image), cv2.COLOR_BGR2GRAY), 0, 1).reshape(model_feature_shape)
        featv = theano.tensor.matrix(name='input')
        predict_target_image = theano.function([featv], theano.tensor.argmax(model.fprop(featv), axis=1)[0])
        target_label = onehot_label(model, predict_target_image(target_image))
        target_image_costf, target_image_gradf, target_image_featv = get_model_gradient(model, target_label, model_feature_shape)
        target_image_cost = theano.function([target_image_featv], target_image_costf)
        print "... target_image label=%d, cost=%f" % (predict_target_image(target_image), target_image_cost(target_image))
    else:        
        target_label = onehot_label(model, target_label)
    target_label_idx = np.argmax(target_label)

    starting_image_file = starting_image
    if starting_image is not None:        
        starting_image = normalize(cv2.cvtColor(cv2.imread(starting_image), cv2.COLOR_BGR2GRAY), 0, 1)

    start = int(round(time.time() * 1000))
    if encoder is not None:
        
        if pixelate_width is None:
            inverted_image, bits, pix_image, pval = image_encoded_inversion(model, encoder, target_label, feature_shape, model_feature_shape, image_shape, num_elements,
                                target_image=target_image, rate_c=rate, patience=patience, p_cutoff=p_cutoff, random_start=random_start, blur=blur,
                                display_scale=scale, make_video=make_video, video_len=video_length, starting_image=starting_image,
                                pixelate_width=pixelate_width, suppress_graphics=suppress_graphics)
        else:            
            inverted_image, bits, pix_image, pval = deblur_image_encoded_inversion(model, encoder, target_label, feature_shape, model_feature_shape, image_shape, num_elements,
                                target_image=target_image, rate_c=rate, patience=patience, p_cutoff=p_cutoff, random_start=random_start, blur=blur,
                                display_scale=scale, make_video=make_video, video_len=video_length, starting_image=starting_image,
                                pixelate_width=pixelate_width, suppress_graphics=suppress_graphics)

    else:
        if not use_blackbox:
            if pixelate_width is None:
                inverted_image, bits, pix_image, pval = image_inversion(model, target_label, feature_shape, image_shape, num_elements,
                                    target_image=target_image, rate_c=rate, patience=patience, p_cutoff=p_cutoff, random_start=random_start, blur=blur,
                                    display_scale=scale, make_video=make_video, video_len=video_length, starting_image=starting_image,
                                    pixelate_width=pixelate_width, suppress_graphics=suppress_graphics)
            else:
                inverted_image, bits, pix_image, pval = deblur_image_inversion(model, target_label, feature_shape, image_shape, num_elements,
                    target_image=target_image, rate_c=rate, patience=patience, p_cutoff=p_cutoff, random_start=random_start, blur=blur,
                    display_scale=scale, make_video=make_video, video_len=video_length, starting_image=starting_image,
                    pixelate_width=pixelate_width, suppress_graphics=suppress_graphics)

        else:
            if pixelate_width is None:
                inverted_image, bits, pix_image, pval = blackbox_image_inversion(model, target_label, feature_shape, image_shape, num_elements,
                        target_image=target_image, rate_c=rate, patience=patience, p_cutoff=p_cutoff, random_start=random_start, blur=blur,
                        display_scale=scale, make_video=make_video, video_len=video_length, starting_image=starting_image,
                        pixelate_width=pixelate_width, suppress_graphics=suppress_graphics)
            else:
                inverted_image, bits, pix_image, pval = deblur_blackbox_image_inversion(model, target_label, feature_shape, image_shape, num_elements,
                        target_image=target_image, rate_c=rate, patience=patience, p_cutoff=p_cutoff, random_start=random_start, blur=blur,
                        display_scale=scale, make_video=make_video, video_len=video_length, starting_image=starting_image,
                        pixelate_width=pixelate_width, suppress_graphics=suppress_graphics)

    end = int(round(time.time() * 1000))
    time_ms = end - start

    if output_dir is not None:
        write_output(output_dir, model_file, inverted_image, pix_image, target_image_file, target_label_idx, rate, patience, 
            p_cutoff, random_start, blur, starting_image_file, pixelate_width, scale, time_ms)
    else:
        cv2.namedWindow('final', cv2.CV_WINDOW_AUTOSIZE)
        cv2.imshow('final', inverted_image)
        cv2.waitKey(0)


if __name__ == '__main__':
    main()
