import sys
import csv
import math
import json

from sklearn import metrics
from bigml.model import Model
from bigml.api import BigML

sys.path.append('../bigml')
import steak
import steak2
import steak3
import steakcheat

def remove_quoted_commas(line):

	in_quote = False
	new_line = ""
	
	for i in range(len(line)):
		if line[i] == '\"':
			in_quote = not in_quote
			continue
		if line[i] == ',' and in_quote:
			continue
		new_line = "%s%s" % (new_line, line[i])
	return new_line

def has_missing(row):

	return True if sum([1 if el != '' else 0 for el in row]) < len(row) else False	

def read_csv(filename, attributes, classindex, ignore_missing=True):

	data = []

	with open(filename, 'rb') as f:
		dialect = csv.Sniffer().sniff(f.read(1024))
		f.seek(0)
		reader = csv.reader(f, dialect)
		reader.next()		
		for row in reader:
			if ignore_missing and has_missing(row):
				continue

			drow = {}
			for i in range(len(row)):
				drow[attributes[i]] = remove_quoted_commas(row[i])
			drow["__target__"]	= remove_quoted_commas(row[classindex])

			data.append(drow)

	return data

def write_csv(filename, data, attributes):

	with open(filename, 'wb') as f:
		first = True
		for attr in attributes:
			quotes = "\"" if (',' in attr) else ""
			f.write("%s%s%s%s" % ("," if not first else "", quotes, attr, quotes))
			first = False
		f.write("\n")
		for row in data:
			first = True
			for attr in attributes:
				quotes = "\"" if (',' in row[attr]) else ""
				f.write("%s%s%s%s" % ("," if not first else "", quotes, row[attr], quotes))
				first = False
			f.write("\n")

def read_attributes(filename):

	attr_indices = {}
	attributes = []

	with open(filename, 'rb') as f:
		line = f.readline().strip()
		line = remove_quoted_commas(line)
		attrs = line.split(',')
		for i in range(len(attrs)):
			attr_indices[attrs[i]] = i
			attributes.append(attrs[i])

	return attr_indices, attributes

def print_row(row, attributes, tab=1):

	rstr = ""
	prefix = "\t"*tab
	for attr in attributes:
		rstr = "%s%s%s: %s\n" % (rstr, prefix, attr, row[attr])

	return rstr

def domains(data, attributes):

	allkeys = attributes[:]
	allkeys.extend(range(len(attributes)))	
	doms = {attr: [] for attr in attributes}

	for row in data:
		for (attr, idx) in zip(attributes, range(len(attributes))):
			if row[attr] not in doms[attr]:
				doms[attr].append(row[attr])
				# doms[idx].append(row[attr])

	return doms

def marginals(data, doms):

	counts = {(k, v): 0.0 for k in doms.keys() for v in doms[k]}
	nrows = float(len(data))

	for row in data:
		for attr in doms.keys():
			counts[(attr, row[attr])] += 1.0/nrows

	ret_marginals = {attr: [(val, counts[(attr, val)]) for val in doms[attr]] for attr in doms.keys()}
	for attr in doms.keys():
		margs = sorted(ret_marginals[attr], key=lambda (val, p): -1*p)
		ret_marginals[attr] = margs

	return ret_marginals

def marg_prob(attr, val, margs):

	assert attr in margs.keys()

	for (cval, p) in margs[attr]:
		if val == cval:
			return p

	return 0.0

def joint_prob(row, marg_attrs, margs):

	p = 0.0
	for attr in marg_attrs:
		p += math.log(marg_prob(attr, row[attr], margs))

	return math.exp(p)

def confidence(data, model, config):

	all_matches = 0.0
	y_matches = 0.0
	for row in data:
		if row["__target__"] == config["__target__"]:
			all_matches += 1.0
			if model(row) == model(config):
				y_matches += 1

	p = max(y_matches/all_matches, 0.0000001)
	return p

def attribute_indices(attributes, attr_indices):

	return map(lambda attr: attr_indices[attr], attributes)

def fillin(row, missing, doms):

	newrows = [dict(row)]
	for m in missing:		
		for el in doms[m]:
			additions = []			
			for r in newrows:
				if r[m] == el:
					continue
				newr = dict(r)
				r[m] = el
				additions.append(newr)
			newrows.extend(additions)
	
	for i in range(len(newrows)-1):
		for j in range(i+1, len(newrows)):
			if str(newrows[i]) == str(newrows[j]):
				newrows.pop(i)
				break

	return newrows

def feasible_target(model, row, target, doms, missing=None, margs=None, data=None):

	allmissing = [] if missing is None else missing[:]
	allmissing.append(target)
	X = fillin(row, allmissing, doms)	

	Y = {val: 0.0 for val in doms[target]}	
	for x in X:
		# print print_row(x, doms.keys())
		if data is None:
			if model.predict(x) == row['__target__']:
				Y[x[target]] += joint_prob(x, allmissing, margs) if margs is not None else 1
		else:
			pr = model.predict(x, with_confidence=True)			
			label, pathconf, dist, total = pr
			conf = 0.0
			for d in dist:
				if d[0] == row['__target__']:
					conf = float(d[1])/float(total)
					break
			if conf > 0.0:
				# print "conf=%.2f, %s" % (100*conf, str(pr))
				Y[x[target]] += math.exp(math.log(joint_prob(x, allmissing, margs)) + math.log(conf) + math.log(pathconf))
				# Y[x[target]] += math.exp(math.log(joint_prob(x, allmissing, margs))+(math.log(conf)+math.log(pathconf)))

	# for v in Y.keys():
	# 	Y[v] = math.exp(Y[v])

	return Y

def predict_target(target, feasible, doms, margs=None):

	if sum([feasible[v] for v in doms[target]]) == 0:		
		return margs[target][0][0] if margs is not None else doms[target][0]

	return max(feasible, key=lambda v: feasible[v])

(attr_indices, attributes) = read_attributes(sys.argv[1])
classidx = attr_indices[sys.argv[2]]
target = sys.argv[3]
data = read_csv(sys.argv[1], attributes, classidx)
doms = domains(data, attributes)
margs = marginals(data, doms)

# api = BigML('mfredrik', 'cfabe94c349954db760425e48d95f6d61568b269')
with open('../bigml/steak-raw.json') as json_data:
	local_model = Model(json.load(json_data))
with open('../bigml/steakcheat-full.json') as json_data:
	cheat_model = Model(json.load(json_data))

# local_model = steak
# print local_model.predict(data[2], with_confidence=True)
# x = dict(data[2])
# for attr in data[0].keys():
# 	del x[attr]
# 	print local_model.predict(data[0], with_confidence=True)
# quit()

# write_csv("/tmp/test.csv", data, attributes)
# quit()

misclass = 0
cheaters = 0
innocent = 0
clean = 0
true = [1 if row[target] == "Yes" else 0 for row in data]
pred = []
for row in data:
	feasible = feasible_target(local_model, row, target, doms, ["lottery_a_or_lottery_b_"], margs, data=1)
	# feasible = feasible_target(local_model, row, target, doms, [], margs, data=1)
	print feasible
	print '-'*60
	predicted = predict_target(target, feasible, doms, margs)
	pred.append(1 if predicted == "Yes" else 0)
	if predicted != row[target]:
		misclass += 1
		if predicted == "Yes":
			innocent += 1
		if predicted == "No":
			clean += 1
	if predicted == row[target] and predicted == "Yes":
		cheaters += 1
print "misclassified: %d of %d (%.2f pct)" % (misclass, len(data), 100.0*float(misclass)/float(len(data)))
print "found %d cheaters with %d false positives, %d false negatives" % (cheaters, innocent, clean)
print "aucroc:", metrics.roc_auc_score(true, pred)

print "\n"*2
misclass = 0
for row in data:
	misclass += 1 if cheat_model.predict(row) != row[target] else 0
print "misclassified: %d of %d (%.2f)" % (misclass, len(data), float(misclass)/float(len(data)))


