#!/usr/bin/env python
# Software License Agreement (BSD License)
#
# Copyright (c) 2012, Philipp Wagner <bytefish[at]gmx[dot]de>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of the author nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# ------------------------------------------------------------------------------------------------
# Note:
# When using the FaceRecognizer interface in combination with Python, please stick to Python 2.
# Some underlying scripts like create_csv will not work in other versions, like Python 3.
# ------------------------------------------------------------------------------------------------

import os
import sys
import cv2
import sympy
import numpy as np
import shutil
import subprocess

from scipy import optimize
from sympy.utilities.lambdify import lambdify

import theano 
import theano.tensor as T

from theano import sandbox

#np.set_printoptions(threshold='nan')

def normalize(X, low, high, dtype=None):
	"""Normalizes a given array in X to a value between low and high."""
	X = np.asarray(X)
	minX, maxX = np.min(X), np.max(X)
	# normalize to [0...1].
	X = X - float(minX)
	X = X / float((maxX - minX))
	# scale to [low...high].
	X = X * (high-low)
	X = X + low
	if dtype is None:
		return np.asarray(X)
	return np.asarray(X, dtype=dtype)

def make_status_image(attacker_raw, attacker_noised, victim, shape):

	height, width = shape
	combined = np.concatenate((normalize(attacker_raw.reshape(shape), 0, 255, dtype=np.uint8), normalize(attacker_noised.reshape(shape), 0, 255, dtype=np.uint8), normalize(victim.reshape(shape), 0, 255, dtype=np.uint8)), axis=1)

	return cv2.resize(normalize(combined, 0, 255, dtype=np.uint8), (2*3*width, 2*height))	

def read_images(path, splitsize=None, splitstart=None, nclasses=None, sz=None):
	"""Reads the images in a given folder, resizes images on the fly if size is given.

	Args:
		path: Path to a folder with subfolders representing the subjects (persons).
		splitsize: Number of images from each directory to read
		splitstart: Offset within each directory to start reading images
		nclasses: Number of image subdirectories (classes) to read
		sz: A tuple with the size Resizes

	Returns:
		A list [X,y]

			X: The images, which is a Python list of numpy arrays.
			y: The corresponding labels (the unique number of the subject, person) in a Python list.
	"""
	c = 0
	X,y = [], []
	print path	
	for dirname, dirnames, filenames in os.walk(path):
		ctr1 = 0
		for subdirname in dirnames:
			if nclasses is not None and ctr1 >= nclasses:
				break
			ctr1 += 1
			ctr2 = 0
			subject_path = os.path.join(dirname, subdirname)
			for filename in os.listdir(subject_path):
				if splitstart is not None and ctr2 < splitstart:
					ctr2 += 1
					continue
				if splitsize is not None and splitstart is not None and ctr2 >= splitsize+splitstart:
					break
				ctr2 += 1				
				if filename.startswith(".") or not (filename.endswith(".jpg") or filename.endswith(".pgm")):
					continue
				try:
					im = cv2.imread(os.path.join(subject_path, filename), cv2.IMREAD_GRAYSCALE)
					# resize to given size (if given)
					if (sz is not None):
						im = cv2.resize(im, sz)
					X.append(np.asarray(im, dtype=np.uint8))									
					y.append(c)
				except IOError, (errno, strerror):
					print "I/O error({0}): {1}".format(errno, strerror)
				except:
					print "Unexpected error:", sys.exc_info()[0]
					raise
			c = c+1
	return [X,y]

def get_nearest(projections, w):

	nearest = None
	tidx = None		
	for i in range(0, len(projections)):
		p = projections[i]
		norm = np.linalg.norm(p-w)
		if nearest is None:
			nearest = norm
			tidx = i				
		if norm < nearest:
			nearest = norm
			tidx = i
		break		

	return (projections[idx], tidx, float(nearest))

def get_weights(t, eigenvectors, mean):
	sub = t - mean
	return sub.dot(eigenvectors)

def mask_fn(idx, img, shape):
	
	height, width = shape

	width_cons = True	
	height_cons = img[0,idx] >= 0.5

	# width_cons = idx%width > width/7 and idx%width < 5*width/6
	# height_cons = idx/width > height/7 and idx/width < 6*height/7

	# width_cons = True
	# height_cons = True

	return width_cons and height_cons	

def attack_eigenfaces(X, y, X_t, y_t, model, idx, target, reg, attack_rate=1, tolerance=0.01, attack_patience=10000, make_video=False):
	
	[p_label, p_confidence] = model.predict(np.asarray(X_t[idx]))	
	print "Attacker's predicted label = %d (confidence=%.2f, real label = %d)" % (p_label, p_confidence, y_t[idx])

	X_shape = X_t[idx].shape
	(height, width) = X_shape
	col_width = height*width
	nent = height*width		
	eigenvectors = model.getMat("eigenvectors")
	training_projections = model.getMatVector("projections")
	mean = model.getMat("mean")
	orig = X_t[idx].reshape(mean.shape)
	init = np.zeros_like(orig, dtype=theano.config.floatX)+255

	attacker = X_t[idx]
	victim = X[5*target]	
	
	target_projections = []	
	for i in range(0, len(training_projections)):		
		if y[i] == target:			
			target_projections.append(training_projections[i])


	(w_nearest, idx_nearest, _) = get_nearest(target_projections, get_weights(orig, eigenvectors, mean))

	mask_native = [1.0 if mask_fn(j, orig, X_shape) else 0.0 for j in range(0,height*width)]
	mask = theano.shared(value=np.asarray(mask_native, dtype=theano.config.floatX), name='mask', borrow=False)	

	rate = T.scalar('rate', dtype=theano.config.floatX)
	delta = T.scalar('delta', dtype=theano.config.floatX)
	lX = T.scalar('minX', dtype=theano.config.floatX)
	hX = T.scalar('maxX', dtype=theano.config.floatX)
	class_conf = T.scalar('class_conf', dtype=theano.config.floatX)
	eig_row, eig_col = eigenvectors.shape

	r_V = T.vector('r')	
		
	#r_S = theano.shared(value=np.zeros_like(orig, dtype=theano.config.floatX).reshape(height*width,), name='r_actual', borrow=False)
	r_S = theano.shared(value=np.random.sample(height*width).astype(theano.config.floatX)*256, name='r_actual', borrow=False)

	original_S = theano.shared(value=orig.astype(theano.config.floatX), name='original', borrow=False)
	mean_S = theano.shared(value=mean.astype(theano.config.floatX), name='mean', borrow=False)
	eigens_S = theano.shared(value=eigenvectors.astype(theano.config.floatX), name='eigenvectors', borrow=False)
	w_S = theano.shared(value=w_nearest.astype(theano.config.floatX), name='w', borrow=False)
	
	X = r_V + original_S
	X_normalized = T.clip(X, 0, 255) # ((X - lX)/(hX - lX))*255
	sub_n = X_normalized - mean_S
	weights_n = sub_n.dot(eigens_S)
	normed = T.max([T.sum((weights_n - w_S)**2) - class_conf*class_conf, 0]) + delta*T.mean(r_V**2)
	#normed = T.sum((weights_n - w_S)**2) - class_conf**2 + reg*T.mean(r_V**2)
	gradient = T.grad(cost=normed, wrt=r_V)
	
	cost_f = theano.function([class_conf, delta], outputs=normed, givens={lX: T.min(r_S+original_S), hX: T.max(r_S+original_S), r_V: r_S}, on_unused_input='ignore')
	grad_f = theano.function([class_conf, delta], outputs=gradient, givens={lX: T.min(r_S+original_S), hX: T.max(r_S+original_S), r_V: r_S}, on_unused_input='ignore')
	
	attack_update = [(r_S, (r_S - (rate*gradient)).astype(theano.config.floatX))]
	attack_f = theano.function([class_conf, delta, rate], outputs=normed, updates=attack_update, givens={lX: T.min(r_S+original_S), hX: T.max(r_S+original_S), r_V: r_S}, on_unused_input='ignore')	

	# random_update = [(r_S, r_S + np.random.randint(-10, 10, height*width))]
	# random_f = theano.function([], outputs=r_S, updates=random_update)

	stilln = 0
	reg_epoch = 1000	
	cur_reg_epoch = 0
	last_good = None
	last_val = None		
	for iter in range(0, attack_patience):		

		#val = attack_f(r_S.get_value().reshape(col_width,).astype(theano.config.floatX), p_confidence, reg)
		val = attack_f(p_confidence, reg, attack_rate)
		r = r_S.get_value().reshape(col_width,)

		if last_val is not None:
			if abs(val - last_val) <= tolerance:
				break
		last_val = val		

		attack_image = np.clip(r+orig, 0, 255).reshape(X_shape).astype(np.uint8) # normalize(r+orig, 0, 255, dtype=np.uint8).reshape(X_shape)
		(cur_predicted, cur_conf) = model.predict(np.asarray(attack_image))		

		if cur_predicted == target:
			last_good = attack_image
			if cur_reg_epoch == 0:
				reg = min(2*reg, 131072)
				attack_rate = max(attack_rate/1.5, 0.001)
				cur_reg_epoch = reg_epoch
				# random_f()
				print "increased regularization constant to", reg, ", rate to", attack_rate
			cur_reg_epoch -= 1
		elif cur_predicted != target and reg > 1:
			reg = max(1, reg/1.1)
			attack_rate = min(2*attack_rate, 65536) if last_good is not None else attack_rate
			cur_reg_epoch = reg_epoch
			print "backing off regularization constant; new value =", reg, ", rate =", attack_rate

		status_image = None
		if iter % 50 == 0:
			print "obj=", val			
			cv2.destroyAllWindows()
			status_image = make_status_image(attacker, attack_image, victim, X_shape)
			cv2.imshow("iter", status_image)

		if iter % 100 == 0 and make_video:
			cv2.imwrite("stills/im%d.jpg" % stilln, status_image)
			stilln += 1
	
	opt_result = r_S.get_value().reshape(col_width,)

	if last_good is None:
		new_image = normalize(X_t[idx] + opt_result.reshape(X_t[idx].shape), 0, 255, dtype=np.uint8)
	else:
		new_image = last_good

	if make_video:
		cv2.imwrite("stills/im%d.jpg" % stilln, new_image)			

	print "*** After attack ***"
	[p_label, p_confidence] = model.predict(np.asarray(new_image))
	print "Attacker classified as = %d (confidence=%.2f, real label = %d, %s)" % (p_label, p_confidence, y_t[idx], "SUCCESS" if p_label == target else "FAILURE")

	cv2.destroyAllWindows()

	return (make_status_image(attacker, attack_image, victim, X_shape), p_label == target)

def attack_eigenfaces2(X, y, X_t, y_t, model, idx, target, reg, attack_rate=1, tolerance=0.01, attack_patience=10000, make_video=False):
	
	[p_label, p_confidence] = model.predict(np.asarray(X_t[idx]))	
	print "Attacker's predicted label = %d (confidence=%.2f, real label = %d)" % (p_label, p_confidence, y_t[idx])

	X_shape = X_t[idx].shape
	(height, width) = X_shape
	nent = height*width		
	eigenvectors = model.getMat("eigenvectors")
	eigsT = eigenvectors.T
	training_projections = model.getMatVector("projections")
	mean = model.getMat("mean")
	orig = X_t[idx].reshape(mean.shape)	
	
	target_projections = []	
	for i in range(0, len(training_projections)):		
		if y[i] == target:			
			target_projections.append(training_projections[i])


	(w_nearest, idx_nearest, _) = get_nearest(target_projections, get_weights(orig, eigenvectors, mean))
	init = np.ones_like(w_nearest, dtype=theano.config.floatX)
	_, col_width = w_nearest.shape	

	delta = T.scalar('delta', dtype=theano.config.floatX)
	lX = T.scalar('minX', dtype=theano.config.floatX)
	hX = T.scalar('maxX', dtype=theano.config.floatX)
	class_conf = T.scalar('class_conf', dtype=theano.config.floatX)
	eig_row, eig_col = eigenvectors.shape
	
	r_V = T.vector('r')	
		
	original_S = theano.shared(value=orig, name='original', borrow=True)
	mean_S = theano.shared(value=mean, name='mean', borrow=True)
	eigens_S = theano.shared(value=eigenvectors, name='eigenvectors', borrow=True)
	eigensT_S = theano.shared(value=eigsT, name='eigensT', borrow=True)
	w_S = theano.shared(value=w_nearest, name='w', borrow=True)
	
	recon_image = r_V.dot(eigensT_S)
	X = recon_image + original_S
	X_normalized = T.clip(X, 0, 255) # ((X - lX)/(hX - lX))*255
	sub_n = X_normalized - mean_S
	weights_n = sub_n.dot(eigens_S)
	normed = T.max([T.sum((weights_n - w_S)**2) - class_conf*class_conf, 0]) + delta*T.mean(recon_image**2)#+ reg*T.mean(r_V**2)
	gradient = T.grad(cost=normed, wrt=r_V)

	r_S = theano.shared(value=np.zeros_like(w_nearest, dtype=theano.config.floatX), name='r_actual', borrow=True)
	#r_S = theano.shared(value=np.random.sample(col_width)*256, name='r_actual', borrow=True)

	cost_f = theano.function([r_V, class_conf, delta], outputs=normed, on_unused_input='ignore')
	grad_f = theano.function([r_V, class_conf, delta], outputs=gradient, on_unused_input='ignore')

	attack_update = [(r_S, r_S - attack_rate*gradient)]
	attack_f = theano.function([r_V, class_conf, delta], outputs=normed, updates=attack_update)	

	stilln = 0
	reg_epoch = 5000
	cur_reg_epoch = 0
	last_good = None	
	last_val = None
	for iter in range(0, attack_patience):		

		val = attack_f(r_S.get_value().reshape(col_width,), p_confidence, reg)
		r = r_S.get_value().reshape(col_width,)

		if last_val is not None:
			if abs(val - last_val) <= tolerance:
				break
		last_val = val		

		attack_image = np.clip(r.dot(eigsT)+orig, 0, 255).astype(np.uint8).reshape(X_shape)
		(cur_predicted, cur_conf) = model.predict(np.asarray(attack_image))

		if cur_predicted == target:
			last_good = attack_image
			if cur_reg_epoch == 0:
				reg *= 2
				cur_reg_epoch = reg_epoch
				print "increased regularization constant to", reg
			cur_reg_epoch -= 1
		elif cur_predicted != target and reg > 1:
			reg = max(1, reg/1.1)
			cur_reg_epoch = reg_epoch
			print "backing off regularization constant; new value =", reg

		if iter % 500 == 0:
			print "obj=", val
			cv2.destroyAllWindows()
			cv2.imshow("iter=%d" % iter, attack_image)

		if iter % 20 == 0 and make_video:
			cv2.imwrite("stills/im%d.jpg" % stilln, attack_image)			
			stilln += 1		

	opt_result = r_S.get_value().reshape(col_width,)
	print '-'*60
	print opt_result
	print '-'*60
	print '\n'	

	if last_good is None:
		new_image = np.clip(opt_result.dot(eigsT)+orig, 0, 255).astype(np.uint8).reshape(X_shape)
	else:
		new_image = last_good

	if make_video:
		cv2.imwrite("stills/im%d.jpg" % stilln, new_image)					

	print "*** After attack ***"
	[p_label, p_confidence] = model.predict(np.asarray(new_image))
	print "Attacker classified as = %d (confidence=%.2f, real label = %d, %s)" % (p_label, p_confidence, y_t[idx], "SUCCESS" if p_label == target else "FAILURE")

	cv2.destroyAllWindows()

	return (new_image, p_label == target)

def attack_eigenfaces3(X, y, X_t, y_t, model, idx, target, reg, attack_rate=1, tolerance=0.01, attack_patience=10000, make_video=False):
	
	[p_label, p_confidence] = model.predict(np.asarray(X_t[idx]))	
	print "Attacker's predicted label = %d (confidence=%.2f, real label = %d)" % (p_label, p_confidence, y_t[idx])

	X_shape = X_t[idx].shape
	(height, width) = X_shape
	col_width = height*width
	nent = height*width		
	eigenvectors = model.getMat("eigenvectors")
	training_projections = model.getMatVector("projections")
	mean = model.getMat("mean")
	orig = X_t[idx].reshape(mean.shape)
	init = np.zeros_like(orig, dtype=theano.config.floatX)+255
	
	target_projections = []	
	for i in range(0, len(training_projections)):		
		if y[i] == target:			
			target_projections.append(training_projections[i])


	(w_nearest, idx_nearest, _) = get_nearest(target_projections, get_weights(orig, eigenvectors, mean))

	rate = T.scalar('rate', dtype=theano.config.floatX)
	delta = T.scalar('delta', dtype=theano.config.floatX)
	lX = T.scalar('minX', dtype=theano.config.floatX)
	hX = T.scalar('maxX', dtype=theano.config.floatX)
	class_conf = T.scalar('class_conf', dtype=theano.config.floatX)
	eig_row, eig_col = eigenvectors.shape

	r_V = T.vector('r')	
		
	#r_S = theano.shared(value=np.zeros_like(orig, dtype=theano.config.floatX).reshape(height*width,), name='r_actual', borrow=False)
	r_S = theano.shared(value=np.random.sample(height*width).astype(theano.config.floatX)*256, name='r_actual', borrow=False)

	proj_mean = np.zeros_like(w_nearest).astype(theano.config.floatX)
	for tp in training_projections:
		proj_mean = proj_mean + tp/len(training_projections)

	original_S = theano.shared(value=orig.astype(theano.config.floatX), name='original', borrow=False)
	mean_S = theano.shared(value=mean.astype(theano.config.floatX), name='mean', borrow=False)
	eigens_S = theano.shared(value=eigenvectors.astype(theano.config.floatX), name='eigenvectors', borrow=False)
	w_S = theano.shared(value=proj_mean.astype(theano.config.floatX), name='w', borrow=False)
	
	X = r_V + original_S
	X_normalized = T.clip(X, 0, 255) # ((X - lX)/(hX - lX))*255
	sub_n = X_normalized - mean_S
	weights_n = sub_n.dot(eigens_S)
	normed = T.max([T.sum((weights_n - T.mean(w_S))**2) - class_conf*class_conf, 0]) + delta*T.mean(r_V**2)
	#normed = T.sum((weights_n - w_S)**2) - class_conf**2 + reg*T.mean(r_V**2)
	gradient = T.grad(cost=normed, wrt=r_V)
	
	cost_f = theano.function([class_conf, delta], outputs=normed, givens={lX: T.min(r_S+original_S), hX: T.max(r_S+original_S), r_V: r_S}, on_unused_input='ignore')
	grad_f = theano.function([class_conf, delta], outputs=gradient, givens={lX: T.min(r_S+original_S), hX: T.max(r_S+original_S), r_V: r_S}, on_unused_input='ignore')
	
	attack_update = [(r_S, r_S - (rate*gradient).astype(theano.config.floatX))]
	attack_f = theano.function([class_conf, delta, rate], outputs=normed, updates=attack_update, givens={lX: T.min(r_S+original_S), hX: T.max(r_S+original_S), r_V: r_S}, on_unused_input='ignore')	

	# random_update = [(r_S, r_S + np.random.randint(-10, 10, height*width))]
	# random_f = theano.function([], outputs=r_S, updates=random_update)

	stilln = 0
	reg_epoch = 5000	
	cur_reg_epoch = 0
	last_good = None
	last_val = None		
	for iter in range(0, attack_patience):		

		#val = attack_f(r_S.get_value().reshape(col_width,).astype(theano.config.floatX), p_confidence, reg)
		val = attack_f(p_confidence, reg, attack_rate)
		r = r_S.get_value().reshape(col_width,)

		# if last_val is not None:
		# 	if abs(val - last_val) <= tolerance:
		# 		break
		last_val = val		

		attack_image = np.clip(r+orig, 0, 255).reshape(X_shape).astype(np.uint8) # normalize(r+orig, 0, 255, dtype=np.uint8).reshape(X_shape)
		(cur_predicted, cur_conf) = model.predict(np.asarray(attack_image))		

		if cur_predicted == target:
			last_good = attack_image
			if cur_reg_epoch == 0:
				reg = min(2*reg, 131072)
				attack_rate = max(attack_rate/1.5, 0.001)
				cur_reg_epoch = reg_epoch
				# random_f()
				print "increased regularization constant to", reg, ", rate to", attack_rate
			cur_reg_epoch -= 1
		elif cur_predicted != target and reg > 1:
			reg = max(1, reg/1.1)
			attack_rate = min(2*attack_rate, 5) if last_good is not None else attack_rate
			cur_reg_epoch = reg_epoch
			print "backing off regularization constant; new value =", reg, ", rate =", attack_rate

		if iter % 500 == 0:
			print "obj=", val, "cur_predicted=", cur_predicted, "actual=", y_t[idx]			
			cv2.destroyAllWindows()
			cv2.imshow("iter", attack_image)

		if iter % 100 == 0 and make_video:
			cv2.imwrite("stills/im%d.jpg" % stilln, attack_image)			
			stilln += 1
	
	opt_result = r_S.get_value().reshape(col_width,)

	if last_good is None:
		new_image = normalize(X_t[idx] + opt_result.reshape(X_t[idx].shape), 0, 255, dtype=np.uint8)
	else:
		new_image = last_good

	if make_video:
		cv2.imwrite("stills/im%d.jpg" % stilln, new_image)			

	print "*** After attack ***"
	[p_label, p_confidence] = model.predict(np.asarray(new_image))
	print "Attacker classified as = %d (confidence=%.2f, real label = %d, %s)" % (p_label, p_confidence, y_t[idx], "SUCCESS" if p_label == target else "FAILURE")

	cv2.destroyAllWindows()

	return (new_image, p_label == target)


if __name__ == "__main__":

	make_video = True
	cv2.startWindowThread()

	if make_video:
		if os.path.exists('./stills'):
			shutil.rmtree('./stills')
			os.makedirs('./stills')

	if len(sys.argv) < 2:
		print "USAGE: opencv.py </path/to/images>"
		sys.exit()

	[X,y] = read_images(sys.argv[1], splitsize=5, splitstart=0, nclasses=10)
	[X_t, y_t] = read_images(sys.argv[1], splitsize=5, splitstart=5, nclasses=10)
	y = np.asarray(y, dtype=np.int32)
	y_t = np.asarray(y_t, dtype=np.int32)

	model = cv2.createEigenFaceRecognizer()	
	model.train(np.asarray(X), np.asarray(y))

	idx = 0
	target = 1
	reg = 16
	attack_image, success = attack_eigenfaces(X, y, X_t, y_t, model, idx=idx, target=target, reg=reg, tolerance=0.00001, attack_patience=100000, attack_rate=2, make_video=make_video)

	cv2.imshow("attacker", X_t[idx])	
	cv2.imshow("victim", X[7*target])
	cv2.imshow("attack face", attack_image)

	cv2.imwrite("attack_image.png", attack_image)

	if make_video:
		if os.path.exists('attack.avi'):
			os.remove('attack.avi')		
		subprocess.call(['ffmpeg', '-start_number', '0', '-i', './stills/im%d.jpg', '-vcodec', 'mpeg4', 'attack.avi'])

	cv2.waitKey(0)
