import sys
import os
import argparse
import gzip
import math
import random
import subprocess as sp
import time
import webcam
import projector

import pylearn2
import theano
import scipy
import numpy as np
import cv2
import cPickle

from pylearn2.config import yaml_parse
from pylearn2.utils import serial
from pylearn2.utils import safe_zip
from pylearn2.utils.data_specs import DataSpecsMapping

from PIL import Image

def load_mnist(location="../images/mnist.pkl.gz"):

	f = gzip.open(location, 'rb')
	(train_set_x, train_set_y), (valid_set_x, valid_set_y), (test_set_x, test_set_y) = cPickle.load(f)
	f.close()	

	return (train_set_x, train_set_y), (valid_set_x, valid_set_y), (test_set_x, test_set_y)

def normalize_relative(X, Y, low, high, dtype=theano.config.floatX):
	"""Normalizes a given array in X to a value between low and high."""
	X = np.asarray(X)
	# minX, maxX = max(np.min(X), 1), max(np.max(X), 1)

	Xm = X.flatten() if len(X.shape) > 1 else X[:]
	Ym = Y.flatten() if len(Y.shape) > 1 else Y[:]
	C = np.concatenate((Xm, Ym))

	minX, maxX = np.min(C), np.max(C)
	# print minX, ",", maxX
	# normalize to [0...1].
	X = X - float(minX)
	X = X / float((maxX - minX))
	# scale to [low...high].
	X = X * (high-low)
	X = X + low
	if dtype is None:
		return np.asarray(X)
	return np.asarray(X, dtype=dtype)

def normalize(X, low, high, dtype=theano.config.floatX):
	"""Normalizes a given array in X to a value between low and high."""
	X = np.asarray(X)
	# minX, maxX = max(np.min(X), 1), max(np.max(X), 1)
	minX, maxX = np.min(X), np.max(X)
	# print minX, ",", maxX
	# normalize to [0...1].
	X = X - float(minX)
	X = X / float((maxX - minX))
	# scale to [low...high].
	X = X * (high-low)
	X = X + low
	if dtype is None:
		return np.asarray(X)
	return np.asarray(X, dtype=dtype)

def format_image_to_model(model, image, grayscale=True, target_shape=None):	

	cost = model.get_default_cost()
	data_specs = cost.get_data_specs(model)
	mapping = DataSpecsMapping(data_specs)

	space_tuple = mapping.flatten(data_specs[0], return_tuple=True)
	source_tuple = mapping.flatten(data_specs[1], return_tuple=True)

	assert source_tuple.index("features") >= 0
	feature_space = space_tuple[source_tuple.index("features")]

	if isinstance(feature_space, pylearn2.space.VectorSpace):

		dim = feature_space.dim

		if grayscale and len(image.shape) > 2:
			image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
			nchannels = 1
			height, width = image.shape
		else:
			if len(image.shape) > 2:
				height, width, nchannels = image.shape
			else:
				nchannels = 1
				height, width = image.shape

		if height*width != dim and target_shape is None:		

			# we need to find a new aspect ratio that fits in this vector space,
			# but doesn't distort the original image too much.
			# current solution is to check factors of vector space dimension,
			# find the best fit.
			aspect = float(height)/float(width)
			closest_aspect = float("inf")
			best_ratio = (None, None)
			for i in range(1, int((dim/nchannels)**0.5)+1):
				new_aspect = abs(float(i)/(float(dim)/float(i)/float(nchannels))-aspect)
				if (dim/nchannels) % i == 0 and new_aspect < closest_aspect:
					closest_aspect = new_aspect
					best_ratio = (int(i), int(float(dim)/float(i)/float(nchannels)), nchannels)

		else:

			if target_shape is None:
				best_ratio = (height, width, nchannels)
			else:
				if len(target_shape) == 2:
					t_height, t_width = target_shape
					best_ratio = (t_height, t_width, 1)
				else:
					best_ratio = (target_shape[0], target_shape[1], 1 if grayscale else target_shape[2])					

		rat_height, rat_width, _ = best_ratio		
		image = cv2.resize(image, (rat_width, rat_height), interpolation=cv2.INTER_NEAREST)
		image = normalize(np.asarray(image, dtype=theano.config.floatX).reshape((1, dim)), 0, 1)

		return image, (height, width, nchannels), best_ratio

	elif isinstance(feature_space, pylearn2.space.Conv2DSpace):

		shape = feature_space.shape
		feat_height, feat_width = shape		

		if grayscale and len(image.shape) > 2:
			image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
			nchannels = 1
			height, width = image.shape
		else:
			if len(image.shape) > 2:
				height, width, nchannels = image.shape
			else:
				nchannels = 1
				height, width = image.shape
		
		resize_shape = feat_width, feat_height

		image = normalize(cv2.resize(image, resize_shape, interpolation=cv2.INTER_NEAREST), 0, 1)
		image = image.reshape([1, feat_height, feat_width, nchannels])

		return image, (height, width, nchannels), (feat_height, feat_width, nchannels)

	else:
		assert False and "unsupported feature space"

def onehot_label(model, label):

	cost = model.get_default_cost()
	data_specs = cost.get_data_specs(model)
	mapping = DataSpecsMapping(data_specs)

	space_tuple = mapping.flatten(data_specs[0], return_tuple=True)
	source_tuple = mapping.flatten(data_specs[1], return_tuple=True)

	assert source_tuple.index("targets") >= 0
	target_space = space_tuple[source_tuple.index("targets")]

	assert isinstance(target_space, pylearn2.space.VectorSpace)
	dim = target_space.dim

	onehot = np.asarray([0]*dim, dtype=theano.config.floatX).reshape(1, dim)	
	onehot[0,label] = 1.0	

	return onehot

def get_model_params(model, attacker_image, target_label):
	"""
	Returns the attack parameter (i.e., noise vector variable), the model cost expression
	(specialized with the attacker_image and target_label), and the gradient with respect to 
	the attack parameter.	

	Assumes that the model outcome (and therefore the target_label) is in one-hot format,
	and that the attack parameter has the same shape as attacker_image.
	"""

	cost = model.get_default_cost()
	data_specs = cost.get_data_specs(model)
	mapping = DataSpecsMapping(data_specs)
	space_tuple = mapping.flatten(data_specs[0], return_tuple=True)
	source_tuple = mapping.flatten(data_specs[1], return_tuple=True)

	r = None
	features_var = None
	targets_var = None

	theano_args = []
	for space, source in safe_zip(space_tuple, source_tuple):	
		name = 'model_[%s]' % source
		arg = space.make_theano_batch(name=name)
		if source == "features":
			features_var = arg			
			r = space.make_theano_batch(name='r')
			o = space.make_theano_batch(name='o')
		elif source == "targets":
			targets_var = arg
		theano_args.append(arg)
	theano_args = tuple(theano_args)

	nested_args = mapping.nest(theano_args)
	fixed_var_descr = cost.get_fixed_var_descr(model, nested_args)

	assert r is not None and o is not None and features_var is not None and targets_var is not None	

	reg = theano.tensor.scalar(name='reg', dtype=theano.config.floatX)

	expr = cost.expr(model, nested_args, ** fixed_var_descr.fixed_vars)
	spec_expr = theano.clone(expr, replace={features_var: theano.tensor.clip(r+o, 0, 1), targets_var: target_label})
	cost_expr = spec_expr + reg*theano.tensor.mean(theano.tensor.sqr(r))
	# spec_grad = theano.tensor.grad(cost=theano.clone(cost_expr, replace={o: attacker_image}), wrt=r)
	spec_grad = theano.tensor.grad(cost=cost_expr, wrt=r)

	return (cost_expr, spec_grad, o, r, reg)

def make_status_image(attacker_raw, mask, victim, shape, scale=1, color=False, add_mask=True):

	height, width, nchannels = shape	

	if victim is None:
		victim = np.zeros(shape, dtype=theano.config.floatX)

	if add_mask:
		attacker_noised = normalize(np.clip(attacker_raw+mask, 0, 1), 0, 255, dtype=np.uint8).reshape((height, width, nchannels))
	else:
		attacker_noised = normalize(np.clip(attacker_raw, 0, 1), 0, 255, dtype=np.uint8).reshape((height, width, nchannels))
	attacker_raw = normalize(attacker_raw, 0, 255, dtype=np.uint8).reshape((height, width, nchannels))
	mask = normalize(mask, 0, 255, dtype=np.uint8).reshape((height, width, nchannels))
	victim = normalize(victim, 0, 255, dtype=np.uint8).reshape((height, width, nchannels))

	combined = np.concatenate((attacker_raw, mask, attacker_noised, victim), axis=1)

	return cv2.resize(combined, (scale*4*width, scale*height), interpolation=cv2.INTER_NEAREST)

def show_status(current_cost, r_value, attacker_image, victim_image, attack_shape, scale, predicted, 
	show_image=True, color=False, time_elapsed=None, print_status=True, grad_norm=None, tolerance=None,
	add_mask=True):

	assert grad_norm is not None
	
	if print_status:
		if time_elapsed is None:
			print "obj=%.4f, class=%d, mean=%.4f, t=NA, step=%.5f > %.7f" % (current_cost, predicted, np.mean(abs(r_value)), grad_norm, tolerance)
		else:
			print "obj=%.4f, class=%d, mean=%.4f, t=%.5f, step=%.5f > %.7f" % (current_cost, predicted, np.mean(abs(r_value)), time_elapsed, grad_norm, tolerance)
	
	status_image = make_status_image(attacker_image, r_value, victim_image, attack_shape, scale=scale, color=color, add_mask=add_mask)
	
	if show_image:	
		# cv2.destroyAllWindows()
		cv2.imshow("status", status_image)
		cv2.waitKey(1)

def open_video(filename, shape, status_scale):

	height, width, nchannels = shape

	pix_fmt = 'gray' if nchannels == 1 else 'bgr24'

	FFMPEG_BIN = "ffmpeg"

	command = [ FFMPEG_BIN,
			'-y',
			'-s',
			'%dx%d' % (status_scale*4*width, status_scale*height),
			'-r', '24',
			# '-vcodec', 'mpeg4',
			'-c:v', 'rawvideo',
			'-f', 'rawvideo',
			'-pix_fmt', pix_fmt,
			'-i',
			'-',
			filename ]
	with open('/dev/null', 'w') as devnull:
		pipe = sp.Popen(command, stdin=sp.PIPE, stdout=devnull, stderr=devnull)	

	return pipe

def skin_mask(image):

	image = normalize(image, 0, 255, np.uint8)

	# Constants for finding range of skin color in YCrCb
	min_YCrCb = np.array([0,133,77],np.uint8)
	max_YCrCb = np.array([255,173,127],np.uint8)	

	imageYCrCb = cv2.cvtColor(image, cv2.COLOR_BGR2YCR_CB)
	skinRegion = cv2.inRange(imageYCrCb,min_YCrCb,max_YCrCb)

	_, _, nchannels = image.shape
	if nchannels > 1:
		skinRegion = cv2.cvtColor(skinRegion, cv2.COLOR_GRAY2BGR)

	skinRegion = normalize(skinRegion, 0, 1)

	return skinRegion

def run_attack(model, attacker_image, target_label, victim_image=None, attack_patience=50000, attack_rate=10, attack_reg=1, 
				status_period=50, reg_epoch=500, reg_scaling=True, color=False, random_start=False, blur=True, out_image=None, 
				status_scale=1.0, status_images=True, gauss_kernel_size=5, make_video=False, video_interval=100, video_file=None,
				target_shape=None, skin=False, tolerance=1e-2, tolerance_scaling=True, steptol_terminate=True, success_terminate=False):

	try:

		assert video_interval > 0 or not make_video

		target_label_int = int(np.where(target_label>0)[1][0])

		cv2.namedWindow("status", cv2.CV_WINDOW_AUTOSIZE)

		attacker_scaled, original_shape, scaled_shape = format_image_to_model(model, attacker_image, grayscale=not color, target_shape=target_shape)
		attacker_flattened = attacker_scaled.flatten()
		if victim_image is not None:
			victim_scaled, _, _ = format_image_to_model(model, victim_image, grayscale=not color, target_shape=target_shape)
		else:
			victim_scaled = None	

		noise_shape = attacker_scaled.shape
		height, width, nchannels = scaled_shape

		if make_video:
			assert video_file is not None
			video_pipe = open_video(video_file, scaled_shape, status_scale)

		if color and skin:
			skin_region = skin_mask(cv2.resize(attacker_image, (width, height))).reshape(noise_shape)

		cost_expr, cost_grad, o, r, reg = get_model_params(model, attacker_scaled, target_label)

		rate = theano.tensor.scalar(name='rate', dtype=theano.config.floatX)
		if not random_start:		
			r_shared = theano.shared(value=np.zeros(noise_shape, dtype=theano.config.floatX), name="r_S", borrow=True)
		else:
			r_shared = theano.shared(value=np.random.sample(height*width).astype(theano.config.floatX).reshape(noise_shape), name="r_S", borrow=True)
		
		attack_updates = [(r_shared, r_shared-rate*cost_grad)]
		attack_step = theano.function([reg, rate, r], outputs=cost_expr, updates=attack_updates, givens={o: attacker_scaled})

		grad_norm = theano.function([reg, r, o], outputs=theano.tensor.sqrt(theano.tensor.sum(theano.tensor.sqr(cost_grad))))

		inputvar = model.get_input_space().make_theano_batch()
		predict_expr = model.fprop(inputvar)
		predict = theano.function([inputvar], theano.tensor.argmax(predict_expr, axis=1)[0])

		initial_label = predict(attacker_scaled)
		print "... initial label is %d" % initial_label

		current_cost = float("inf")
		current_epoch = 0
		target_label = np.argmax(target_label)
		epoch_time = None

		for i in range(0, attack_patience):

			epoch_start = time.time()

			r_value = r_shared.get_value()
			if color and skin:				
				r_value = r_value*skin_region			
			if blur:
				r_value = cv2.GaussianBlur(r_value.reshape(scaled_shape), (gauss_kernel_size,gauss_kernel_size), 0).reshape(noise_shape)
				r_shared.set_value(r_value)

			current_predicted = predict(np.clip(r_value+attacker_scaled, 0, 1))

			if current_predicted == target_label_int and success_terminate:
				break

			if reg_scaling:
				if current_predicted == target_label:
					current_epoch -= 1			
					if current_epoch <= 0:
						attack_reg = min(2*attack_reg, 1024)						
						current_epoch = reg_epoch
						tolerance = tolerance*2.0 if tolerance_scaling else tolerance
				else:
					attack_reg = max(attack_reg/1.1, 1)
					tolerance = tolerance/1.1 if tolerance_scaling else tolerance

			if i % status_period == 0:			
				show_status(current_cost, r_value, attacker_scaled, victim_scaled, scaled_shape, scale=status_scale, 
					predicted=current_predicted, show_image=status_images, color=color, time_elapsed=epoch_time,
					grad_norm=grad_norm(attack_reg, r_value, attacker_scaled), tolerance=tolerance)				

			if i % video_interval == 0 and make_video:
				video_frame = make_status_image(attacker_scaled, r_value, victim_scaled, scaled_shape, status_scale)
				video_pipe.stdin.write(video_frame.tostring())
			
			current_cost = attack_step(attack_reg, attack_rate/math.log(i+2), r_value)

			if grad_norm(attack_reg, r_value, attacker_scaled) < tolerance and current_predicted == target_label_int and steptol_terminate:
				break;

			epoch_end = time.time()
			epoch_time = epoch_end - epoch_start

		print "... final label is %d" % predict(attacker_scaled + r_shared.get_value())
		
		if out_image is not None:
			status_image = make_status_image(attacker_scaled, r_value, victim_scaled, scaled_shape, scale=2, color=color)
			cv2.imwrite(out_image, status_image)

		cv2.waitKey(0)

	finally:

		if make_video:
			video_pipe.terminate()

def run_attack_webcam(model, attacker_image, target_label, victim_image=None, attack_patience=50000, attack_rate=10, attack_reg=1, 
						status_period=50, reg_epoch=500, reg_scaling=True, color=False, random_start=False, blur=True, out_image=None, 
						status_scale=1.0, status_images=True, gauss_kernel_size=5, make_video=False, video_interval=100, video_file=None,
						target_shape=None, skin=False, tolerance=1e-2, tolerance_scaling=True, steptol_terminate=True, success_terminate=False):

	try:

		assert video_interval > 0 or not make_video

		video_pipe = None

		webcam_face = None
		cv2.namedWindow("webcam", cv2.CV_WINDOW_AUTOSIZE)

		frame_buffer = []
		frame_buffer_nbytes = 0

		print "\n\n"
		print "*"*80
		print "* !!! beginning webcam initialization.\n*" 
		print "* Put keyboard focus on the video (i.e., click on the webcam window)\n* Press q when your face is centered, and the frame has resized\n* to fit your face."
		print "*"*80
		print "\n"

		webcam_face = webcam.WebcamFace(grayscale=not color)		
		webcam_face.begin_capture()
		init_frame = None
		try:
			while True:
				frame = webcam_face.get_frame()
				cv2.imshow("webcam", frame)
				c = cv2.waitKey(10)
				if c == 113:
					init_frame = frame
					break
		finally:
			webcam_face.end_capture()
		
		target_label_int = int(np.where(target_label>0)[1][0])

		cv2.destroyWindow("webcam")
		cv2.namedWindow("status", cv2.CV_WINDOW_AUTOSIZE)
		print "\n\n!!! Now hold still while the attack works."
		print "!!! When the status indicates class=%s, the attack has succeeded.\n\n" % target_label_int		

		attacker_image = init_frame

		attacker_scaled, original_shape, scaled_shape = format_image_to_model(model, attacker_image, grayscale=not color, target_shape=target_shape)		
		attacker_flattened = attacker_scaled.flatten()
		if victim_image is not None:
			victim_scaled, _, _ = format_image_to_model(model, victim_image, grayscale=not color, target_shape=target_shape)
		else:
			victim_scaled = None	

		noise_shape = attacker_scaled.shape
		height, width, nchannels = scaled_shape
		orig_height, orig_width, _ = original_shape

		webcam_face.set_size((width, height))		

		cost_expr, cost_grad, o, r, reg = get_model_params(model, attacker_scaled, target_label)

		rate = theano.tensor.scalar(name='rate', dtype=theano.config.floatX)
		if not random_start:		
			r_shared = theano.shared(value=np.zeros(noise_shape, dtype=theano.config.floatX), name="r_S", borrow=True)
		else:
			r_shared = theano.shared(value=np.random.sample(height*width).astype(theano.config.floatX).reshape(noise_shape), name="r_S", borrow=True)

		attack_updates = [(r_shared, r_shared-rate*cost_grad)]
		attack_step = theano.function([reg, rate, r, o], outputs=cost_expr, updates=attack_updates)

		grad_norm = theano.function([reg, r, o], outputs=theano.tensor.sqrt(theano.tensor.sum(theano.tensor.sqr(cost_grad))))

		inputvar = model.get_input_space().make_theano_batch()
		predict_expr = model.fprop(inputvar)
		predict = theano.function([inputvar], theano.tensor.argmax(predict_expr, axis=1)[0])
		class_prob = theano.function([inputvar], 100*predict_expr[0][target_label_int])

		initial_label = predict(attacker_scaled)
		print "... initial label is %d" % initial_label

		webcam_face.begin_capture()

		current_cost = float("inf")
		current_epoch = 0
		target_label = np.argmax(target_label)
		epoch_time = None
		last_failure = 0
		skin_region = np.ones_like(attacker_scaled)

		for i in range(0, attack_patience):

			epoch_start = time.time()

			frame = webcam_face.get_frame(doKill=True)
			attacker_scaled = frame.reshape(noise_shape)

			r_value = r_shared.get_value()
			if color and skin:
				skin_region = skin_mask(frame).reshape(noise_shape)
				r_value = r_value*skin_region
			if blur:
				r_value = cv2.GaussianBlur(r_value.reshape(scaled_shape), (gauss_kernel_size,gauss_kernel_size), 0).reshape(noise_shape)
				r_shared.set_value(r_value)

			current_predicted = predict(np.clip(r_value+attacker_scaled, 0, 1))

			if current_predicted == target_label_int and success_terminate:
				break			

			if reg_scaling:
				if current_predicted == target_label:
					current_epoch -= 1			
					if current_epoch <= 0:
						attack_reg = min(2*attack_reg, 1024)
						tolerance = tolerance*2.0 if tolerance_scaling else tolerance
						current_epoch = reg_epoch				
				else:
					attack_reg = max(attack_reg/1.1, 1)
					tolerance = tolerance/1.1 if tolerance_scaling else tolerance

			if current_predicted != target_label:
				last_failure = i
			
			show_status(current_cost, r_value, frame.reshape(noise_shape), victim_scaled, scaled_shape, scale=status_scale, 
				predicted=current_predicted, show_image=status_images, color=color, time_elapsed=epoch_time, print_status=(i % status_period == 0),
				grad_norm=grad_norm(attack_reg, r_value, attacker_scaled), tolerance=tolerance)

			# if i % status_period == 0:
			# 	print class_prob(np.clip(r_value+attacker_scaled, 0, 1))
			# 	print "%f megabytes in frame buffer" % frame_buffer_nbytes

			if i % video_interval == 0 and make_video:
				video_frame = make_status_image(attacker_scaled, r_value, victim_scaled, scaled_shape, status_scale)
				frame_buffer.append(video_frame)
				frame_buffer_nbytes += float(video_frame.nbytes)/1048576.0
			
			current_cost = attack_step(attack_reg, attack_rate/math.log((i-last_failure)+2), r_value, attacker_scaled)

			if grad_norm(attack_reg, r_value, attacker_scaled) < tolerance and current_predicted == target_label_int and steptol_terminate:
				break;

			epoch_end = time.time()
			epoch_time = epoch_end - epoch_start

		print "... final label is %d" % predict(attacker_scaled + r_shared.get_value())
		
		if out_image is not None:
			status_image = make_status_image(attacker_scaled, r_value, victim_scaled, scaled_shape, scale=2, color=color)
			cv2.imwrite(out_image, status_image)

		if make_video:
			assert video_file is not None
			video_pipe = open_video(video_file, scaled_shape, status_scale)
			cnt = 0
			print "\n\n!!! writing output to %s" % video_file
			print "!!! %d frames" % len(frame_buffer)
			for frame in frame_buffer:
				video_pipe.stdin.write(frame.tostring())
				print "%d," % cnt,
				cnt += 1
			print "\n\n"

		cv2.waitKey(0)

	finally:

		if make_video and video_pipe is not None:
			video_pipe.terminate()
		if webcam_face is not None:
			webcam_face.end_capture()

def run_attack_projector(model, attacker_image, target_label, victim_image=None, attack_patience=50000, attack_rate=10, attack_reg=1, 
						status_period=50, reg_epoch=500, reg_scaling=True, color=False, random_start=False, blur=True, out_image=None, 
						status_scale=1.0, status_images=True, gauss_kernel_size=5, make_video=False, video_interval=100, video_file=None,
						target_shape=None, skin=False, tolerance=1e-2, tolerance_scaling=True, steptol_terminate=True, success_terminate=False,
						success_stop_cnt=300, success_cond=95.0):

	try:

		assert video_interval > 0 or not make_video

		webcam_face = None
		video_pipe = None
		cv2.namedWindow("webcam", cv2.CV_WINDOW_AUTOSIZE)		

		proj = projector.Projector(screen=1)

		frame_buffer = []
		frame_buffer_nbytes = 0		

		print "\n\n"
		print "*"*80
		print "* !!! beginning webcam initialization.\n*" 
		print "* Put keyboard focus on the video (i.e., click on the webcam window)\n* Press q when your face is centered, and the frame has resized\n* to fit your face."
		print "*"*80
		print "\n"

		webcam_face = webcam.WebcamFace(grayscale=not color)		
		webcam_face.begin_capture()
		init_frame = None
		try:
			while True:
				frame = webcam_face.get_frame()
				cv2.imshow("webcam", frame)
				c = cv2.waitKey(1)
				if c == 113:
					init_frame = frame
					break
		finally:
			webcam_face.end_capture()
		
		target_label_int = int(np.where(target_label>0)[1][0])

		cv2.destroyWindow("webcam")
		cv2.namedWindow("status", cv2.CV_WINDOW_AUTOSIZE)
		print "\n\n!!! Now hold still while the attack works."
		print "!!! When the status indicates class=%s, the attack has succeeded.\n\n" % target_label_int		

		attacker_image = init_frame

		attacker_scaled, original_shape, scaled_shape = format_image_to_model(model, attacker_image, grayscale=not color, target_shape=target_shape)		
		attacker_flattened = attacker_scaled.flatten()
		if victim_image is not None:
			victim_scaled, _, _ = format_image_to_model(model, victim_image, grayscale=not color, target_shape=target_shape)
		else:
			victim_scaled = None	

		noise_shape = attacker_scaled.shape
		height, width, nchannels = scaled_shape
		orig_height, orig_width, _ = original_shape		

		webcam_face.set_size((width, height))

		# if make_video:
		# 	assert video_file is not None
		# 	video_pipe = open_video(video_file, scaled_shape, status_scale)

		cost_expr, cost_grad, o, r, reg = get_model_params(model, attacker_scaled, target_label)

		rate = theano.tensor.scalar(name='rate', dtype=theano.config.floatX)
		if not random_start:		
			r_shared = theano.shared(value=np.zeros(noise_shape, dtype=theano.config.floatX), name="r_S", borrow=True)
		else:
			r_shared = theano.shared(value=np.random.sample(height*width).astype(theano.config.floatX).reshape(noise_shape), name="r_S", borrow=True)
		zero_mask = np.zeros_like(attacker_scaled)

		proj_scale = 5
		proj_frame = np.zeros((proj_scale*height, proj_scale*width, nchannels), dtype=np.uint8)

		attack_updates = [(r_shared, r_shared-rate*cost_grad)]
		attack_step = theano.function([reg, rate, r, o], outputs=cost_expr, updates=attack_updates)

		grad_norm = theano.function([reg, r, o], outputs=theano.tensor.sqrt(theano.tensor.sum(theano.tensor.sqr(cost_grad))))

		inputvar = model.get_input_space().make_theano_batch()
		predict_expr = model.fprop(inputvar)
		predict = theano.function([inputvar], theano.tensor.argmax(predict_expr, axis=1)[0])
		class_prob = theano.function([inputvar], 100*predict_expr[0][target_label_int])

		initial_label = predict(attacker_scaled)
		print "... initial label is %d" % initial_label

		proj.start_window(proj_frame)
		webcam_face.begin_capture()

		current_cost = float("inf")
		current_epoch = 0
		target_label = np.argmax(target_label)
		epoch_time = None
		last_failure = 0
		skin_region = np.ones_like(attacker_scaled)
		tol_cnt = 0
		success_cnt = 0

		for i in range(0, attack_patience):

			epoch_start = time.time()

			r_value = r_shared.get_value()
			
			if color and skin:
				skin_region = skin_mask(frame).reshape(noise_shape)
				r_value = r_value*skin_region
			if blur:
				r_value = cv2.GaussianBlur(r_value.reshape(scaled_shape), (gauss_kernel_size,gauss_kernel_size), 0).reshape(noise_shape)
				r_shared.set_value(r_value)

			# r_value = normalize_relative(r_value, r_value, 0, 1, theano.config.floatX)

			noise_projection = r_value[:]
			noise_projection = cv2.resize(normalize(r_value.reshape(scaled_shape), 0, 255, np.uint8), (proj_scale*width, proj_scale*height))
			noise_projection = noise_projection
			noise_projection = noise_projection.astype(np.uint8)			
			proj.project_frame(noise_projection)
			zero_mask = r_value
			
			frame = webcam_face.get_frame()
			attacker_scaled = frame.reshape(noise_shape)
			as_mean = np.mean(attacker_scaled)
			attacker_scaled *= 0.05/as_mean	

			current_predicted = predict(attacker_scaled)

			if current_predicted == target_label_int and success_terminate and success_cnt >= success_stop_cnt:
				print "terminated on success"
				break			

			if reg_scaling:
				if current_predicted == target_label:
					current_epoch -= 1			
					if current_epoch <= 0:
						attack_reg = min(2*attack_reg, 1024)
						tolerance = tolerance*2.0 if tolerance_scaling else tolerance
						current_epoch = reg_epoch				
				else:
					attack_reg = max(attack_reg/1.1, 1)
					tolerance = tolerance/1.1 if tolerance_scaling else tolerance

			if current_predicted != target_label:
				last_failure = i
				success_cnt = 0
			else:
				if class_prob(np.clip(attacker_scaled, 0, 1)) >= success_cond:
					success_cnt += 1

			show_status(class_prob(np.clip(attacker_scaled, 0, 1)), r_value, frame.reshape(noise_shape), victim_scaled, scaled_shape, scale=status_scale, 
				predicted=current_predicted, show_image=status_images, color=color, time_elapsed=epoch_time, print_status=(i % status_period == 0),
				grad_norm=grad_norm(attack_reg, r_value, attacker_scaled), tolerance=tolerance, add_mask=False)

			# if i % status_period == 0:
			# 	print class_prob(np.clip(attacker_scaled, 0, 1))

			if i % video_interval == 0 and make_video:
				video_frame = make_status_image(attacker_scaled, r_value, victim_scaled, scaled_shape, status_scale, add_mask=False)
				frame_buffer.append(video_frame)
				frame_buffer_nbytes += float(video_frame.nbytes)/1048576.0

			current_cost = attack_step(attack_reg, attack_rate, attacker_scaled, zero_mask) #, attacker_scaled)
			# current_cost = attack_step(attack_reg, attack_rate, zero_mask, attacker_scaled) # normal

			if grad_norm(attack_reg, zero_mask, attacker_scaled) < tolerance and current_predicted == target_label_int and steptol_terminate: # normal
			# if grad_norm(attack_reg, attacker_scaled, zero_mask) < tolerance and current_predicted == target_label_int and steptol_terminate:
				if tol_cnt > 10:
					print "terminated on steptol (%.7f)" % grad_norm(attack_reg, zero_mask, attacker_scaled)
					break;
				tol_cnt += 1
			else:
				tol_cnt = 0

			epoch_end = time.time()
			epoch_time = epoch_end - epoch_start

		print "... final label is %d" % predict(attacker_scaled)
		
		if out_image is not None:
			status_image = make_status_image(attacker_scaled, r_value, victim_scaled, scaled_shape, scale=2, color=color, add_mask=False)
			cv2.imwrite(out_image, status_image)

		if make_video:
			assert video_file is not None
			video_pipe = open_video(video_file, scaled_shape, status_scale)
			cnt = 0
			print "\n\n!!! writing output to %s" % video_file
			print "!!! %d frames" % len(frame_buffer)
			for frame in frame_buffer:
				video_pipe.stdin.write(frame.tostring())
				print "%d," % cnt,
				cnt += 1
			print "\n\n"

		cv2.waitKey(0)

	finally:

		if make_video and video_pipe is not None:
			video_pipe.terminate()
		if webcam_face is not None:
			webcam_face.end_capture()

def process_args():
	parser = argparse.ArgumentParser(description='Image recognition masquerade attack.')
	parser.add_argument('model', metavar='model_file', type=str, help='recognition model (.pkl produced by pylearn2)')
	parser.add_argument('attack_image_file', metavar='attack_image', type=str, help='image modified by attack')
	parser.add_argument('target_label', metavar='target_label', type=int, help='target classification label')
	parser.add_argument('--victim', metavar='VICTIM_IMAGE', type=str, help='example image from target class')
	parser.add_argument('--output_image', metavar='OUTPUT_IMAGE', type=str, default="attack.png", help='output attack image')	
	parser.add_argument('--attack_rate', type=float, default=10.0, help='attack rate (default=10.0)')
	parser.add_argument('--attack_reg', type=float, default=1.0, help='regularization constant (default=1.0)')
	parser.add_argument('--max_epochs', type=int, default=50000, help='maximum number of epochs (default=500000)')
	parser.add_argument('--gauss_kernel', type=int, default=5, help='size of Gaussian smoothing kernel (default=5)')
	parser.add_argument('--status_period', type=int, default=500, help='epochs between status updates (default=500)')
	parser.add_argument('--status_scale', type=int, default=1, help='scale factor for status image (default=1)')	
	parser.add_argument('--make_video', action='store_true', help='make video of attack (default=false)')
	parser.add_argument('--video_interval', type=int, default=100, help='interval (in epochs) between video frames (default=100)')
	parser.add_argument('--video_file', type=str, default='attack.mp4', help='video output file (default=attack.mp4)')
	parser.add_argument('--size_x', type=int, default=None, help='noise mask width')
	parser.add_argument('--size_y', type=int, default=None, help='noise mask height')
	parser.add_argument('--channels', type=int, default=None, help='noise mask channels')
	parser.add_argument('--steptol', type=float, default=1e-3, help='step size tolerance')
	parser.add_argument('--color', action='store_true', help='use color channels (default=false)')
	parser.add_argument('--no_reg_scaling', action='store_true', help='disable regularization scaling (default=false)')
	parser.add_argument('--no_steptol_scaling', action='store_true', help='don\'t scale steptol with regularization (default=false)')
	parser.add_argument('--no_steptol_terminate', action='store_true', help='disable steptol termination')
	parser.add_argument('--random_mask', action='store_true', help='start with a random noise mask (default starts with 0)')
	parser.add_argument('--zero_image', action='store_true', help='make attack image all zeros')
	parser.add_argument('--success_stop', action='store_true', help='terminate on first success')
	parser.add_argument('--blur', action='store_true', help='apply gaussian blur to noise mask')
	parser.add_argument('--skin', action='store_true', help='use skintone detection (color only)')
	parser.add_argument('--no_status_images', action='store_true', help='suppress status images (default=false)')
	parser.add_argument('--mnist', action='store_true', help='use mnist dataset')
	parser.add_argument('--webcam', action='store_true', help='use live webcam for the attack image')
	parser.add_argument('--projector', action='store_true', help='use webcam and projector for the attack image')

	return parser.parse_args()	

def main():	
	
	args = process_args()
	model_file = args.model
	attack_image_file = args.attack_image_file
	victim_image_file = args.victim
	output_image_file = args.output_image
	target_label = args.target_label
	status_period = args.status_period
	attack_rate = args.attack_rate
	attack_reg = args.attack_reg
	max_epochs = args.max_epochs
	gauss_kernel_size = args.gauss_kernel
	status_scale = args.status_scale
	use_reg_scaling = not args.no_reg_scaling
	use_color = args.color
	use_mnist = args.mnist
	use_blur = args.blur
	use_skin = args.skin
	use_random = args.random_mask
	use_status_images = not args.no_status_images
	make_video = args.make_video
	video_interval = args.video_interval
	video_file = args.video_file
	use_zero_image = args.zero_image
	use_webcam = args.webcam
	use_projector = args.projector
	size_x = args.size_x
	size_y = args.size_y
	nchannels = args.channels
	step_tolerance = args.steptol
	tol_scale = not args.no_steptol_scaling
	tol_terminate = not args.no_steptol_terminate
	success_terminate = args.success_stop

	cv2.startWindowThread()	

	if use_mnist:
		(_, _, (mnist_test_x, mnist_test_y)) = load_mnist()
		victim_idx = np.where(mnist_test_y == target_label)[0][1]
		attack_idx = np.where(mnist_test_y == 9)[0][1]
		attack_image = mnist_test_x[attack_idx].reshape((28, 28))*255.0
		victim_image = mnist_test_x[victim_idx].reshape((28,28))*255.0		
	else:
		attack_image = cv2.imread(attack_image_file)		
		victim_image = cv2.imread(victim_image_file) if victim_image_file is not None else None

	model = serial.load(model_file)
	attack_label = onehot_label(model, target_label)

	if use_zero_image:
		attack_image = 0*attack_image

	target_shape = None
	if size_x is not None and size_y is not None and nchannels is not None:
		target_shape = (size_y, size_x, nchannels)

	if use_webcam:
		run_attack_webcam(model, attack_image, attack_label, victim_image, attack_rate=attack_rate, attack_reg=attack_reg, blur=use_blur, 
			gauss_kernel_size=gauss_kernel_size, status_period=status_period, attack_patience=max_epochs, random_start=use_random, 
			out_image=output_image_file, status_scale=status_scale, status_images=use_status_images, make_video=make_video, skin=use_skin,
			video_interval=video_interval, video_file=video_file, reg_scaling=use_reg_scaling, color=use_color, target_shape=target_shape,
			tolerance=step_tolerance, tolerance_scaling=tol_scale, steptol_terminate=tol_terminate, success_terminate=success_terminate)
	elif use_projector:
		run_attack_projector(model, attack_image, attack_label, victim_image, attack_rate=attack_rate, attack_reg=attack_reg, blur=use_blur, 
			gauss_kernel_size=gauss_kernel_size, status_period=status_period, attack_patience=max_epochs, random_start=use_random, 
			out_image=output_image_file, status_scale=status_scale, status_images=use_status_images, make_video=make_video, skin=use_skin,
			video_interval=video_interval, video_file=video_file, reg_scaling=use_reg_scaling, color=use_color, target_shape=target_shape,
			tolerance=step_tolerance, tolerance_scaling=tol_scale, steptol_terminate=tol_terminate, success_terminate=success_terminate)
	else:
		run_attack(model, attack_image, attack_label, victim_image, attack_rate=attack_rate, attack_reg=attack_reg, blur=use_blur, 
			gauss_kernel_size=gauss_kernel_size, status_period=status_period, attack_patience=max_epochs, random_start=use_random, 
			out_image=output_image_file, status_scale=status_scale, status_images=use_status_images, make_video=make_video, skin=use_skin,
			video_interval=video_interval, video_file=video_file, reg_scaling=use_reg_scaling, color=use_color, target_shape=target_shape,
			tolerance=step_tolerance, tolerance_scaling=tol_scale, steptol_terminate=tol_terminate, success_terminate=success_terminate)


if __name__ == '__main__':
	main()
