import sys
import os
import pprint
import math
import csv
import json
import pdb
import numpy
import random
import curses
import argparse
import traceback
import time
import datetime

import Orange as orange

import bigmlutil
import orangeutil
import treeinversion
from treeinversion import whitebox_tree_inversion
from treeinversion import blackbox_tree_inversion
from treeinversion import nominal_perf_stats

from nltk.metrics import ConfusionMatrix
from sklearn import metrics

def mean(x):

  return sum(x)/float(len(x)) if len(x) > 0.0 else 0.0

class InversionStats(object):

  def __init__(self, accuracy=None, aucroc=None, precision=None, recall=None, cm=None, count=None, times=None):

    if accuracy is None:
      accuracy = []
    if aucroc is None:
      aucroc = []
    if precision is None:
      precision = []
    if recall is None:
      recall = []
    if cm is None:
      cm = []
    if count is None:
      count = []
    if times is None:
      times = []

    self.accuracy = accuracy if isinstance(accuracy, list) else [accuracy]
    self.aucroc = aucroc if isinstance(aucroc, list) else [aucroc]
    self.precision = precision if isinstance(precision, list) else [precision]
    self.recall = recall if isinstance(recall, list) else [recall]
    self.cm = cm if isinstance(cm, list) else [cm]
    self.count = count if isinstance(count, list) else [count]
    self.times = times if isinstance(times, list) else [times]

  def update(self, stats):

    assert isinstance(stats.accuracy, list)
    assert isinstance(stats.aucroc, list)
    assert isinstance(stats.precision, list)
    assert isinstance(stats.recall, list)
    assert isinstance(stats.cm, list)
    assert isinstance(stats.count, list)
    assert isinstance(stats.times, list)

    self.accuracy.extend(stats.accuracy)
    self.aucroc.extend(stats.aucroc)
    self.precision.extend(stats.precision)
    self.recall.extend(stats.recall)
    self.cm.extend(stats.cm)
    self.count.extend(stats.count)
    self.times.extend(stats.times)

  def csv(self, aggregate=True):

    if aggregate:
      return "%.4f,%.4f,%.4f,%.4f,%.4f,%.4f\n" % (mean(self.accuracy), mean(self.aucroc), mean(self.precision), mean(self.recall), mean(self.count), mean(self.times))
    else:
      zipped_stats = zip(self.accuracy, self.aucroc, self.precision, self.recall, self.count, self.times)
      return ["accuracy,aucroc,precision,recall,count,time\n"]+["%.4f,%.4f,%.4f,%.4f,%.4f,%.4f\n" % (el[0], el[1], el[2], el[3], el[4], el[5]) for el in zipped_stats]

  def __str__(self, do_tab=1, do_cm=False, do_count=False, do_times=False):

    tabwidth = "\t"*do_tab if do_tab is not None else ""

    if len(self.accuracy) == 0 or len(self.aucroc) == 0 or len(self.precision) == 0 or len(self.recall) == 0 or len(self.count) == 0 or len(self.times) == 0:
      return "%s<no data>" % tabwidth

    cmstr = self.cm[len(self.cm)-1].pp(sort_by_count=True)
    cmstr_tabbed = "\n" + "\n".join(["%s%s" % (tabwidth, splt) for splt in cmstr.split("\n")]).rstrip() if do_cm else ""
    countstr = ", count = %.4f" % mean(self.count) if do_count else ""
    timestr = ", times = %.4f" % mean(self.times) if do_times else ""
    return "%sacc = %.4f, aucroc = %.4f, precision = %.4f, recall = %.4f%s%s%s" % (tabwidth, mean(self.accuracy), mean(self.aucroc), mean(self.precision), mean(self.recall), countstr, timestr, cmstr_tabbed)

def simplestatistics(guesses, data, target, count=0, time_ms=0, do_print=False):

  guesses = [g for g in guesses if g is not None]
  truth = [row[target] for row in data if row[target] is not None]

  try:
    cm = ConfusionMatrix(truth, guesses)
  except:
    pdb.set_trace()
  accuracy = metrics.accuracy_score(truth, guesses)

  if do_print:
    print "accuracy = %.4f, count=%d" % (accuracy, count)
    print
    print cm.pp(sort_by_count=True)

  return InversionStats(accuracy, 0.0, 0.0, 0.0, cm, count, time_ms)

def statistics(guesses, data, config, count=0, time_ms=0, do_flip=False, do_print=False):

  guesses = [str(g) for g in guesses if g is not None]
  target = str(config["target"])
  pos_label = str(config["pos_label"])
  neg_label = str(config["neg_label"])

  truth = [str(row[target]) for row in data if row[target] is not None]

  binary_truth = [1 if val == pos_label else 0 for val in truth]
  binary_guesses = [1 if val == pos_label else 0 for val in guesses]
  if do_flip and metrics.roc_auc_score(binary_truth, binary_guesses) < 0.5:
    old_guesses = binary_guesses[:]
    binary_guesses = [1 if el == 0 else 0 for el in old_guesses]
    old_guesses = guesses[:]
    guesses = [neg_label if el == pos_label else pos_label for el in old_guesses]

  try:
    cm = ConfusionMatrix(truth, guesses)
  except:
    pdb.set_trace()
  aucroc = metrics.roc_auc_score(binary_truth, binary_guesses) if (sum(binary_truth) < len(binary_truth) and sum(binary_truth) > 0) else 0.5
  precision = metrics.precision_score(truth, guesses, pos_label=pos_label)
  recall = metrics.recall_score(truth, guesses, pos_label=pos_label)
  accuracy = metrics.accuracy_score(truth, guesses)

  pos = sum(binary_truth)
  neg = len(truth) - pos
  tp = sum([1 if tr == guess and tr == pos_label else 0 for (tr, guess) in zip (truth, guesses)])
  fp = sum([1 if tr != guess and tr != pos_label else 0 for (tr, guess) in zip (truth, guesses)])
  tn = sum([1 if tr == guess and tr != pos_label else 0 for (tr, guess) in zip (truth, guesses)])
  fn = sum([1 if tr != guess and tr == pos_label else 0 for (tr, guess) in zip (truth, guesses)])
  print "pos=%d, tp=%d, fp=%d, tn=%d, fn=%d" % (pos, tp, fp, tn, fn)
  # pdb.set_trace()

  if do_print:
    print "accuracy = %.4f, aucroc = %.4f, precision = %.4f, recall = %.4f, count=%d" % (accuracy, aucroc, precision, recall, count)
    print
    print cm.pp(sort_by_count=True)

  return InversionStats(accuracy, aucroc, precision, recall, cm, count, time_ms)

def bayesopt_predict(data, sample, config, info):

  target = str(config["target"])
  missing = [str(feat) for feat in config["missing"]]
  missing.append(target)
  known = [str(feat) for feat in info.field_names if not feat in missing]

  matching = {val: 0 for val in info.field_values[target]}
  for row in data:
    matches = reduce(lambda matches, feat: matches and row[feat] == sample[feat], known, True)
    if matches:
      matching[row[target]] += 1 #info.marginals[target][row[target]]

  return max(matching, key=lambda v: matching[v])

def bestcase_eval_tree(config):

  target = str(config["target"])

  if not config.has_key("seed"):
    config["seed"] = 1
  rseed = config["seed"]

  data, values, feats = orangeutil.csv2orangetable(config["datafile"], target, str(config["class"]))
  training = orange.data.Table(data[:])
  bml_training = orangeutil.orangetable2bigml(training)

  tree = orange.regression.tree.TreeLearner(training) #, binarization=0, same_majority_pruning=True, m_pruning=1, max_depth=100)

  bml, bml_model = orangeutil.orange2bigml(tree, training, values)
  info = orangeutil.OrangeModelInfo(bml)

  guesses = []
  for row in bml_training:
    guesses.append(bml_model(row))

  return statistics(guesses, bml_training, config, count=0, time_ms=0, do_flip=False, do_print=False)

def bestcase_eval(config, info):

  target = str(config["target"])

  if not config.has_key("seed"):
    config["seed"] = 1
  rseed = config["seed"]

  data, values, feats = orangeutil.csv2orangetable(config["datafile"], target, str(config["class"]))
  training = orange.data.Table(data[:])
  bml_training = orangeutil.orangetable2bigml(training)

  guesses = []
  for row in bml_training:
    guesses.append(bayesopt_predict(bml_training, row, config, info))

  return statistics(guesses, bml_training, config, count=0, time_ms=0, do_flip=False, do_print=False)

def info_line(config, con, setting, split):

  line = "%s.%s,split=%.3f" % (setting, con, split)

  return line

def init_curses_info(stdscr, settings):

  maxy, maxx = stdscr.getmaxyx()

  window = curses.newpad(maxy*3, maxx)
  curses.curs_set(0)
  curses.use_default_colors()
  curses.init_pair(1, -1, -1)

  cur_line = 0
  lines = {setting: {con: {} for con in ["train", "test"]} for setting in settings}

  return window, lines, maxy, maxx, cur_line

def calibrate_curses_lines(lines, done_splits, stdscr, settings, cur_line, split, display_most_recent=False):

  maxy, maxx = stdscr.getmaxyx()

  if display_most_recent:
    for s in settings:
      for c in ["train", "test"]:
        for d in [d for d in done_splits if d != split]:
          lines[s][c][d] = maxy
    cur_line = 0
    stdscr.clear()

  for setting in settings:
    for con in ["train", "test"]:
      lines[setting][con][split] = cur_line
      cur_line += 2
      if cur_line >= maxy-3:
        cur_line = maxy

  return lines, cur_line

def write_stats(config, window, stdscr, stats, lines, con, split, setting):

  maxy, maxx = stdscr.getmaxyx()

  window.addstr(lines[setting][con][split], 0, info_line(config, con, setting, split), curses.color_pair(1))
  window.addstr(lines[setting][con][split]+1, 0, str(stats[setting][con][split]), curses.color_pair(1))
  window.refresh(0, 0, 0, 0, maxy-1, maxx-1)

def write_stats_plain(config, stats, con, split, setting):

  print info_line(config, con, setting, split)
  print str(stats[setting][con][split])

def write_output(config, settings, stats):

  ret_lines = []
  for setting in settings:
    for con in stats[setting].keys():
      for splt in sorted(stats[setting][con].keys()):
        ret_lines.append(info_line(config, con, setting, splt))
        ret_lines.append(stats[setting][con][splt].__str__(do_count=True, do_times=True))

  if config.has_key("output_dir") and config.has_key("output_prefix"):

    out_prefix = os.path.join(config["output_dir"], "%s-%s" % (datetime.datetime.fromtimestamp(time.time()).strftime('%Y.%m.%d-%H:%M'), config["output_prefix"]))
    if not debug_mode:
      if not os.path.exists(out_prefix):
        os.mkdir(out_prefix)

      with open(os.path.join(out_prefix, "transcript.txt"), "w") as out:
        out.write("\n".join(ret_lines))
        out.write("\n")

    aggregate_lines = ["algorithm,configuration,split,accuracy,aucroc,precision,recall,calls,time\n"]

    for setting in settings:
      for con in stats[setting].keys():
        for splt in sorted(stats[setting][con].keys()):
          aggregate_lines.append("%s,%s,%s,%s" % (setting, con, splt, stats[setting][con][splt].csv(aggregate=True)))

          cur_file = os.path.join(out_prefix, "%s.%s.split=%s.csv" % (setting, con, splt))
          csvlines = stats[setting][con][splt].csv(aggregate=False)
          if not debug_mode:
            with open(cur_file, "w") as out:
              out.writelines(csvlines)

    if not debug_mode:
      with open(os.path.join(out_prefix, "aggstats.csv"), "w") as out:
        out.writelines(aggregate_lines)

    if not debug_mode:
      with open(os.path.join(out_prefix, "config.json"), "w") as out:
        json.dump(config, out, sort_keys=True, indent=4)

  return ret_lines

def timed(method):
    def wrapper(*args, **kw):
        start = int(round(time.time() * 1000))
        result = method(*args, **kw)
        end = int(round(time.time() * 1000))

        time_ms = end - start
        return result, time_ms

    return wrapper

@timed
def timed_whitebox_tree_inversion(data, target, missing, info):
  return whitebox_tree_inversion(data, target, missing, info)
  # return treeinversion.modified_whitebox(data, target, missing, info)

@timed
def timed_blackbox_tree_inversion(model, data, perf_stats, target, objective_name, field_values, missing, marginals):
  return blackbox_tree_inversion(model, data, perf_stats, target, objective_name, field_values, missing, marginals)

def get_model_info(config, split, shuffle=False):

  if config.has_key("tree"):
    return bigmlutil.get_bigml_tree(config, split, shuffle)
  else:
    return orangeutil.train_orange_tree(config, split, shuffle=shuffle)

def main(stdscr, config_file, display_most_recent=False, use_curses=True):

  with open(config_file) as config_fp:
    config = json.load(config_fp)

  target = str(config["target"])
  missing = [str(feat) for feat in config["missing"]]
  iters = int(config["num_iters"]) if config.has_key("num_iters") else 1

  do_whitebox = bool(config["do_whitebox"]) if config.has_key("do_whitebox") else False
  do_blackbox = bool(config["do_blackbox"]) if config.has_key("do_blackbox") else False

  splits = [float(csplit) for csplit in config["split"]] if config.has_key("split") else [1.0]
  do_shuffle = bool(config["do_shuffle"]) if config.has_key("do_shuffle") else False

  settings = []
  if do_whitebox:
    settings.append("whitebox")
  if do_blackbox:
    settings.append("blackbox")

  stats = {}
  for setting in settings:
    stats[setting] = {}
    for con in ["train", "test"]:
      stats[setting][con] = {splt: InversionStats() for splt in splits}

  if use_curses:
    window, lines, maxy, maxx, cur_line = init_curses_info(stdscr, settings)

  done_splits = []
  cur_iter = 0
  total_iters = iters*len(splits)

  counts = {setting: {con: {split: [0] for split in splits} for con in ["train", "test"]} for setting in settings}
  models = {setting: {con: {split: lambda x, cnt=counts[setting][con][split], get_cnt=False: cnt.insert(0, cnt.pop()+1) or bml_model(x) if not get_cnt else cnt[0] for split in splits} for con in ["train", "test"]} for setting in settings}
  except_obj = None

  try:

    for split in splits:

      done_splits.append(split)
      if use_curses:
        lines, cur_line = calibrate_curses_lines(lines, done_splits, stdscr, settings, cur_line, split, display_most_recent)

      for i in range(iters):

        cur_iter += 1
        if use_curses:
          window.addstr(maxy-2, 0, "iteration %-8d of %-8d" % (cur_iter, total_iters))
          window.refresh(0, 0, 0, 0, maxy-1, maxx-1)

        do_split = True if split < 1.0 else False

        configs = ["train"]
        if do_split:
          configs.append("test")

        bml, bml_model, info, bml_training, bml_test = get_model_info(config, split, shuffle=do_shuffle)
        data = {"train": bml_training, "test": bml_test}

        # bestcase_stats = bestcase_eval(config, info)
        # pdb.set_trace()

        if do_blackbox:
          perf_stats = nominal_perf_stats(bml_model, data["train"], info.field_values[info.objective_name], info.objective_name)

        # treeinversion.test_sensitivity(bml_model, bml_training, target, info, ignore=lambda row, targ=target: row[targ] == "No")
        # print "acc [all]:", str(treeinversion.model_accuracy(bml_model, bml_training, info, perf_stats))
        # print "acc [t=No]:", str(treeinversion.model_accuracy(bml_model, bml_training, info, perf_stats, ignore=lambda row, targ=target: row[targ] == "No"))

        for con in configs:

          if do_blackbox:
            guesses, time_ms = timed_blackbox_tree_inversion(models["blackbox"][con][split], data[con], perf_stats, target, info.objective_name, info.field_values, missing, info.marginals)
            stats["blackbox"][con][split].update(statistics(guesses, data[con], config, count=counts["blackbox"][con][split], time_ms=time_ms, do_flip=False, do_print=False))
            if use_curses:
              write_stats(config, window, stdscr, stats, lines, con, split, "blackbox")
            else:
              write_stats_plain(config, stats, con, split, "blackbox")

          if do_whitebox:
            guesses, time_ms = timed_whitebox_tree_inversion(data[con], target, missing, info)
            stats["whitebox"][con][split].update(statistics(guesses, data[con], config, count=counts["whitebox"][con][split], time_ms=time_ms, do_flip=False, do_print=False))
            if use_curses:
              write_stats(config, window, stdscr, stats, lines, con, split, "whitebox")
            else:
              write_stats_plain(config, stats, con, split, "whitebox")

        if use_curses:
          window.clear()
          for setting in settings:
            for con in stats[setting].keys():
              for splt in lines[setting][con].keys():
                write_stats(config, window, stdscr, stats, lines, con, splt, setting)

    if use_curses:
      window.addstr(maxy-2, 0, "iteration %-8d of %-8d" % (cur_iter, total_iters))
      window.addstr(maxy-1, 0, "Press any key to exit", curses.color_pair(1)|curses.A_STANDOUT)
      window.refresh(0, 0, 0, 0, maxy-1, maxx-1)
      window.getkey()
  except BaseException:
    except_obj = traceback.format_exc()
  finally:
    ret_lines = write_output(config, settings, stats)
    return "\n".join(ret_lines), except_obj

def process_args():

  parser = argparse.ArgumentParser(description='Model inversion experiment harness for nominal-valued trees.')
  parser.add_argument('config', metavar='config_file', type=str, help='json configuration script')
  parser.add_argument('--no_history', action='store_true', help='only display most-recent stats')
  parser.add_argument('--no_curses', action='store_true', help='append output to standard out')
  parser.add_argument('--debug', action='store_true', help='disable file output')
  parser.add_argument('--bestcase', action='store_true', help='run bayes-optimal evaluation')

  return parser.parse_args()

debug_mode = False
if __name__ == '__main__':

  args = process_args()
  if args.debug:
    debug_mode = True

  if args.bestcase:
    with open(args.config) as config_fp:
      config = json.load(config_fp)
    bml, bml_model, info, bml_training, bml_test = get_model_info(config, 1.0, shuffle=False)
    bestcase_stats = bestcase_eval(config, info) # bestcase_eval_tree(config, info)
    print bestcase_stats
    quit()

  if not args.no_curses:
    output, except_obj = curses.wrapper(lambda stdscr: main(stdscr, args.config, display_most_recent=args.no_history))
    print output
  else:
    output, except_obj = main(None, args.config, display_most_recent=args.no_history, use_curses=False)

  if except_obj is not None:
    print "\n"
    print except_obj
