import sys
import pprint
import math
import csv
import json
import pdb
import numpy
import random
import time

# import orangeutil

import Orange as orange
from nltk.metrics import ConfusionMatrix

class ModelInfo(object):

  def __init__(self, model_object):

    # Number of rows in the training data
    self.training_size = None

    # Objective feature of the original model --- NOT the objective of model inversion
    self.objective_field = None

    # Text for field names
    self.field_names = None

    '''
    Domain for each feature
    Structure:
      {
        field_name_1: [value_1, value_2, ...]   (in no particular order)
        field_name_2: [value_1, value_2, ...]
        ...
      }
    '''
    self.field_values = None

    '''
    Marginals for each field
    Structure:
      {
        field_name_1:
          {
            value_1: probability (float)
            value_2: probability
            ...
          }
        ...
      }
    '''
    self.marginals = None

    '''
    Mapping from field names to identifiers
    Structure:
      {
        field_name_1: field_id_1 (usually an integer)
        ...
      }
    '''
    self.name_to_id = None

    '''
    Mapping from field identifiers to names
    Structure:
      {
        field_id_1: field_name_1
        ...
      }
    '''
    self.id_to_name = None

    # Name of the model's objective field
    self.objective_name = None

    # Index of the model's objective field in the field_names list
    self.objective_index = None

    # Root of the decision tree
    self.root = None

def paths(node, path=[]):

  if not node.has_key("children"):
    yield path+[node]
  else:
    for child in node["children"]:
      for newpath in paths(child, path+[node]):
        yield newpath

def objective_configs(path, info):

  leaf = path[len(path)-1]
  assert leaf.has_key("objective_summary")

  for cat in leaf["objective_summary"]["categories"]:
    for i in range(int(cat[1])):
      yield {info.objective_name: cat[0]}

def node_str(node, info, tabs=""):

  if not isinstance(node["predicate"], dict):
    r = "(root)"
  else:
    count = node["count"]
    op = node["predicate"]["operator"]
    pfield = info.id_to_name[node["predicate"]["field"]]
    pval = node["predicate"]["value"]
    r = "%s-> [%s %s %s] n=%s" % (tabs, pfield, op, pval, str(count))
    if not node.has_key("children"):
      for cat in node["objective_summary"]["categories"]:
        r = "%s%s\n%s\t-> [%s = %s] n=%s" % (r, tabs, tabs, info.objective_name, cat[0], str(cat[1]))

  return r

def path_str(path, info):

  r = ""

  for node in path:
    r = "%s%s\n" % (r, node_str(node, info))

  return r

def row_str(row):

  r = ""
  for field in row.keys():
    r = "%s\t%s = %s\n" % (r, field, row[field])

  return r

def rows_equal(r1, r2, info):

  for field in info.field_names:
    if r1[field] != r2[field]:
      return False
  return True

def data_contains(data, row, info):

  cnt = 0
  for r in data:
    if rows_equal(row, r, info):
      cnt += 1
  return cnt

def mode_from_samples(samples, field, info):

  counts = {val: info.marginals[field][val] for val in info.field_values[field]}

  for samp in samples:
    counts[samp[field]] += 1.0

  return max(counts, key=lambda v: counts[v])

def samples_from_path_and_partial(path, partial_config, target, info):

  samples = []
  possible_values = {field: info.field_values[field][:] if field not in partial_config.keys() else [partial_config[field]] for field in info.field_names}

  for node in path:

    assert node.has_key("predicate")

    predicate = node["predicate"]
    if not isinstance(predicate, dict):
      continue

    fieldname = info.id_to_name[predicate["field"]]
    if predicate["operator"] == "=":
      if fieldname in partial_config.keys():
        if predicate["value"] != partial_config[fieldname]:
          possible_values[fieldname] = []
      else:
        possible_values[fieldname] = [predicate["value"]]
    elif predicate["operator"] == "!=":
      if predicate["value"] in possible_values[fieldname]:
        possible_values[fieldname].remove(predicate["value"])
    elif predicate["operator"] == "<=":
      to_remove = []
      for val in possible_values[fieldname]:
        if not float(val) <= float(predicate["value"]):
          to_remove.append(val)
      for val in to_remove:
        possible_values[fieldname].remove(val)
    elif predicate["operator"] == "<":
      to_remove = []
      for val in possible_values[fieldname]:
        if not float(val) < float(predicate["value"]):
          to_remove.append(val)
      for val in to_remove:
        possible_values[fieldname].remove(val)
    elif predicate["operator"] == ">=":
      to_remove = []
      for val in possible_values[fieldname]:
        if not float(val) >= float(predicate["value"]):
          to_remove.append(val)
      for val in to_remove:
        possible_values[fieldname].remove(val)
    elif predicate["operator"] == ">":
      to_remove = []
      for val in possible_values[fieldname]:
        if not float(val) > float(predicate["value"]):
          to_remove.append(val)
      for val in to_remove:
        possible_values[fieldname].remove(val)
    else:
      assert False and "Unrecognized predicate: %s." % predicate["operator"]

  predicate_sizes = [len(possible_values[field]) for field in possible_values.keys()]
  if min(predicate_sizes) == 0:
    return (possible_values[target], 0)

  leaf = path[len(path)-1]
  assert leaf.has_key("objective_summary")
  n = sum([int(cat[1]) for cat in leaf["objective_summary"]["categories"]])
  for cat in leaf["objective_summary"]["categories"]:
    if info.objective_name in partial_config.keys() and cat[0] != partial_config[info.objective_name]:
      n -= int(cat[1])

  return (possible_values[target], n)

'''
  White-box tree inversion for trees with only nominal features.

  arguments:
    data: dataset to evaluate on
    target: feature name for inversion target
    missing: list of missing attributes
    info: ModelInfo class for tree to invert
'''
def whitebox_tree_inversion(data, target, missing, info):

  N = float(info.root["count"])
  known = [feature for feature in info.field_names if (feature not in missing and feature != target)]
  guesses = []
  n_paths = 0

  times = []

  for row in data:
    start = int(round(time.time() * 1000))
    samples = []
    partial = {field: row[field] for field in known if row.has_key(field)}

    allpaths = paths(info.root)
    n_paths = 0
    for path in allpaths:
      n_paths += 1
      newsamps = samples_from_path_and_partial(path, dict(partial), target, info)
      if newsamps[1] > 0: # this doesn't affect anything but performance
        samples.append(newsamps)
    # p_all = float(sum([samp[1] for samp in samples]))/N
    p_v = {v: 0.0 for v in info.field_values[target]}    

    for v in info.field_values[target]:
      denom = sum([float(samp[1])/N if v in samp[0] else 0.0 for samp in samples])
      if denom == 0.0:
        p_v[v] = 0.0
        continue
      for samp in samples:
        p_i = float(samp[1])/N
        num = p_i*info.marginals[target][v] if v in samp[0] else 0.0
        p_v[v] += (num/denom)

    tied = True if sum([abs(p_v.values()[0] - v) for v in p_v.values()]) == 0.0 else False

    if not tied:
      guess = max(p_v, key=lambda v: p_v[v])
    else:
      guess = max(info.marginals[target], key=lambda v: info.marginals[target][v])

    guesses.append(guess)
    end = int(round(time.time() * 1000))  
    time_ms = end - start    
    times.append(time_ms)
    # print time_ms, "runtime"

  print float(sum(times))/float(len(times)), "ms on average"

  return guesses

'''
  Computes "performance statistics" for a nominal-valued model.
  Performance statistics estimate Pr[x_class = y | tree(x) = y']

  arguments:
    model: the nominal-valued model
    data: the training data for the model
    dom: domain for features in the training data
    classfeat: objective class feature for model
'''
def nominal_perf_stats(model, data, dom, classfeat):

  truth = [row[classfeat] for row in data]
  guesses = [model(row) for row in data]

  cm = ConfusionMatrix(truth, guesses)
  inds = cm._indices
  mat = cm._confusion

  stats = {val: {} for val in dom}
  # stats = {val: {val2: 0.0 if val2 != val else 1.0 for val2 in dom} for val in dom}

  for given in dom:
    if given not in inds.keys():
      for out in dom:
        stats[given][out] = 0.0
      continue
    n = float(sum([matrow[inds[given]] for matrow in mat]))
    if n == 0.0:
      for out in dom:
        stats[given][out] = 0.0
      continue
    for out in dom:
      if out not in inds.keys():
        stats[given][out] = 0.0
      else:
        stats[given][out] = mat[inds[out]][inds[given]]/n

  return stats

'''
  Black-box inversion for trees with only nominal features.

  arguments:
    model: the tree model (same format as BigML's json models)
    data: dataset to evaluate on
    perf_stats: performance statistics derived from model's performance on the training set (as generated by nominal_perf_stats)
    target: feature name for inversion target
    classfeat: feature name for model objective class feature
    doms: domain of each feature in the data set
    missing: list of unknown features
    margs: marginal distributions for training data
'''
gcnt = 0
bcnt = 0
def blackbox_tree_inversion(model, data, perf_stats, target, classfeat, doms, missing=None, margs=None):

  def fillin(row, missing, doms):

    if missing is None or len(missing) == 0:
      yield row
    else:
      curattr = missing[0]
      for val in doms[curattr]:
        newrow = dict(row)
        newrow[curattr] = val
        for yrow in fillin(newrow, missing[1:], doms):
          yield yrow

  def joint_prob(row, marg_attrs, margs):

    p = 0.0
    for attr in marg_attrs:
      p += math.log(margs[attr][row[attr]]) if margs[attr][row[attr]] > 0.0 else 0.0

    return math.exp(p)

  def feasible_target(model, perf_stats, row, target, classfeat, doms, missing=None, margs=None, data=None):

    global gcnt
    global bcnt
    original_prediction = model(row)

    allmissing = [] if missing is None else missing[:]
    allmissing.append(target)
    X = fillin(row, allmissing, doms)

    Y = {val: 0.0 for val in doms[target]}
    for x in X:
      # if model(x) == model(row):
      #   Y[x[target]] += joint_prob(x, allmissing, margs)
      Y[x[target]] += joint_prob(x, allmissing, margs) * perf_stats[model(x)][row[classfeat]]
      # pdb.set_trace()

    return Y

  def predict_target(target, feasible, doms, margs):

    if sum([feasible[v] for v in doms[target]]) == 0:
      return max(margs[target], key=lambda v: margs[target][v])

    return max(feasible, key=lambda v: feasible[v])

  guesses = []

  # pdb.set_trace()
  for row in data:
    feasible_set = feasible_target(model, perf_stats, row, target, classfeat, doms, missing, margs)
    guess = predict_target(target, feasible_set, doms, margs)
    guesses.append(guess)

  return guesses

def predicate_matches(predicate, row, info):

  fieldname = info.id_to_name[predicate["field"]]
  if not row.has_key(fieldname):
    return True

  val = predicate["value"]
  rowval = row[fieldname]

  if predicate["operator"] == "=":
    return val == rowval
  elif predicate["operator"] == "!=":
    return val != rowval
  elif predicate["operator"] == "<=":
    return float(rowval) <= float(val)
  elif predicate["operator"] == "<":
    return float(rowval) < float(val)
  elif predicate["operator"] == ">=":
    return float(rowval) >= float(val)
  elif predicate["operator"] == ">":
    return float(rowval) > float(val)

  assert False and "Unknown predicate operator: \"%s\"" % predicate["operator"]

def path_matches(path, row, info):

  predicates = [node["predicate"] for node in path[1:]]

  leaf = path[len(path)-1]
  leaf_pred = {"operator": "=", "value": leaf["output"], "field": info.name_to_id[info.objective_name]}
  predicates.append(leaf_pred)

  return reduce(lambda matches, predicate: matches and predicate_matches(predicate, row, info), predicates, True)

def path_samples(path):

  leaf = path[len(path)-1]
  output_value = leaf["output"]
  category = [cat for cat in leaf["objective_summary"]["categories"] if cat[0] == output_value]

  return int(category[0][1])

def path_mentions_field(path, field, info):

  for node in path[1:]:
    assert node.has_key("predicate")
    predicate = node["predicate"]
    fieldname = info.id_to_name[predicate["field"]]
    if fieldname != field:
      continue
    if predicate["operator"] == "=":
      return [predicate["value"]]
    elif predicate["operator"] == "!=":
      return [val for val in info.field_values[field] if val != predicate["value"]]
    else:
      assert False and "Unsupported predicate in \"path mentions\" lookup: %s" % predicate["operator"]

  return info.field_values[field][:]

def break_ties(val_probs, target, info):
  
  val_probs = dict(val_probs)
  maxp = max(val_probs.values())
  nmax = sum([1 if val_probs[v] == maxp else 0 for v in val_probs.keys()])
  if nmax > 1 and info.marginals is not None:
    for val in info.field_values[target]:
      if val_probs[val] == maxp:
        val_probs[val] = (float(val_probs[val]) if maxp > 0 else 1.0)*info.marginals[target][val]
      else:
        val_probs[val] = 0.0

  return max(val_probs, key=lambda v: val_probs[v])

def modified_whitebox(data, target, missing, info):

  guesses = []
  if target not in missing:
    missing.append(target)

  for row in data:
    target_vals = {val: 0.0 for val in info.field_values[target]}
    allpaths = paths(info.root)
    mod_row = dict(row)
    for field in missing:
      del mod_row[field]
    for path in allpaths:
      mentioned = path_mentions_field(path, target, info)
      if mentioned is None:
        continue      
      if path_matches(path, mod_row, info):
        # print path_str(path, info)
        n = path_samples(path)
        total_mass = sum([info.marginals[target][val] for val in mentioned])
        for val in mentioned:
          target_vals[val] += float(n)*(info.marginals[target][val]/total_mass)
        # pdb.set_trace()

    guess = break_ties(target_vals, target, info)
    # print target_vals, guess
    # pdb.set_trace()
    guesses.append(guess)

  return guesses

def model_accuracy(model, data, info, perf_stats, ignore=lambda x: False):

  p = []
  correct = 0
  total = 0
  for row in data:
    if ignore(row):
      continue
    p.append(perf_stats[model(row)][row[info.objective_name]])
    if model(row) == row[info.objective_name]:
      correct += 1
    total += 1

  return float(correct)/float(total), sum(p)/float(len(p))

def test_sensitivity(model, data, target, info, ignore=lambda x: False):

  n = []

  for row in data:
    cur_n = 0
    original_prediction = model(row)

    if ignore(row):
      continue

    for val in info.field_values[target]:
      newrow = dict(row)
      newrow[target] = val
      # if not model(newrow) == row[info.objective_name]:
      #   continue
      if model(newrow) == original_prediction:
        cur_n += 1
    n.append(cur_n)

  cnts = {}
  for i in range(len(info.field_values[target])+1):
    cnts[i] = sum([1 if cur_n == i else 0 for cur_n in n])

  avg = sum(n)/float(len(n)) if len(n) > 0 else 0.0

  pdb.set_trace()
  return n, cnts, avg
