import cv2
import numpy as np
import theano

def normalize(X, low, high, dtype=theano.config.floatX):
    """Normalizes a given array in X to a value between low and high."""
    X = np.asarray(X)
    minX, maxX = np.min(X), np.max(X)    
    # normalize to [0...1].
    X = X - float(minX)
    X = X / float((maxX - minX))
    # scale to [low...high].
    X = X * (high-low)
    X = X + low
    if dtype is None:
        return np.asarray(X)
    return np.asarray(X, dtype=dtype)

class WebcamFace(object):

    def __init__(self, size=None, grayscale=True, camera_index=0, cascade='../classifiers/haarcascade_frontalface_alt.xml',
                    xoffset=-30, yoffset=-70, woffset=40, hoffset=120, fix_frame=True, normalize255=True):

        self.size = size
        self.face_cascade = cv2.CascadeClassifier(cascade)
        self.camera_index = 0
        self.capture = None
        self.xoffset = xoffset
        self.yoffset = yoffset
        self.woffset = woffset
        self.hoffset = hoffset
        self.grayscale = grayscale
        self.fix_frame = fix_frame
        self.normalize255 = normalize255

        self.lx = None
        self.ly = None
        self.lw = None
        self.lh = None

        self.last_frame = None

    def begin_capture(self):

        if self.capture is None:
            self.capture = cv2.VideoCapture(self.camera_index)
            
            # let the camera adjust to the room's lighting.
            # running the attack in a dark room resulted in some of
            # the most confusing errors I've ever encountered.
            meanf = 0.0
            while meanf < 10:
                _, frame = self.capture.read()
                meanf = np.mean(frame)

    def end_capture(self):

        if self.capture is not None:
            self.capture.release()
            self.capture = None

    def set_size(self, size):

        self.size = size

    def get_frame(self, doKill=False):

        if self.capture is None:
            return self.last_frame

        _, frame = self.capture.read()

        faces = ()
        gray_frame = None
        if self.lx is None and self.fix_frame:
            gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            faces = self.face_cascade.detectMultiScale(gray_frame, 1.3, 5)

        x, y, w, h = self.lx, self.ly, self.lw, self.lh

        if self.grayscale and gray_frame is None:
            gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
            frame = gray_frame

        if len(faces) > 0:
            (x, y, w, h) = faces[0]
            x += self.xoffset
            y += self.yoffset
            w += self.woffset
            h += self.hoffset
            x, y, w, h = max(0, x), max(0, y), max(0, w), max(0, h)
            self.lx, self.ly, self.lw, self.lh = x, y, w, h
        else:
            x, y, w, h = self.lx, self.ly, self.lw, self.lh

        if x is not None and y is not None and w > 0 and h > 0:            
            frame = frame[y:max(y+h, y+100), x:max(x+w, x+100)]
            self.last_Frame = frame            

        if self.size is not None:
            frame = cv2.resize(frame, self.size, interpolation=cv2.INTER_NEAREST)

        if self.normalize255:
            frame = normalize(frame, 0, 1, theano.config.floatX)

        return frame
